@@ -19,9 +19,10 @@ import subprocess
 import sys
 import traceback
 from pathlib import Path
-from typing import Optional
+from typing import Optional, List
 
 import click
+import rich
 
 from . import DefaultNames, __version__, get_home_config_path, is_version_unreleased
 from . import init as initmod
@@ -33,7 +34,7 @@ from .support import Status
 from .util import filter_env, get_package_digest
 
 _latestJobs = []  # for testing
-_args = []  # for testing
+_args: List[str] = []  # for testing
 
 
 def option_group(*options):
@@ -64,10 +65,12 @@ def option_group(*options):
     is_flag=True,
     help="Only output errors to the stdout",
 )
-@click.option("--logfile",
-  default=None,
-  envvar="UNFURL_LOGFILE",
-  help="Log messages to file (at DEBUG level)")
+@click.option(
+    "--logfile",
+    default=None,
+    envvar="UNFURL_LOGFILE",
+    help="Log messages to file (at DEBUG level)",
+)
 @click.option(
     "--tmp",
     envvar="UNFURL_TMPDIR",
@@ -120,15 +123,20 @@ def cli(
             msg = f"{sys.executable} -m pip install -U 'git+https://github.com/onecommons/unfurl.git#egg=unfurl'"
         else:
             msg = f"{sys.executable} -m pip install -U unfurl=={version_check}"
-        logging.error("Use --no-version-check to ignore or run this command to upgrade: \n %s", msg)
+        logging.error(
+            "Use --no-version-check to ignore or run this command to upgrade: \n %s",
+            msg,
+        )
         raise click.Abort()
 
+
 def detect_log_level(loglevel: Optional[str], quiet: bool, verbose: int) -> Levels:
     if quiet:
         effective_log_level = Levels.CRITICAL
     else:
-        if os.getenv("UNFURL_LOGGING"):
-            effective_log_level = Levels[os.getenv("UNFURL_LOGGING").upper()]
+        loglevel_env = os.getenv("UNFURL_LOGGING")
+        if loglevel_env:
+            effective_log_level = Levels[loglevel_env.upper()]
         else:
             levels = [Levels.INFO, Levels.VERBOSE, Levels.DEBUG, Levels.TRACE]
             effective_log_level = levels[min(verbose, 3)]
@@ -175,12 +183,7 @@ readonlyJobControlOptions = option_group(
         "--jobexitcode",
         type=click.Choice(["error", "degraded", "never"]),
         default="never",
-        help="Set exit code to 1 if job status is not ok. (Default: never)",
-    ),
-    click.option(
-        "--use-environment",
-        default=None,
-        help="Run this job in the given environment.",
+        help="Set exit code to 64 if job ends at given status. (Default: never)",
     ),
 )
 jobControlOptions = option_group(
@@ -212,6 +215,11 @@ commonJobFilterOptions = option_group(
         is_flag=True,
         help="(Re)run operation regardless of instance's status or state",
     ),
+    click.option(
+        "--use-environment",
+        default=None,
+        help="Run this job in the given environment.",
+    ),
 )
 destroyUnmanagedOption = click.option(
     "--destroyunmanaged",
@@ -264,7 +272,7 @@ def _get_runtime(options, ensemblePath):
     return runtime, localEnv
 
 
-def _run(ensemble, options, workflow=None):
+def _run(ensemble: str, options, workflow=None):
     if workflow:
         options["workflow"] = workflow
 
@@ -375,9 +383,7 @@ class DockerCmd:
         cmd.extend(self.env_vars_to_args())
         cmd.extend(self.default_volumes())
         cmd.extend(self.docker_args)
-        cmd.extend(
-            [self.image, "unfurl", "--no-runtime"]
-        )
+        cmd.extend([self.image, "unfurl", "--no-runtime"])
         return cmd
 
     def env_vars_to_args(self) -> list:
@@ -414,7 +420,7 @@ def _run_remote(runtime, options, localEnv):
         print(f"TESTING: running remote with _args {_args}")
 
     if options.get("no_version_check"):
-        cmdLine.remove('--no-version-check')
+        cmdLine.remove("--no-version-check")
         version_check = []
     else:
         version_check = ["--version-check", __version__(True)]
@@ -432,7 +438,7 @@ def _print_summary(job, options):
     jsonSummary = {}
     summary = options.get("output")
     if summary == "text" and not job.jobOptions.planOnly:
-        click.echo(job.summary())
+        job.print_summary_table()
     elif summary == "json":
         if job.jobOptions.planOnly:
             jsonSummary = job._json_plan_summary()
@@ -447,8 +453,11 @@ def _print_summary(job, options):
             jsonSummary["result"] = result
         else:
             click.echo("query: " + query)
-            click.echo(result)
-    if jsonSummary is not None:
+            if result is None:
+                click.echo("No results found")
+            else:
+                click.echo(result)
+    if jsonSummary:
         click.echo(json.dumps(jsonSummary, indent=2))
 
 
@@ -479,9 +488,9 @@ def _stop_logging(job, options, verbose, tmplogfile):
         click.echo("Done, full log written to " + log_path)
 
 
-def _run_local(ensemble, options):
+def _run_local(ensemble: str, options):
     logger = logging.getLogger("unfurl")
-    logger.verbose('Running command: %s', sys.argv[1:])
+    logger.verbose("Running command: %s", sys.argv[1:])  # type: ignore
     verbose = options.get("verbose", 0)
     tmplogfile = options["logfile"]
     job, rendered, proceed = start_job(ensemble, options)
@@ -507,15 +516,16 @@ def _run_local(ensemble, options):
 
 
 def _exit(job, options):
+    # https://tldp.org/LDP/abs/html/exitcodes.html
     if not job or (
         "jobexitcode" in options
         and options["jobexitcode"] != "never"
         and Status[options["jobexitcode"]] <= job.status
     ):
         if options.get("standalone_mode") is False:
-            return 1
+            return 64
         else:
-            sys.exit(1)
+            sys.exit(64)
     else:
         return 0
 
@@ -800,38 +810,64 @@ def home(ctx, init=False, render=False, replace=False, **options):
 @cli.command(short_help="Print or manage the project's runtime")
 @click.pass_context
 @click.option("--init", default=False, is_flag=True, help="Create a new runtime")
+@click.option(
+    "--update",
+    default=False,
+    is_flag=True,
+    help="Update Python requirements to match this instance of Unfurl.",
+)
 @click.argument(
     "project_folder",
     type=click.Path(exists=False),
     default=".",
 )
-def runtime(ctx, project_folder, init=False, **options):
+def runtime(ctx, project_folder, init=False, update=False, **options):
     """If no options are set, display the runtime currently used by the project. To create
     a new runtime in the project root use --init and the global --runtime option.
     """
     options.update(ctx.obj)
     project_path = os.path.abspath(project_folder)
+    runtime_ = options.get("runtime") or "venv:"
     if not init:
-        # just display the current runtime
         try:
-            runtime_, local_env = _get_runtime(options, project_folder)
+            # XXX if --runtime is set venv_location likely is wrong
+            venv_location, local_env = _get_runtime(options, project_folder)
         except:
             # if the current path isn't a folder
             if project_folder == ".":
-                runtime_, local_env = _get_runtime(
+                venv_location, local_env = _get_runtime(
                     options, get_home_config_path(options.get("home"))
                 )
             else:
                 raise
-        click.echo(f"\nCurrent runtime: {runtime_}")
-        return
+        if venv_location:
+            # venv_location will be "venv:project/path/.venv" or None
+            project_path = os.path.dirname(venv_location[len("venv:") :])
+        if not update:
+            # just display the current runtime
+            click.echo(f"\nCurrent runtime: {venv_location}")
+            return
+
+    if init:
+        error = initmod.init_engine(project_path, runtime_)
+    else:  # update
+        venv_path = Path(venv_location[len("venv:") :])
+        env = _venv(venv_path, None)
+        import importlib.metadata
+
+        packages = [
+            req
+            for req in importlib.metadata.requires("unfurl")
+            if "extra ==" not in req
+        ]
+        error = subprocess.run(["pipenv", "install"] + packages, env=env).returncode
 
-    runtime_ = options.get("runtime") or "venv:"
-    error = initmod.init_engine(project_path, runtime_)
     if not error:
-        click.echo(f'Created runtime "{runtime_}" in "{project_path}"')
+        action = "Updated" if update else "Created"
+        click.echo(f'{action} runtime "{runtime_}" in "{project_path}"')
     else:
-        click.echo(f'Failed to create runtime "{runtime_}": {error}')
+        action = "update" if update else "create"
+        click.echo(f'Failed to {action} runtime "{runtime_}": {error}')
 
 
 @cli.command(short_help="Clone a project, ensemble or service template")
@@ -876,6 +912,13 @@ def runtime(ctx, project_folder, init=False, **options):
     "--use-deployment-blueprint",
     help="Use this deployment blueprint.",
 )
+@click.option(
+    "--var",
+    nargs=2,
+    type=click.Tuple([str, str]),
+    multiple=True,
+    help="name/value pair to pass to skeleton (multiple times ok).",
+)
 def clone(ctx, source, dest, **options):
     """Create a new ensemble or project from a service template or an existing ensemble or project.
 
@@ -969,11 +1012,19 @@ def get_commit_message(committer):
     is_flag=True,
     help="Don't add files for committing (user must add using git)",
 )
+@click.option(
+    "--use-environment",
+    default=None,
+    help="Use this environment.",
+)
 def commit(ctx, project_or_ensemble_path, message, skip_add, no_edit, **options):
-    """Commit any outstanding changes to the given project or ensemble."""
+    """Commit any changes to the given project or ensemble."""
     options.update(ctx.obj)
     localEnv = LocalEnv(
-        project_or_ensemble_path, options.get("home"), can_be_empty=True
+        project_or_ensemble_path,
+        options.get("home"),
+        can_be_empty=True,
+        override_context=options.get("use_environment") or "",
     )
     if localEnv.manifestPath:
         committer = localEnv.get_manifest()
@@ -1002,11 +1053,19 @@ def commit(ctx, project_or_ensemble_path, message, skip_add, no_edit, **options)
     is_flag=True,
     help="Only show repositories with uncommitted changes",
 )
+@click.option(
+    "--use-environment",
+    default=None,
+    help="Use this environment.",
+)
 def git_status(ctx, project_or_ensemble_path, dirty, **options):
     "Show the git status for paths relevant to the given project or ensemble."
     options.update(ctx.obj)
     localEnv = LocalEnv(
-        project_or_ensemble_path, options.get("home"), can_be_empty=True
+        project_or_ensemble_path,
+        options.get("home"),
+        can_be_empty=True,
+        override_context=options.get("use_environment") or "",
     )
     if localEnv.manifestPath:
         committer = localEnv.get_manifest()
@@ -1025,24 +1084,38 @@ def git_status(ctx, project_or_ensemble_path, dirty, **options):
 @click.option(
     "--format",
     default="deployment",
-    type=click.Choice(["blueprint", "environments", "deployment"]),
+    type=click.Choice(["blueprint", "environments", "deployment", "deployments"]),
 )
 @click.option(
     "--use-environment",
     default=None,
     help="Export using this environment.",
 )
-def export(ctx, project_or_ensemble_path, format, **options):
+@click.option(
+    "--file",
+    default=None,
+    help="Write json export to this file instead of the console.",
+)
+def export(ctx, project_or_ensemble_path, format, file, **options):
     """Export ensemble in a simplified json format."""
     from . import to_json
 
     options.update(ctx.obj)
-    localEnv = LocalEnv(project_or_ensemble_path, options.get("home"), override_context=options.get("use_environment"))
+    localEnv = LocalEnv(
+        project_or_ensemble_path,
+        options.get("home"),
+        override_context=options.get("use_environment") or "",
+    )
     exporter = getattr(to_json, "to_" + format)
-    jsonSummary = exporter(localEnv)
-    click.echo(json.dumps(jsonSummary, indent=2))
+    jsonSummary = exporter(localEnv, file)
+    output = json.dumps(jsonSummary, indent=2)
+    if file:
+        with open(file, "w") as f:
+            f.write(output)
+    else:
+        click.echo(output)
     logger = logging.getLogger("unfurl")
-    logger.info('Export complete.')
+    logger.info("Export complete.")
 
 
 @cli.command()
@@ -1050,20 +1123,40 @@ def export(ctx, project_or_ensemble_path, format, **options):
 @click.argument("ensemble", default=".", type=click.Path(exists=False))
 @click.option("--query", help="Run the given expression")
 @click.option("--trace", default=0, help="Set the query's trace level")
+@click.option(
+    "--use-environment",
+    default=None,
+    help="Use this environment.",
+)
 def status(ctx, ensemble, **options):
-    """ "Show the status of deployed resources in the given ensemble."""
+    """Show the status of deployed resources in the given ensemble."""
     options.update(ctx.obj)
-    localEnv = LocalEnv(ensemble, options.get("home"))
+    localEnv = LocalEnv(
+        ensemble,
+        options.get("home"),
+        override_context=options.get("use_environment") or "",
+    )
     manifest = localEnv.get_manifest()
     click.echo(manifest.status_summary())
     query = options.get("query")
     if query:
-        from .eval import eval_for_func, RefContext
-
         trace = options.get("trace")
-        result = eval_for_func(query, RefContext(manifest.rootResource, trace=trace))
+        result = manifest.rootResource.query(query, trace=trace)
         click.echo("query: " + query)
-        click.echo(result)
+        if result is None:
+            click.echo("No results found")
+        else:
+            click.echo(result)
+
+
+@cli.command()
+@click.pass_context
+@click.argument("ensemble", default=".", type=click.Path(exists=False))
+def validate(ctx, ensemble, **options):
+    """Validate the given ensemble."""
+    options.update(ctx.obj)
+    localEnv = LocalEnv(ensemble, options.get("home"))
+    localEnv.get_manifest()
 
 
 @cli.command()
@@ -1110,6 +1203,29 @@ def help(ctx, cmd=""):
         raise click.UsageError(f"no help found for unknown command '{cmd}'", ctx=ctx)
 
 
+@cli.command()
+@click.pass_context
+@click.argument("project_or_ensemble_path", default=".", type=click.Path(exists=True))
+@click.option("--port", default=8081, help="Port to listen on")
+@click.option("--address", default="localhost", help="Host to listen on")
+@click.option(
+    "--secret",
+    envvar="UNFURL_SERVE_SECRET",
+    help="Secret required to access the server",
+)
+@click.option(
+    "--clone-root",
+    default=".",
+    help="Where to clone all repositories",
+    type=click.Path(exists=True),
+)
+def serve(ctx, port, address, secret, clone_root, project_or_ensemble_path, **options):
+    options.update(ctx.obj)
+    from .server import serve
+
+    serve(address, port, secret, clone_root, project_or_ensemble_path, options)
+
+
 def main():
     obj = {"standalone_mode": False}
     try:
