@@ -1,19 +1,32 @@
 # Copyright (c) 2020 Adam Souzis
 # SPDX-License-Identifier: MIT
-from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple, Union, ValuesView, cast
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Dict,
+    Generator,
+    List,
+    Optional,
+    Tuple,
+    Union,
+    ValuesView,
+    cast,
+    MutableSequence,
+)
 from typing_extensions import TypedDict
 import six
-from collections.abc import Mapping, MutableSequence
+from collections.abc import Mapping
 import os
 import copy
 
-from unfurl.logs import UnfurlLogger
+from unfurl.logs import UnfurlLogger, Levels, LogExtraLevels
+
 if TYPE_CHECKING:
     from unfurl.manifest import Manifest
     from unfurl.job import ConfigTask
 
 
-from .support import Status, ResourceChanges, Priority, TopologyMap
+from .support import Status, ResourceChanges, Priority, set_context_vars
 from .result import (
     ChangeRecord,
     ResultsList,
@@ -33,11 +46,11 @@ from .util import (
     to_enum,
     wrap_sensitive_value,
     sensitive,
-    truncate_str
+    truncate_str,
 )
 from . import merge
 from .eval import Ref, map_value, RefContext
-from .runtime import NodeInstance, RelationshipInstance, Operational
+from .runtime import EntityInstance, NodeInstance, RelationshipInstance, Operational
 from .yamlloader import yaml
 from .projectpaths import WorkFolder, Folders
 from .planrequests import (
@@ -48,12 +61,14 @@ from .planrequests import (
     find_operation_host,
     create_instance_from_spec,
 )
+from .tosca import find_env_vars
 
 import logging
 
 # Tell mypy the logger is of type UnfurlLogger
 logger = cast(UnfurlLogger, logging.getLogger("unfurl.task"))
 
+
 class ConfiguratorResult:
     """
     Modified indicates whether the underlying state of configuration,
@@ -66,11 +81,11 @@ class ConfiguratorResult:
         self,
         success: bool,
         modified: Optional[bool],
-        status: Optional[Status]=None,
-        configChanged: bool=None,
-        result: object=None,
-        outputs: Optional[dict]=None,
-        exception: Optional[UnfurlTaskError]=None,
+        status: Optional[Status] = None,
+        configChanged: bool = None,
+        result: object = None,
+        outputs: Optional[dict] = None,
+        exception: Optional[UnfurlTaskError] = None,
     ) -> None:
         self.modified = modified
         self.status = to_enum(Status, status)
@@ -123,7 +138,7 @@ class Configurator:
     as an alternative to using the full name ("module.class") when setting the implementation on an operation.
     (Titlecase recommended)"""
 
-    exclude_from_digest = ()
+    exclude_from_digest: Tuple[str, ...] = ()
 
     @classmethod
     def set_config_spec_args(klass, kw: dict, target: None) -> dict:
@@ -224,6 +239,7 @@ class Configurator:
         values = [inputs[key] for key in keys]
 
         for dep in task.dependencies:
+            assert isinstance(dep, Dependency)
             if not isinstance(dep.expected, sensitive):
                 keys.append(dep.expr)
                 values.append(dep.expected)
@@ -299,7 +315,7 @@ class Configurator:
                 "digests didn't match for %s with %s: old %s, new %s",
                 task.target.name,
                 _parameters,
-                changeset.digestValue, # type: ignore
+                changeset.digestValue,  # type: ignore
                 newDigest,
             )
         return mismatch
@@ -312,12 +328,15 @@ class MockConfigurator(Configurator):
     def can_dry_run(self, task: "TaskView") -> bool:
         return True
 
+
 class _ConnectionsMap(dict):
     def by_type(self) -> ValuesView:
         # return unique connection by type
         # reverse so nearest relationships replace less specific ones that have matching names
+        # XXX why is rel sometimes a Result?
         by_type = {  # the list() is for Python 3.7
-            rel.type: rel for rel in reversed(list(self.values()))
+            rel.resolved.type if isinstance(rel, Result) else rel.type: rel
+            for rel in reversed(list(self.values()))
         }
         return by_type.values()
 
@@ -325,6 +344,7 @@ class _ConnectionsMap(dict):
         # the more specific connections are inserted first so this should find
         # the most relevant connection of the given type
         for value in self.values():
+            # XXX why is value sometimes a Result?
             if isinstance(value, Result):
                 value = value.resolved
             if (
@@ -336,6 +356,30 @@ class _ConnectionsMap(dict):
         raise KeyError(key)
 
 
+class TaskLoggerAdapter(logging.LoggerAdapter, LogExtraLevels):
+    def log(self, level, msg, *args, **kwargs):
+        """
+        Delegate a log call to the underlying logger, after adding
+        contextual information from this adapter instance.
+        """
+        # we override this instead of process() because we need to change the log level
+        # note that `self.extra` does NOT replace kwargs['extra']
+        if self.isEnabledFor(level):
+            # task for {resource name}: {operation} (reason: {resource})
+            # e.g. Task for test-node: configure (reason: missing)
+            task: TaskView = self.extra
+            task_id = f"for {task.target.name}: {task.configSpec.operation}"
+            if task.reason:
+                task_id += f" (reason: {task.reason})"
+            if task._rendering:
+                msg = f"Rendering task {task_id} (errors expected): {msg}"
+                if level >= Levels.VERBOSE:
+                    level = Levels.VERBOSE
+            else:
+                msg = f"Running task {task_id}: {msg}"
+            self.logger.log(level, msg, *args, **kwargs)
+
+
 class TaskView:
     """The interface presented to configurators.
 
@@ -349,30 +393,33 @@ class TaskView:
         verbose (int): Verbosity level set for this job (-1 error, 0 normal, 1 verbose, 2 debug)
     """
 
-    def __init__(self,
+    def __init__(
+        self,
         manifest: "Manifest",
         configSpec: ConfigurationSpec,
-        target: NodeInstance,
-        reason: Optional[str]=None,
-        dependencies: Optional[List["Dependency"]]=None
+        target: EntityInstance,
+        reason: Optional[str] = None,
+        dependencies: Optional[List["Operational"]] = None,
     ) -> None:
         # public:
         self.configSpec = configSpec
         self.target = target
         self.reason = reason
-        self.logger = logger
+        self.logger = TaskLoggerAdapter(logger, self)  # type: ignore
         self.cwd = os.path.abspath(self.target.base_dir)
-        self.rendered = None
+        self.rendered: Any = None
         self.dry_run = None
         # private:
-        self._errors: List[UnfurlTaskError] = []  # UnfurlTaskError objects appends themselves to this list
+        self._errors: List[
+            UnfurlTaskError
+        ] = []  # UnfurlTaskError objects appends themselves to this list
         self._inputs: Optional[ResultsMap] = None
         self._environ = None
         self._manifest = manifest
         self.messages: List[object] = []
         self._addedResources: List[NodeInstance] = []
         self._dependenciesChanged = False
-        self.dependencies = dependencies or []
+        self.dependencies: List["Operational"] = dependencies or []
         self._resourceChanges = ResourceChanges()
         self._workFolders: dict = {}
         self._rendering = False
@@ -393,11 +440,16 @@ class TaskView:
         if self._inputs is None:
             assert self._attributeManager  # type: ignore
             assert self.target.root.attributeManager is self._attributeManager  # type: ignore
+            ctx = RefContext(self.target, task=self)
             # deepcopy because ResultsMap might modify interior maps and lists
             inputs = copy.deepcopy(self.configSpec.inputs)
-            relationship = isinstance(self.target, RelationshipInstance)
-            if relationship:
-                target = self.target.target  # type: ignore  # checked with isinstance above
+            relationship = None
+            if isinstance(self.target, RelationshipInstance):
+                relationship = self.target
+                if self.target.target:
+                    target: EntityInstance = self.target.target
+                else:
+                    target = self.target.root
             else:
                 target = self.target
             HOST = (target.parent or target).attributes
@@ -406,11 +458,6 @@ class TaskView:
                 inputs=inputs,
                 task=self.get_settings(),
                 connections=self._get_connections(),
-                TOPOLOGY=dict(
-                    inputs=target.root._attributes["inputs"],
-                    outputs=target.root._attributes["outputs"],
-                ),
-                NODES=TopologyMap(target.root),
                 SELF=self.target.attributes,
                 HOST=HOST,
                 ORCHESTRATOR=ORCHESTRATOR and ORCHESTRATOR.attributes or {},
@@ -418,13 +465,17 @@ class TaskView:
                 and self.operation_host.attributes
                 or {},
             )
+            set_context_vars(vars, target)
             if relationship:
-                vars["SOURCE"] = self.target.source.attributes  # type: ignore
+                if relationship.source:
+                    vars["SOURCE"] = relationship.source.attributes  # type: ignore
                 vars["TARGET"] = target.attributes
             # expose inputs lazily to allow self-referencee
-            ctx = RefContext(self.target, vars, task=self, strict=not self._rendering)
+            ctx.vars = vars
             if self.configSpec.artifact and self.configSpec.artifact.base_dir:
                 ctx.base_dir = self.configSpec.artifact.base_dir
+            elif self.configSpec.base_dir:
+                ctx.base_dir = self.configSpec.base_dir
             self._inputs = ResultsMap(inputs, ctx)
         return self._inputs
 
@@ -435,8 +486,14 @@ class TaskView:
         """
         return self.inputs.context.vars
 
+    @property
+    def connections(self) -> _ConnectionsMap:
+        return self.inputs.context.vars["connections"]
+
     @staticmethod
-    def _get_connection(source: NodeInstance, target: Optional[NodeInstance], seen: dict) -> None:
+    def _get_connection(
+        source: NodeInstance, target: Optional[NodeInstance], seen: dict
+    ) -> None:
         """
         Find the requirements on source that match the target
         If source is root, requirements will be the connections that are the default_for the target.
@@ -457,6 +514,8 @@ class TaskView:
         """
         seen: Dict[int, Any] = {}
         for parent in self.target.ancestors:
+            if not isinstance(parent, NodeInstance):
+                continue
             if parent is self.target.root:
                 break
             if self.operation_host:
@@ -481,9 +540,8 @@ class TaskView:
         respectively and the operation's implementation will probably need both those set when it executes.
         """
         env = {}
-        t = lambda datatype: datatype.type == "unfurl.datatypes.EnvVar"
-        for rel in self._get_connections().by_type():  # only one per connection type
-            env.update(rel.merge_props(t, True))
+        for rel in self.connections.by_type():  # only one per connection type
+            env.update(rel.merge_props(find_env_vars, True))
 
         return env
 
@@ -525,7 +583,11 @@ class TaskView:
 
         # add the variables required by TOSCA 1.3 spec
         targets: List[str] = []
-        if isinstance(self.target, RelationshipInstance):
+        if (
+            isinstance(self.target, RelationshipInstance)
+            and self.target.capability
+            and self.target.target
+        ):
             targets = [
                 c.tosca_id
                 for c in self.target.target.get_capabilities(
@@ -536,17 +598,18 @@ class TaskView:
                 dict(
                     TARGETS=",".join(targets),
                     TARGET=self.target.target.tosca_id,
-                    SOURCES=",".join(
-                        [
-                            r.tosca_id
-                            for r in self.target.source.get_requirements(
-                                self.target.requirement.template.name
-                            )
-                        ]
-                    ),
-                    SOURCE=self.target.source.tosca_id,
                 )
             )
+            if self.target.source:
+                SOURCES = ",".join(
+                    [
+                        r.tosca_id
+                        for r in self.target.source.get_requirements(
+                            self.target.requirement.template.name
+                        )
+                    ]
+                )
+                SOURCE = self.target.source.tosca_id
         return env
 
     def get_settings(self) -> dict:
@@ -562,7 +625,21 @@ class TaskView:
             cwd=self.cwd,
         )
 
-    def find_connection(self, target: NodeInstance, relation: str="tosca.relationships.ConnectsTo") -> Any:
+    def find_connection(
+        self, target: NodeInstance, relation: str = "tosca.relationships.ConnectsTo"
+    ) -> Optional[RelationshipInstance]:
+        """
+        Find a relationship that this task can use to connect to the given instance.
+        First look for relationship between the task's target instance and the given instance.
+        If none is found, see if there a default connection of the given type.
+
+        Args:
+            target (NodeInstance): The instance to connect to.
+            relation (str, optional): The relationship type. Defaults to "tosca.relationships.ConnectsTo".
+
+        Returns:
+            RelationshipInstance or None: The connection instance.
+        """
         connection = self.query(
             f"$OPERATION_HOST::.requirements::*[.type={relation}][.target=$target]",
             vars=dict(target=target),
@@ -573,7 +650,10 @@ class TaskView:
             endpoints = target.get_default_relationships(relation)
             if endpoints:
                 connection = endpoints[0]
-        return connection
+        if connection:  # type: ignore
+            assert isinstance(connection, RelationshipInstance)
+            return connection
+        return None
 
     def sensitive(self, value: object) -> Union[sensitive, object]:
         """Mark the given value as sensitive. Sensitive values will be encrypted or redacted when outputed.
@@ -600,14 +680,37 @@ class TaskView:
     #     outputs to share operation outputs so far
     #     """
 
+    def _set_outputs(self, outputs):
+        for key, value in outputs.items():
+            mapping = self.configSpec.outputs.get(key)
+            if mapping:
+                invalid = False
+                if isinstance(mapping, list):
+                    # XXX support more TOSCA mapping forms, e.g. to capabilities
+                    if len(mapping) == 2 and mapping[0] == "SELF":
+                        mapping = mapping[1]
+                    else:
+                        invalid = True
+                if not isinstance(mapping, str):
+                    invalid = True
+                if invalid:
+                    UnfurlTaskError(
+                        self,
+                        f"invalid or unsupported mapping for output '{key}': {mapping}",
+                    )
+                else:
+                    self.target.attributes[mapping] = value
+
     def done(
         self,
-        success: Optional[bool]=None,
-        modified: Optional[bool]=None,
-        status: Optional[Status]=None,
-        result: Optional[Union[dict, str]]=None,
-        outputs: Optional[dict]=None,  # so the docstring says dict, but ConfiguratorResult
-        captureException: Optional[object]=None,
+        success: Optional[bool] = None,
+        modified: Optional[bool] = None,
+        status: Optional[Status] = None,
+        result: Optional[Union[dict, str]] = None,
+        outputs: Optional[
+            dict
+        ] = None,  # so the docstring says dict, but ConfiguratorResult
+        captureException: Optional[object] = None,
     ) -> ConfiguratorResult:
         """:py:meth:`unfurl.configurator.Configurator.run` should call this method and yield its return value before terminating.
 
@@ -639,14 +742,15 @@ class TaskView:
 
         # Typechecking
         class kwTypeBase(TypedDict):
-            result: Optional[dict]
+            result: Optional[Union[dict, str]]
             outputs: Optional[dict]
 
         class kwType(kwTypeBase, total=False):
             exception: UnfurlTaskError
 
         kw = cast(kwType, kw)
-
+        if outputs:
+            self._set_outputs(outputs)
         return ConfiguratorResult(success, modified, status, **kw)
 
     # updates can be marked as dependencies (changes to dependencies changed) or required (error if changed)
@@ -656,15 +760,15 @@ class TaskView:
     def query(
         self,
         query: str,
-        dependency: bool=False,
-        name: Optional[str]=None,
-        required: bool=False,
-        wantList: bool=False,
-        resolveExternal: bool=True,
-        strict: bool=True,
-        vars: Optional[dict]=None,
-        throw: bool=False,
-    ) -> Optional[Ref]:
+        dependency: bool = False,
+        name: Optional[str] = None,
+        required: bool = False,
+        wantList: bool = False,
+        resolveExternal: bool = True,
+        strict: bool = True,
+        vars: Optional[dict] = None,
+        throw: bool = False,
+    ) -> Union[ResultsList, Result, List[Result], None]:
         # XXX pass resolveExternal to context?
         try:
             result = Ref(query, vars=vars).resolve(
@@ -687,12 +791,12 @@ class TaskView:
     def add_dependency(
         self,
         expr: Union[str, Mapping],
-        expected: Optional[Union[list, ResultsList, Result]]=None,
-        schema: Optional[Mapping]=None,
-        name: Optional[str]=None,
-        required: bool=True,
-        wantList: bool=False,
-        target: Optional[NodeInstance]=None,
+        expected: Optional[Union[list, ResultsList, Result]] = None,
+        schema: Optional[Mapping] = None,
+        name: Optional[str] = None,
+        required: bool = True,
+        wantList: bool = False,
+        target: Optional[NodeInstance] = None,
     ) -> "Dependency":
         getter = getattr(expr, "as_ref", None)
         if getter:
@@ -703,6 +807,7 @@ class TaskView:
             expr, expected, schema, name, required, wantList, target
         )
         for i, dep in enumerate(self.dependencies):
+            assert isinstance(dep, Dependency)
             if dep.expr == expr or dep.name == name:
                 self.dependencies[i] = dependency
                 break
@@ -713,6 +818,7 @@ class TaskView:
 
     def remove_dependency(self, name: str) -> Optional["Dependency"]:
         for i, dep in enumerate(self.dependencies):
+            assert isinstance(dep, Dependency)
             if dep.name == name:
                 self.dependencies.pop(i)
                 self._dependenciesChanged = True
@@ -726,11 +832,11 @@ class TaskView:
 
     def create_sub_task(
         self,
-        operation: Optional[str]=None,
-        resource: Optional[NodeInstance]=None,
-        inputs: Optional[dict]=None,
-        persist: bool=False,
-        required: Optional[bool]=None
+        operation: Optional[str] = None,
+        resource: Optional[EntityInstance] = None,
+        inputs: Optional[dict] = None,
+        persist: bool = False,
+        required: Optional[bool] = None,
     ) -> Optional[TaskRequest]:
         """Create a subtask that will be executed if yielded by :py:meth:`unfurl.configurator.Configurator.run`
 
@@ -777,7 +883,9 @@ class TaskView:
         # otherwise operation should be a ConfigurationSpec
         return TaskRequest(operation, resource, "subtask", persist, required)
 
-    def _update_instance(self, existingResource: NodeInstance, resourceSpec: Mapping) -> bool:
+    def _update_instance(
+        self, existingResource: EntityInstance, resourceSpec: Mapping
+    ) -> bool:
         from .manifest import Manifest
 
         updated = False
@@ -793,22 +901,28 @@ class TaskView:
                 existingResource.state = operational.state  # type: ignore
             updated = True
 
+        protected = resourceSpec.get("protected")
+        if protected is not None:
+            existingResource.protected = bool(protected)
+            updated = True
+
         attributes = resourceSpec.get("attributes")
         if attributes:
+            ctx = self.inputs.context.copy(existingResource)
             for key, value in attributes.items():
-                existingResource.attributes[key] = map_value(value, existingResource)
+                existingResource.attributes[key] = map_value(value, ctx)
                 self.logger.debug(
                     "setting attribute %s on %s with %s",
                     key,
                     existingResource.name,
-                    truncate_str(value)
+                    truncate_str(value),
                 )
             updated = True
         return updated
 
-    def _parse_instances_tpl(self,
-        instances: Union[str, Mapping, MutableSequence]
-    ) -> Union[Tuple[None, UnfurlTaskError], Tuple[MutableSequence, None]]:
+    def _parse_instances_tpl(
+        self, instances: Union[str, Mapping, MutableSequence]
+    ) -> Tuple[Optional[MutableSequence[Mapping]], Optional[UnfurlTaskError]]:
         if isinstance(instances, six.string_types):
             try:
                 instances = yaml.load(instances)
@@ -828,12 +942,16 @@ class TaskView:
 
     # # XXX how can we explicitly associate relations with target resources etc.?
     # # through capability attributes and dependencies/relationship attributes
-    def update_instances(self, instances: Union[str, List]) -> Union[Tuple[JobRequest, List], Tuple[None, List]]:
+    def update_instances(
+        self, instances: Union[str, List]
+    ) -> Tuple[Optional[JobRequest], List[UnfurlTaskError]]:
         """Notify Unfurl of new or changes to instances made while the configurator was running.
 
         Operational status indicates if the instance currently exists or not.
         This will queue a new child job if needed.
 
+        To run the job based on the supplied spec immediately, yield the returned JobRequest.
+
         .. code-block:: YAML
 
           - name:     aNewInstance
@@ -851,15 +969,12 @@ class TaskView:
         Args:
           instances (list or str): Either a list or string that is parsed as YAML.
 
-        Returns:
-          :class:`JobRequest`: To run the job based on the supplied spec
-              immediately, yield the returned JobRequest.
         """
         instances, err = self._parse_instances_tpl(instances)  # type: ignore
         if err:
             return None, [err]
 
-        errors = []
+        errors: List[UnfurlTaskError] = []
         newResources = []
         newResourceSpecs = []
         for resourceSpec in instances:
@@ -881,16 +996,25 @@ class TaskView:
                 if existingResource:
                     updated = self._update_instance(existingResource, resourceSpec)
                     if updated:
-                        self.logger.info("updating instances %s", existingResource.name)
+                        self.logger.info(
+                            "updating dynamic instance %s", existingResource.name
+                        )
+                    else:
+                        self.logger.debug(
+                            "no change to dynamic instance %s", existingResource.name
+                        )
                 else:
                     newResource = create_instance_from_spec(
                         self._manifest, self.target, rname, resourceSpec
                     )
+                    if "priority" not in resourceSpec:
+                        # set priority so this resource isn't ignored even if no one if referencing it
+                        newResource.priority = Priority.required
 
                 # XXX
                 # if resource.required or resourceSpec.get("dependent"):
                 #    self.add_dependency(resource, required=resource.required)
-            except:
+            except Exception:
                 errors.append(
                     UnfurlAddingResourceError(self, originalResourceSpec, rname)
                 )
@@ -910,7 +1034,9 @@ class TaskView:
             return jobRequest, errors
         return None, errors
 
-    def set_work_folder(self, location: str="operation", preserve: Optional[bool]=None) -> WorkFolder:
+    def set_work_folder(
+        self, location: str = "operation", preserve: Optional[bool] = None
+    ) -> WorkFolder:
         if location in self._workFolders:
             return self._workFolders[location]
         if preserve is None:
@@ -923,12 +1049,16 @@ class TaskView:
         #     self, location, preserve
         # )
 
-    def get_work_folder(self, location: Optional[str]=None) -> WorkFolder:
+    def get_work_folder(self, location: Optional[str] = None) -> WorkFolder:
         # return self.job.getFolder(self, location)
         if location is None:
+            if not self._workFolders:
+                raise UnfurlTaskError(self, f"No task folder was set.")
             # XXX error if there is more than one?
             return next(iter(self._workFolders.values()))
         else:
+            if location not in self._workFolders:
+                raise UnfurlTaskError(self, f'missing "{location}" task folder')
             return self._workFolders[location]
 
     def discard_work_folders(self) -> None:
@@ -961,13 +1091,13 @@ class Dependency(Operational):
 
     def __init__(
         self,
-        expr: Union[str, dict],
-        expected: Optional[Union[ResultsList, list, Result, None]]=None,
-        schema: Optional[Mapping]=None,
-        name: Optional[str]=None,
-        required: bool=False,
-        wantList: bool=False,
-        target: Optional[NodeInstance]=None,
+        expr: Union[str, Mapping],
+        expected: Optional[Union[ResultsList, list, Result, None]] = None,
+        schema: Optional[Mapping] = None,
+        name: Optional[str] = None,
+        required: bool = False,
+        wantList: bool = False,
+        target: Optional[EntityInstance] = None,
     ) -> None:
         """
         if schema is not None, validate the result using schema
@@ -985,11 +1115,15 @@ class Dependency(Operational):
         self.target = target
 
     @property
-    def local_status(self) -> Status:
-        if self.target and self.target is not self.target.root and not self.target.is_computed():
+    def local_status(self) -> Optional[Status]:
+        if (
+            self.target
+            and self.target is not self.target.root
+            and not self.target.is_computed()
+        ):
             # (only care about local status of instances with live attribute, not their full operational status)
             # (reduces circular dependencies)
-            return self.target.local_status  # type: ignore  # mypy has trouble with @property calls
+            return self.target.local_status
         else:  # the root has inputs which don't have operational status
             return Status.ok
 
@@ -1004,11 +1138,28 @@ class Dependency(Operational):
                 config.target, dict(val=self.expected, changeId=changeId)
             )
             result = Ref(self.expr).resolve(context, wantList=self.wantList)
+            assert isinstance(context._lastResource, EntityInstance)
             self.target = context._lastResource
             self.expected = result
 
+    def validate(self):
+        if not self.schema or not self.target:
+            return True
+        value = Ref(self.expr).resolve(RefContext(self.target))
+        if isinstance(self.schema, dict):
+            return not validate_schema(value, self.schema)
+        else:  # ProperyDef
+            try:
+                self.schema._validate(value)
+            except Exception:
+                return False
+        return True
+
     @staticmethod
-    def has_value_changed(value: Union[Results, Mapping, MutableSequence, tuple, ChangeAware, Any], changeset: ChangeRecord) -> bool:
+    def has_value_changed(
+        value: Union[Results, Mapping, MutableSequence, tuple, ChangeAware, Any],
+        changeset: ChangeRecord,
+    ) -> bool:
         if isinstance(value, Results):
             return Dependency.has_value_changed(value._attributes, changeset)
         elif isinstance(value, Mapping):
@@ -1023,9 +1174,10 @@ class Dependency(Operational):
             return False
         return False  # assuming this is what is meant by the function
 
-    def has_changed(self, config: "ConfigTask") -> bool:
+    def has_changed(self, config: "ChangeRecord") -> bool:
         changeId = config.changeId
-        context = RefContext(config.target, dict(val=self.expected, changeId=changeId))
+        # Manifest.load_config_change sets config.target
+        context = RefContext(config.target, dict(val=self.expected, changeId=changeId))  # type: ignore
         result = Ref(self.expr).resolve_one(context)  # resolve(context, self.wantList)
 
         if self.schema:
