@@ -6,13 +6,14 @@ from ..util import register_short_names
 from ..support import Status
 from ..planrequests import set_default_command
 import importlib
+from typing import Tuple
 
 # need to define these now because these configurators are lazily imported
 # and so won't register themselves through AutoRegisterClass
 register_short_names(
     {
         name: f"unfurl.configurators.{name.lower()}.{name}Configurator"
-        for name in "Ansible Shell Supervisor Terraform DNS".split()
+        for name in "Ansible Shell Supervisor Terraform DNS Kompose".split()
     }
 )
 
@@ -30,13 +31,13 @@ class PythonPackageCheckConfigurator(Configurator):
             status = Status.ok
         except ImportError:
             status = Status.absent
-        except:
+        except Exception:
             status = Status.error
         yield task.done(True, status=status)
 
 
 class TemplateConfigurator(Configurator):
-    exclude_from_digest = ("resultTemplate", "done")
+    exclude_from_digest: Tuple[str, ...] = ("resultTemplate", "done")
 
     def process_result_template(self, task, result):
         """
@@ -70,8 +71,7 @@ class TemplateConfigurator(Configurator):
                     jr, errors = task.update_instances(results)
             if errors:
                 task.logger.warning(
-                    "error processing resultTemplate for %s: %s",
-                    task.target.name,
+                    "error processing resultTemplate: %s",
                     errors[0],
                 )
         return errors
@@ -102,7 +102,7 @@ class TemplateConfigurator(Configurator):
         done = task.inputs.get_copy("done", {})
         if "result" not in done:
             if not isinstance(runResult, dict):
-                done["result"] = {"run": runResult}
+                done["result"] = {"run": runResult, "outputs": done.get("outputs")}
             else:
                 done["result"] = runResult
         if self.process_result_template(task, done.get("result") or {}):
@@ -132,8 +132,17 @@ class DelegateConfigurator(Configurator):
             return False
 
     def run(self, task):
-        subtaskRequest = task.rendered
-        assert subtaskRequest
-        # note: this will call canRun() and if needed canDryRun() on subtask but not shouldRun()
-        subtask = yield subtaskRequest
-        yield subtask.result
+        if "when" in task.inputs and not task.inputs["when"]:
+            # check this here instead of should_run() so that we always create and run the task
+            # (really just a hack so we save the digest in the job log for future reconfigure operations)
+            task.logger.debug(
+                "skipping subtask: 'when' input evaluated to false: %s",
+                task.configSpec.inputs["when"],
+            )
+            yield task.done(True, modified=False)
+        else:
+            subtaskRequest = task.rendered
+            assert subtaskRequest
+            # note: this will call canRun() and if needed canDryRun() on subtask but not shouldRun()
+            subtask = yield subtaskRequest
+            yield subtask.result
