@@ -1,7 +1,7 @@
 # Copyright (c) 2020 Adam Souzis
 # SPDX-License-Identifier: MIT
 from ..util import save_to_file, UnfurlTaskError, wrap_var, which
-from .shell import ShellConfigurator, clean_output
+from .shell import ShellConfigurator, clean_output, make_regex_filter
 from ..support import Status
 from ..result import Result
 from ..projectpaths import get_path, FilePath, Folders
@@ -17,7 +17,7 @@ def _get_env(env, verbose, dataDir):
     env["TF_INPUT"] = "0"
     # see https://www.terraform.io/plugin/log/managing
     # env["TF_LOG"] = "ERROR WARN INFO DEBUG TRACE".split()[verbose + 1]
-    if verbose > 0:
+    if verbose >= 0:
         # providers can be very verbose, don't set them to debug
         env["TF_LOG_CORE"] = "DEBUG"
 
@@ -27,6 +27,19 @@ def _get_env(env, verbose, dataDir):
     return env
 
 
+def get_echo_args(verbosity):
+    if verbosity == -1:  # quiet mode
+        return dict(echo=False) # no stdout or stderr
+    else:
+        logregex = re.compile(r"\[(TRACE|DEBUG|INFO|WARN|ERROR)\]")
+        if verbosity == 0:  # default
+            levels = "INFO|WARN|ERROR"
+        else:  # verbose == 1
+            levels = "TRACE|DEBUG|INFO|WARN|ERROR"
+        stderr_filter = make_regex_filter(logregex, levels.split('|'))
+    return dict(echo=True, stderr_filter=stderr_filter)
+
+
 def mark_block(schema, items, task, sensitive_names):
     blockTypes = schema.get("block_types", {})
     attributes = schema.get("attributes", {})
@@ -94,12 +107,14 @@ module "main" {
 
 
 def generate_main(relpath, tfvars, outputs):
+    # XXX until we can check if an output is sensitive
+    # we always need to set them all to sensitive to avoid a terraform error
+    # when the referenced output is sensitive
+    sensitive = True
     # if tfvars are hcl:
-    # XXX check if outputs are sensitive
     if isinstance(tfvars, six.string_types):
         output = ""
         for name in outputs:
-            sensitive = False
             if sensitive:
                 sensitive_str = "sensitive=true\n"
             else:
@@ -114,8 +129,9 @@ def generate_main(relpath, tfvars, outputs):
         if outputs:
             output = root["output"] = {}
             for name in outputs:
-                sensitive = False
-                output[name] = dict(value=f"${{module.main.{name}}}", sensitive=sensitive)
+                output[name] = dict(
+                    value=f"${{module.main.{name}}}", sensitive=sensitive
+                )
         return "main.tmp.tf.json", root
 
 
@@ -158,7 +174,9 @@ class TerraformConfigurator(ShellConfigurator):
         ]
         # folder is "tasks" WorkFolder
         cwd = folder.cwd
-        lock_file = task.set_work_folder(Folders.artifacts).permanent_path(".terraform.lock.hcl", False)
+        lock_file = task.set_work_folder(Folders.artifacts).permanent_path(
+            ".terraform.lock.hcl", False
+        )
         if os.path.exists(lock_file):
             folder.copy_from(lock_file)
 
@@ -204,15 +222,16 @@ class TerraformConfigurator(ShellConfigurator):
         # generated tf.json get written to as main.unfurl.tmp.tf.json
         write_vars = True
         contents = None
-        if "main" not in task.inputs:
-            main = get_path(task.inputs.context, "", "spec.home")
+        main = task.inputs.get_copy("main")
+        if not main:
+            main = get_path(task.inputs.context, task.target.name, "src")
             if not os.path.exists(main):
                 raise UnfurlTaskError(
                     task,
                     f'Input parameter "main" not specifed and default terraform module directory does not exist at "{main}"',
                 )
-        else:
-            main = task.inputs.get_copy("main")
+        if task._errors:
+            main = None  # assume render failed
         if isinstance(main, six.string_types):
             if "\n" in main:
                 # assume its HCL and not a path
@@ -264,6 +283,8 @@ class TerraformConfigurator(ShellConfigurator):
         # XXX .tfvars can be sensitive
         # we need to output the plan and convert it to json to see which variables are marked sensitive
         tfvars = task.inputs.get_copy("tfvars")
+        if task._errors:
+            return None  # assume render failed
         if tfvars:
             if isinstance(tfvars, six.string_types):
                 # assume the contents of a tfvars file
@@ -274,7 +295,9 @@ class TerraformConfigurator(ShellConfigurator):
         return None
 
     def _get_workfolder_name(self, task):
-        return task.inputs.get("stateLocation") or Folders.artifacts # XXX global option for secrets
+        return (
+            task.inputs.get("stateLocation") or Folders.secrets
+        )  # XXX global option for secrets
 
     def _prepare_state(self, task, cwd):
         # the terraform state file is associate with the current instance
@@ -283,19 +306,17 @@ class TerraformConfigurator(ShellConfigurator):
         folderName = self._get_workfolder_name(task)
         if folderName == "remote":  # don't use local state file
             return ""
-        yamlPath = task.set_work_folder(folderName).permanent_path("terraform.tfstate.yaml", False)
+        yamlPath = task.set_work_folder(folderName).permanent_path(
+            "terraform.tfstate.yaml", False
+        )
         if os.path.exists(yamlPath):
-            task.logger.debug(
-                "Found existing terraform.tfstate file at %s", yamlPath
-            )
+            task.logger.debug("Found existing terraform.tfstate file at %s", yamlPath)
             # if exists in home, load and write out state file as json
             with open(yamlPath, "r") as f:
                 state = task._manifest.yaml.load(f.read())
             cwd.write_file(state, "terraform.tfstate")
         else:
-            task.logger.debug(
-                "Couldn't find terraform.tfstate file at %s", yamlPath
-            )
+            task.logger.debug("Couldn't find terraform.tfstate file at %s", yamlPath)
         return "terraform.tfstate"
 
     def _get_plan_path(self, task, cwd):
@@ -357,7 +378,6 @@ class TerraformConfigurator(ShellConfigurator):
     def run(self, task):
         cwd = task.get_work_folder(Folders.tasks)
         cmd, terraform, statePath = task.rendered
-        echo = task.verbose > -1
         current_path = cwd.cwd
         dataDir = os.getenv("TF_DATA_DIR", os.path.join(current_path, ".terraform"))
         env = _get_env(task.get_environment(False), task.verbose, dataDir)
@@ -376,12 +396,13 @@ class TerraformConfigurator(ShellConfigurator):
         else:
             providerSchema = {}
 
+        echo_args = get_echo_args(task.verbose)
         result = self.run_process(
             cmd,
             timeout=task.configSpec.timeout,
             env=env,
             cwd=cwd.cwd,
-            echo=echo,
+            **echo_args
         )
         if result.returncode and _needs_init(clean_output(result.stderr)):
             # modules or plugins out of date, re-run terraform init
@@ -394,7 +415,7 @@ class TerraformConfigurator(ShellConfigurator):
                     timeout=task.configSpec.timeout,
                     env=env,
                     cwd=cwd.cwd,
-                    echo=echo,
+                    **echo_args
                 )
             else:
                 raise UnfurlTaskError(
@@ -405,14 +426,30 @@ class TerraformConfigurator(ShellConfigurator):
         # process the result
         status = None
         success = self._handle_result(task, result, cwd.cwd, (0, 2))
+        # plan -detailed-exitcode: 2 - Succeeded, but there is a diff
+        needs_changes = False
         if result.returncode == 2:
-            # plan -detailed-exitcode: 2 - Succeeded, but there is a diff
-            success = True
-            if task.configSpec.operation == "check":
-                status = Status.absent
-            else:
+            success = True  # command succeeded despite non-zero return code
+            # we check for "Plan:" in the output because sensitive outputs will always be marked as changed
+            # so to know if changes are really needed we also look for a message like: Plan: 1 to add, 0 to change, 0 to destroy.
+            needs_changes = "Plan:" in result.stdout
+            if task.configSpec.operation != "check":
+                # changes were applied so set to OK
                 status = Status.ok
 
+        if success and task.configSpec.operation == "check":
+            if needs_changes:  # treat as missing
+                status = Status.absent
+            elif task.target.status in [Status.pending, Status.unknown]:
+                # no changes needed so set to known state
+                status = Status.ok
+
+        modified = (
+            "Modifying..." in result.stdout
+            or "Creating..." in result.stdout
+            or "Destroying..." in result.stdout
+        )
+
         if not task.dry_run and task.configSpec.operation != "check":
             outputs = {}
             current_path = cwd.cwd
@@ -435,6 +472,8 @@ class TerraformConfigurator(ShellConfigurator):
                 }
                 state.update(result.__dict__)
                 state["outputs"] = outputs  # replace outputs
+                state["success"] = success
+                state["modified"] = modified
                 errors = self.process_result_template(task, state)
                 if success:
                     success = not errors
@@ -443,19 +482,15 @@ class TerraformConfigurator(ShellConfigurator):
         else:
             outputs = None
 
-        modified = (
-            "Modifying..." in result.stdout
-            or "Creating..." in result.stdout
-            or "Destroying..." in result.stdout
-        )
         yield self.done(
-                task,
-                success=success,
-                modified=modified,
-                status=status,
-                result=result.__dict__,
-                outputs=outputs,
-            )
+            task,
+            success=success,
+            modified=modified,
+            status=status,
+            result=result.__dict__,
+            outputs=outputs,
+        )
+
 
 # XXX implement discover:
 # terraform import -allow-missing-config {type.name}
