@@ -15,7 +15,18 @@ eval_ref() given expression (string or dictionary) return list of Result
 Expr.resolve() given expression string, return list of Result
 Results._map_value same as map_value but with lazily evaluation
 """
-from typing import Any, Dict, List, Optional, Union, cast, TYPE_CHECKING
+from typing import (
+    Any,
+    Dict,
+    List,
+    Tuple,
+    Optional,
+    Union,
+    Iterable,
+    cast,
+    TYPE_CHECKING,
+)
+from typing import Mapping as MappingType
 import six
 import re
 import operator
@@ -42,8 +53,8 @@ logger = cast(UnfurlLogger, logging.getLogger("unfurl.eval"))
 def map_value(
     value: Any,
     resourceOrCxt: Union["RefContext", "NodeInstance"],
-    applyTemplates: bool=True
-    ) -> Optional[Union[ResultsList, Result, List[Result]]]:
+    applyTemplates: bool = True,
+) -> Optional[Union[ResultsList, Result, List[Result]]]:
     if not isinstance(resourceOrCxt, RefContext):
         resourceOrCxt = RefContext(resourceOrCxt)
     return _map_value(value, resourceOrCxt, False, applyTemplates)
@@ -122,17 +133,18 @@ class RefContext:
 
     def __init__(
         self,
-        currentResource: "NodeInstance",
-        vars: Optional[dict]=None,
-        wantList: bool=False,
-        resolveExternal: bool=False,
-        trace: Optional[int]=None,
-        strict: bool=_defaultStrictness,
-        task: "TaskView"=None,
+        currentResource: "ResourceRef",
+        vars: Optional[dict] = None,
+        wantList: Optional[Union[bool, str]] = False,
+        resolveExternal: bool = False,
+        trace: Optional[int] = None,
+        strict: bool = _defaultStrictness,
+        task: "TaskView" = None,
     ) -> None:
         self.vars = vars or {}
         # the original context:
         self.currentResource = currentResource
+        assert isinstance(currentResource, ResourceRef)
         # the last resource encountered while evaluating:
         self._lastResource = currentResource
         # current segment is the final segment:
@@ -140,33 +152,42 @@ class RefContext:
         self.wantList = wantList
         self.resolveExternal = resolveExternal
         self._trace = self.DefaultTraceLevel if trace is None else trace
-        self.strict = strict
+        self._strict = strict
         self.base_dir = currentResource.base_dir
         self.templar = currentResource.templar
         self.referenced = _Tracker()
         self.task = task
+        self.kw: MappingType[str, Any] = {}
+
+    @property
+    def strict(self):
+        if self.task:
+            return not self.task._rendering
+        else:
+            return self._strict
 
     def copy(
         self,
-        resource: Optional["NodeInstance"]=None,
-        vars: dict=None,
-        wantList: bool=None,
-        trace:  int=0,
-        strict: bool=None
-        ) -> "RefContext":
+        resource: Optional["ResourceRef"] = None,
+        vars: dict = None,
+        wantList: Optional[Union[bool, str]] = None,
+        trace: int = 0,
+        strict: bool = None,
+    ) -> "RefContext":
+        if not isinstance(resource or self.currentResource, ResourceRef) and isinstance(
+            self._lastResource, ResourceRef
+        ):
+            resource = self._lastResource
         copy = RefContext(
             resource or self.currentResource,
             self.vars,
             self.wantList,
             self.resolveExternal,
             max(self._trace, trace),
-            self.strict if strict is None else strict,
+            self._strict if strict is None else strict,
         )
-        if not isinstance(copy.currentResource, ResourceRef) and isinstance(
-            self._lastResource, ResourceRef
-        ):
-            copy._lastResource = self._lastResource
         if vars:
+            copy.vars = copy.vars.copy()
             copy.vars.update(vars)
         if wantList is not None:
             copy.wantList = wantList
@@ -179,7 +200,7 @@ class RefContext:
     def trace(self, *msg: Any) -> None:
         if self._trace:
             log = logger.info if self._trace == 2 else logger.trace
-            log(f"{' '.join(str(a) for a in msg)} (ctx: {self._lastResource})")
+            log(f"{' '.join(str(a) for a in msg)} (ctx: {self._lastResource})")  # type: ignore
 
     def add_external_reference(self, external: ExternalValue) -> Result:
         result = Result(external)
@@ -192,7 +213,7 @@ class RefContext:
     def add_result_reference(self, ref: str, result: Result) -> None:
         self.referenced.add_result_reference(ref, result)
 
-    def resolve_var(self, key: str) -> Optional[Union[Results, ResultsList, List[Result], ResultsMap]]:
+    def resolve_var(self, key: str) -> Any:  # Result asserts resolved is not a Result
         return self._resolve_var(key[1:]).resolved
 
     def _resolve_var(self, key: str) -> Result:
@@ -207,7 +228,9 @@ class RefContext:
             self.vars[key] = val
             return val
 
-    def resolve_reference(self, key: str) -> Union[Results, ResultsList, List[Result], ResultsMap]:
+    def resolve_reference(
+        self, key: str
+    ) -> Union[Results, ResultsList, List[Result], ResultsMap]:
         val = self._resolve_var(key)
         self.add_result_reference(key, val)
         assert not isinstance(val.resolved, Result)
@@ -215,9 +238,9 @@ class RefContext:
 
     def query(
         self,
-        expr: Mapping,
-        vars: Optional[dict]=None,
-        wantList: Union[bool, str]=False
+        expr: Union[str, Mapping],
+        vars: Optional[dict] = None,
+        wantList: Union[bool, str] = False,
     ) -> Optional[Union[ResultsList, Result, List[Result]]]:
         return Ref(expr, vars).resolve(self, wantList)
 
@@ -284,17 +307,21 @@ class Expr:
 class Ref:
     """A Ref objects describes a path to metadata associated with a resource."""
 
-    def __init__(self, exp: Union[str, Mapping], vars: dict=None, trace: Optional[int]=None) -> None:
+    def __init__(
+        self, exp: Union[str, Mapping], vars: dict = None, trace: Optional[int] = None
+    ) -> None:
         self.vars = {"true": True, "false": False, "null": None}
 
         self.foreach = None
+        self.select = None
         trace = RefContext.DefaultTraceLevel if trace is None else trace
         self.trace = trace
         if isinstance(exp, Mapping):
             keys = list(exp)
             if keys and keys[0] not in _FuncsTop:
                 self.vars.update(exp.get("vars", {}))
-                self.foreach = exp.get("foreach", exp.get("select"))
+                self.foreach = exp.get("foreach")
+                self.select = exp.get("select")
                 self.trace = exp.get("trace", trace)
                 exp = exp.get("eval", exp.get("ref", exp))
 
@@ -305,8 +332,8 @@ class Ref:
     def resolve(
         self,
         ctx: RefContext,
-        wantList: Union[bool, str]=True,
-        strict: bool=_defaultStrictness
+        wantList: Union[bool, str] = True,
+        strict: bool = _defaultStrictness,
     ) -> Optional[Union[ResultsList, Result, List[Result]]]:
         """
         If wantList=True (default) returns a ResultList of matches
@@ -321,20 +348,28 @@ class Ref:
         if base_dir:
             ctx.base_dir = base_dir
         # $start is set in eval_ref but the user use that to override currentResource
-        if 'start' in self.vars:
-            ctx.currentResource = ctx.resolve_var('$start')
+        if "start" in self.vars:
+            ctx.currentResource = ctx.resolve_var("$start")
         ctx.trace(
             f"Ref.resolve(wantList={wantList}) start strict {ctx.strict}",
             self.source,
         )
-        results = eval_ref(self.source, ctx, True)
-        ctx.trace(f"Ref.resolve(wantList={wantList}) evalRef", self.source, results)
-        if results and self.foreach:
-            results = for_each(self.foreach, results, ctx)
-        assert not isinstance(results, ResultsList), results
-        results = ResultsList(results, ctx)
+        ref_results = eval_ref(self.source, ctx, True)
+        assert isinstance(ref_results, list)
+        ctx.trace(f"Ref.resolve(wantList={wantList}) evalRef", self.source, ref_results)
+        select = self.foreach or self.select
+        if ref_results and select:
+            results = for_each(select, ref_results, ctx)
+        else:
+            results = ResultsList(ref_results, ctx)
         ctx.add_ref_reference(self, results)
         ctx.trace(f"Ref.resolve(wantList={wantList}) results", self.source, results)
+        if self.foreach:
+            # foreach always returns a list
+            if wantList == "result":
+                return Result(results)
+            else:
+                return results
         if wantList and not wantList == "result":
             return results
         else:
@@ -351,7 +386,9 @@ class Ref:
                 else:
                     return list(results)
 
-    def resolve_one(self, ctx: RefContext, strict: bool=_defaultStrictness) -> Optional[Union[ResultsList, Result, List[Result]]]:
+    def resolve_one(
+        self, ctx: RefContext, strict: bool = _defaultStrictness
+    ) -> Optional[Union[ResultsList, Result, List[Result]]]:
         """
         If no match return None
         If more than one match return a list of matches
@@ -440,11 +477,23 @@ def validate_schema_func(arg, ctx):
     assert_form(args, MutableSequence, len(args) == 2)
     return validate_schema(args[0], args[1])
 
+if sys.version_info >= (3, 9):
+    MutableSequence_Result = MutableSequence[Result]
+else:
+    MutableSequence_Result = MutableSequence
 
-def _for_each(foreach, results, ctx):
-    if isinstance(foreach, six.string_types):
-        keyExp = None
-        valExp = foreach
+ResultOrResultList = Union[Result, MutableSequence_Result]
+PairResultOrResultList = Tuple[Any, ResultOrResultList]
+MaybePairOrResultOrResultList = Union[PairResultOrResultList, ResultOrResultList]
+
+def _for_each(
+    foreach: Union[MappingType[str, Any], str],
+    results: Iterable[Tuple[Any, Any]],
+    ctx: RefContext,
+) -> List[ResultOrResultList]:
+    if isinstance(foreach, str):
+        keyExp: Union[Mapping, str, None] = None
+        valExp: Union[Mapping, str, None] = foreach
     else:
         keyExp = foreach.get("key")
         valExp = foreach.get("value")
@@ -457,8 +506,10 @@ def _for_each(foreach, results, ctx):
     Break = object()
     Continue = object()
 
-    def make_items():
+    def make_items() -> Iterable[MaybePairOrResultOrResultList]:
         for i, (k, v) in enumerate(results):
+            # XXX stop setting any kind of value to currentResource
+            # assert isinstance(v, ResourceRef)
             ictx.currentResource = v
             ictx.vars["collection"] = results
             ictx.vars["index"] = i
@@ -472,35 +523,42 @@ def _for_each(foreach, results, ctx):
                     break
                 elif key is Continue:
                     continue
-            valResults = eval_ref(valExp, ictx)
-            if not valResults:
+            assert valExp
+            evalResults = eval_ref(valExp, ictx)
+            if not evalResults:
                 continue
-            if len(valResults) == 1:
-                val = valResults[0].resolved
+            if len(evalResults) == 1:
+                val = evalResults[0].resolved
                 if val is Break:
                     break
                 elif val is Continue:
                     continue
-                valResults = valResults[0]
+                valResult: ResultOrResultList = evalResults[0]
+            else:
+                valResult = evalResults
 
             if keyExp:
-                yield (key, valResults)
+                yield (key, valResult)
             else:
-                yield valResults
+                yield valResult
 
     if keyExp:
-        # use CommentedMap to preserve order
-        return [CommentedMap(make_items())]
+        return [Result(CommentedMap( cast(Iterable[PairResultOrResultList], make_items()) ))]
     else:
-        return list(make_items())
+        return list(cast(Iterable[ResultOrResultList], make_items()))
 
 
-def for_each(foreach, results, ctx):
-    # results will be list of Result
-    return _for_each(foreach, enumerate(r.external or r.resolved for r in results), ctx)
+def for_each(
+    foreach: Union[Mapping, str], results: list, ctx: RefContext
+) -> ResultsList:
+    if len(results) == 1 and isinstance(results[0].resolved, MutableSequence):
+        result = _for_each(foreach, enumerate(results[0].resolved), ctx)  # hack!
+    else:
+        result = _for_each(foreach, enumerate(r.external or r.resolved for r in results), ctx)
+    return ResultsList(result, ctx)  # ResultsList[List[ResultOrResultList]]]
 
 
-def for_each_func(foreach, ctx):
+def for_each_func(foreach: Union[Mapping, str], ctx: RefContext):
     results = ctx.currentResource
     if results:
         if isinstance(results, Mapping):
@@ -535,11 +593,10 @@ def set_eval_func(name, val, topLevel=False):
     if topLevel:
         _FuncsTop.append(name)
 
-if sys.version_info >= (3, 9):
-    MutableSequence_Result = MutableSequence[Result]
-else:
-    MutableSequence_Result = MutableSequence
-def eval_ref(val: Union[Mapping, str], ctx: RefContext, top: bool=False) -> MutableSequence_Result:
+
+def eval_ref(
+    val: Union[Mapping, str], ctx: RefContext, top: bool = False
+) -> List[Result]:
     "val is assumed to be an expression, evaluate and return a list of Result"
     from .support import is_template, apply_template
 
@@ -554,7 +611,7 @@ def eval_ref(val: Union[Mapping, str], ctx: RefContext, top: bool=False) -> Muta
             func = _Funcs.get(key)
             if func:
                 args = val[key]
-                ctx.kw = val  # type: ignore
+                ctx.kw = val
                 ctx.currentFunc = key  # type: ignore
                 if "var" in val:
                     unexpected: Union[bool, str] = "var"
@@ -589,7 +646,7 @@ def eval_ref(val: Union[Mapping, str], ctx: RefContext, top: bool=False) -> Muta
         return [Result(mappedVal)]
 
 
-def eval_for_func(val, ctx):
+def eval_for_func(val, ctx) -> Any:
     "like `eval_ref` except it returns the resolved value"
     results = eval_ref(val, ctx)
     if not results:
@@ -648,10 +705,12 @@ def lookup(result, key, context):
         return result
     except (KeyError, IndexError, TypeError, ValueError):
         if context._trace:
-            context.trace("lookup return None due to exception:")
             import traceback
 
-            traceback.print_exc()
+            context.trace(
+                f"lookup of '{key}' returned None due to exception:\n",
+                traceback.format_exc(),
+            )
         return None
 
 
