@@ -8,6 +8,7 @@ import os
 import os.path
 import shutil
 import sys
+import json
 from typing import Any
 import uuid
 import logging
@@ -17,7 +18,7 @@ from pathlib import Path
 from . import DefaultNames, __version__, get_home_config_path, is_version_unreleased
 from .localenv import LocalEnv, Project, LocalConfig
 from .repo import GitRepo, Repo, is_url_or_git_path, split_git_url, commit_secrets
-from .util import UnfurlError, get_random_password
+from .util import UnfurlError, get_random_password, substitute_env
 from .yamlloader import make_yaml, make_vault_lib
 
 _templatePath = os.path.join(os.path.abspath(os.path.dirname(__file__)), "templates")
@@ -151,7 +152,7 @@ def write_ensemble_manifest(
         specDir = os.path.abspath(specDir)
     else:
         specDir = ""
-    vars = dict(specRepoUrl=specRepo.get_url_with_path(specDir))
+    vars = dict(specRepoUrl=specRepo.get_url_with_path(specDir, True))
     if extraVars:
         vars.update(extraVars)
     return write_template(destDir, manifestName, "manifest.yaml.j2", vars, templateDir)
@@ -300,7 +301,7 @@ def render_project(
 
     if ensembleRepo:
         extraVars = dict(
-            ensembleUri=ensembleRepo.get_url_with_path(ensemblePath),
+            ensembleUri=ensembleRepo.get_url_with_path(ensemblePath, True),
             # include the ensembleTemplate in the root of the specDir
             ensembleTemplate=names.EnsembleTemplate,
         )
@@ -452,7 +453,7 @@ def create_project(
         assert newProject
         homeProject = newProject.parentProject
         assert homeProject
-        homeProject.localConfig.register_project(newProject, create_context)
+        homeProject.register_project(newProject, create_context)
 
     if not kw.get("render"):
         if password_vault:
@@ -479,6 +480,7 @@ def clone_local_repos(manifest, sourceProject, targetProject):
         repoSpec = repoView.repository
         if repoSpec.name == "self":
             continue
+        # XXX should look in home project too
         repo = sourceProject.find_git_repo_from_repository(repoSpec)
         if repo:
             targetProject.find_or_clone(repo)
@@ -491,7 +493,7 @@ def _create_ensemble_repo(manifest, repo, commit=True):
     elif not os.path.isdir(destDir):
         os.makedirs(destDir)
 
-    manifest.metadata["uri"] = repo.get_url_with_path(manifest.manifest.path)
+    manifest.metadata["uri"] = repo.get_url_with_path(manifest.manifest.path, True)
     with open(manifest.manifest.path, "w") as f:
         manifest.dump(f)
 
@@ -510,7 +512,7 @@ def _looks_like(path, name):
     return None
 
 
-def _get_ensemble_paths(sourcePath, sourceProject, want_init):
+def _get_ensemble_paths(sourcePath, sourceProject, want_init, use_environment):
     """
     Returns either a pointer to the ensemble to clone
     or a dict of variables to pass to an ensemble template to create a new one
@@ -533,7 +535,9 @@ def _get_ensemble_paths(sourcePath, sourceProject, want_init):
     isServiceTemplate = sourcePath.endswith(DefaultNames.ServiceTemplate)
     if not want_init and not isServiceTemplate:
         try:
-            localEnv = LocalEnv(relPath, project=sourceProject)
+            localEnv = LocalEnv(
+                relPath, project=sourceProject, override_context=use_environment
+            )
             sourceDir = sourceProject.get_relative_path(
                 os.path.dirname(localEnv.manifestPath)
             )
@@ -599,6 +603,7 @@ class EnsembleBuilder:
         self.ensemble_name = ensemble_name
         self.mono = options.get("mono") or options.get("existing")
         self.home_path = get_home_config_path(options.get("home"))
+        self.skeleton_vars = dict((n, v) for n, v in options.get("var", []))
 
         self.source_project = None  # step 1
         self.source_path = None  # step 1 relative path in source_project
@@ -633,17 +638,22 @@ class EnsembleBuilder:
         # source is a path into the project relative to the current directory
         source_path = os.path.join(self.source_project.projectRoot, self.source_path)
         self.templateVars = _get_ensemble_paths(
-            source_path, self.source_project, self.options.get("want_init")
+            source_path,
+            self.source_project,
+            self.options.get("want_init"),
+            self.options.get("use_environment"),
         )
         if self.options.get("use_deployment_blueprint"):
-            self.templateVars['deployment_blueprint'] = self.options['use_deployment_blueprint']
+            self.templateVars["deployment_blueprint"] = self.options[
+                "use_deployment_blueprint"
+            ]
         (self.environment, self.shared_repo) = _get_context_and_shared_repo(
             self.dest_project, self.options
         )
 
     @staticmethod
     def _get_ensemble_dir(targetPath):
-        assert not os.path.isabs(targetPath)
+        assert not os.path.isabs(targetPath), targetPath
         if not targetPath or targetPath == ".":
             destDir, manifestName = (
                 DefaultNames.EnsembleDirectory,
@@ -689,8 +699,14 @@ class EnsembleBuilder:
             self.templateVars,
             self.options.get("skeleton"),
         )
-        localEnv = LocalEnv(manifestPath, project=project)
-        manifest = yamlmanifest.ReadOnlyManifest(localEnv=localEnv)
+        localEnv = LocalEnv(
+            manifestPath,
+            project=project,
+            override_context=self.options.get("use_environment"),
+        )
+        manifest = yamlmanifest.ReadOnlyManifest(
+            localEnv=localEnv, vault=localEnv.get_vault()
+        )
         return localEnv, manifest
 
     def create_new_ensemble(self):
@@ -733,7 +749,9 @@ class EnsembleBuilder:
             # we found a template file to clone
             templateVars["inputs"] = self._get_inputs_template()
             localEnv, manifest = self._create_ensemble_from_template(
-                destProject, destDir, manifestName,
+                destProject,
+                destDir,
+                manifestName,
             )
         else:
             # look for an ensemble at the given path or use the source project's default
@@ -756,6 +774,17 @@ class EnsembleBuilder:
             )
         else:
             destProject.register_ensemble(manifest.path, context=self.environment)
+        environments_json_path = os.path.join(
+            destProject.projectRoot, "environments.json"
+        )
+        if os.path.exists(environments_json_path):
+            # temp hack for unfurl.cloud integration
+            from .to_json import set_deploymentpaths
+
+            envjson = set_deploymentpaths(destProject, environments_json_path)
+            with open(environments_json_path, "w") as f:
+                f.write(json.dumps(envjson, indent=2))
+
         self.manifest = manifest
         return destDir
 
@@ -772,7 +801,7 @@ class EnsembleBuilder:
                 search = os.path.join(newrepo.working_dir, self.source_path)
             else:
                 # just register the source project's repo as a localRepository
-                currentProject.localConfig.register_project(sourceProject, save_project=False)
+                currentProject.register_project(sourceProject, save_project=False)
                 self.source_project = sourceProject
                 return sourceProject
         else:
@@ -816,33 +845,39 @@ class EnsembleBuilder:
         self.source_path = self.source_project.get_relative_path(targetDir)
         return self.source_project
 
+    def _needs_local_config(self, clonedProject):
+        return self.skeleton_vars and not os.path.exists(
+            os.path.join(clonedProject.projectRoot, "local", DefaultNames.LocalConfig)
+        )
+
     def set_dest_project_and_path(
         self, existingSourceProject, existingDestProject, dest
-    ):
+    ) -> bool:
         assert self.dest_project is None
-        new_project = self.source_project is not existingSourceProject
         if existingDestProject:
             #  set that as the dest_project
             self.dest_project = existingDestProject
         else:
             # otherwise set source_project as the dest_project
             self.dest_project = self.source_project
-            if new_project:
-                # finishing creating the new project
-                # create local/unfurl.yaml in the new project
-                if _create_local_config(self.source_project, self.logger):
-                    # reload project with the new local project config
-                    self.dest_project = find_project(
-                        self.source_project.projectRoot, self.home_path
-                    )
+            if self.source_project is not existingSourceProject:
                 # set "" as dest because we already "consumed" dest by cloning the project to that location
                 dest = ""
 
+        assert self.dest_project is not None
+        if not existingDestProject or self._needs_local_config(self.dest_project):
+            # create local/unfurl.yaml in the new project
+            if _create_local_config(self.dest_project, self.logger, self.skeleton_vars):
+                # reload project with the new local project config
+                self.dest_project = find_project(
+                    self.dest_project.projectRoot, self.home_path
+                )
+
         if os.path.isabs(dest):
             relDestDir = self.dest_project.get_relative_path(dest)
             assert not relDestDir.startswith(".."), relDestDir
         else:
-            relDestDir = dest.lstrip(".")
+            relDestDir = dest.lstrip("./")
         if (
             self.ensemble_name
             and self.ensemble_name != DefaultNames.EnsembleDirectory
@@ -852,7 +887,7 @@ class EnsembleBuilder:
             relDestDir = self.ensemble_name
         self.dest_path = relDestDir
 
-    def set_existing_ensemble(self, sourceProject):
+    def has_existing_ensemble(self, sourceProject):
         from unfurl import yamlmanifest
 
         if self.source_project is not sourceProject and not self.shared_repo:
@@ -860,10 +895,8 @@ class EnsembleBuilder:
                 Path(self.dest_project.projectRoot) / self.dest_path
             ):
                 # the ensemble is already part of the source project repository or a submodule
-                localEnv = self.templateVars["localEnv"]
-                self.manifest = yamlmanifest.ReadOnlyManifest(localEnv=localEnv)
-                return self.manifest
-        return None
+                return True
+        return False
 
     def set_source(self, sourceProject):
         self.source_project = sourceProject
@@ -875,7 +908,7 @@ class EnsembleBuilder:
     def set_ensemble(self, isRemote, existingSourceProject, existingDestProject):
         sourceWasCloned = self.source_project is not existingSourceProject
         destIsNew = not existingDestProject
-        if destIsNew and self.set_existing_ensemble(existingSourceProject):
+        if destIsNew and self.has_existing_ensemble(existingSourceProject):
             # if dest_project is new (we just cloned it)
             # check if we cloned the ensemble already
             # if so we done, we don't need to create a new one
@@ -903,7 +936,12 @@ class EnsembleBuilder:
         return f'Created new ensemble at "{os.path.abspath(destDir)}"'
 
 
-def clone(source: str, dest: str, ensemble_name: str=DefaultNames.EnsembleDirectory, **options: Any) -> str:
+def clone(
+    source: str,
+    dest: str,
+    ensemble_name: str = DefaultNames.EnsembleDirectory,
+    **options: Any,
+) -> str:
     """
     Clone the ``source`` ensemble to ``dest``. If ``dest`` isn't in a project, create one.
     ``source`` can point to an ensemble_template, a service_template, an existing ensemble
@@ -923,11 +961,15 @@ def clone(source: str, dest: str, ensemble_name: str=DefaultNames.EnsembleDirect
     ======================= =============================================
 
     """
+    builder = EnsembleBuilder(source, ensemble_name, options)
     if not dest:
         dest = Repo.get_path_for_git_repo(source)  # choose dest based on source url
-    # XXX else: # we're assuming dest is directory, handle case where filename is included
-
-    builder = EnsembleBuilder(source, ensemble_name, options)
+        if os.path.exists(dest):
+            raise UnfurlError(
+                'Can not clone project to "'
+                + dest
+                + '": file already exists with that name'
+            )
     currentProject = find_project(dest, builder.home_path)
 
     ### step 1: clone the source repository and set the the source path
@@ -957,6 +999,7 @@ def clone(source: str, dest: str, ensemble_name: str=DefaultNames.EnsembleDirect
             # dest is in the source project's repo
             # so don't need to clone, just need to create an ensemble
             builder.set_source(sourceProject)
+            dest = os.path.abspath(dest)
 
     assert builder.source_project
 
@@ -974,11 +1017,10 @@ def clone(source: str, dest: str, ensemble_name: str=DefaultNames.EnsembleDirect
     return builder.set_ensemble(isRemote, sourceProject, currentProject)
 
 
-def _create_local_config(clonedProject, logger):
+def _create_local_config(clonedProject, logger, vars):
     local_template = os.path.join(
         clonedProject.projectRoot, DefaultNames.LocalConfigTemplate
     )
-    PLACEHOLDER = "$generate_new_vault_password"
     if os.path.isfile(local_template):
         with open(local_template) as s:
             contents = s.read()
@@ -989,9 +1031,12 @@ def _create_local_config(clonedProject, logger):
         dest = os.path.join(
             clonedProject.projectRoot, "local", DefaultNames.LocalConfig
         )
-        if PLACEHOLDER in contents:
-            contents = contents.replace(PLACEHOLDER, get_random_password())
-            _warn_about_new_password(dest)
+        # log warning if invoked
+        vars["generate_new_vault_password"] = (
+            lambda: _warn_about_new_password(dest) or get_random_password()
+        )
+        # replace ${var}
+        contents = substitute_env(contents, vars)
 
         _write_file(
             os.path.join(clonedProject.projectRoot, "local"),
@@ -1102,6 +1147,27 @@ def _add_unfurl_to_venv(projectdir):
     return ""
 
 
+def pipfile_template_dir(pythonPath):
+    from pipenv.utils import python_version
+
+    versionStr = python_version(pythonPath)
+    assert versionStr, versionStr
+    version = versionStr.rpartition(".")[0]  # 3.8.1 => 3.8
+    # version = subprocess.run([pythonPath, "-V"]).stdout.decode()[
+    #     7:10
+    # ]  # e.g. Python 3.8.1 => 3.8
+    return os.path.join(_templatePath, "python" + version)  # e.g. templates/python3.8
+
+
+def copy_pipfiles(pipfileLocation, projectDir):
+    # copy Pipfiles to project root
+    if os.path.abspath(projectDir) != os.path.abspath(pipfileLocation):
+        for filename in ["Pipfile", "Pipfile.lock"]:
+            path = os.path.join(pipfileLocation, filename)
+            if os.path.isfile(path):
+                shutil.copy(path, projectDir)
+
+
 def create_venv(projectDir, pipfileLocation, unfurlLocation):
     """Create a virtual python environment for the given project."""
 
@@ -1120,31 +1186,15 @@ def create_venv(projectDir, pipfileLocation, unfurlLocation):
         # need to set env vars and change current dir before importing pipenv
         from pipenv import environments
         from pipenv.core import do_install
-        from pipenv.utils import python_version
         from pipenv.project import Project as PipEnvProject
 
         pythonPath = os.environ["PIPENV_PYTHON"]
         assert pythonPath, pythonPath
         if not pipfileLocation:
-            versionStr = python_version(pythonPath)
-            assert versionStr, versionStr
-            version = versionStr.rpartition(".")[0]  # 3.8.1 => 3.8
-            # version = subprocess.run([pythonPath, "-V"]).stdout.decode()[
-            #     7:10
-            # ]  # e.g. Python 3.8.1 => 3.8
-            pipfileLocation = os.path.join(
-                _templatePath, "python" + version
-            )  # e.g. templates/python3.8
-
+            pipfileLocation = pipfile_template_dir(pythonPath)
         if not os.path.isdir(pipfileLocation):
             return f'Pipfile location is not a valid directory: "{pipfileLocation}"'
-
-        # copy Pipfiles to project root
-        if os.path.abspath(projectDir) != os.path.abspath(pipfileLocation):
-            for filename in ["Pipfile", "Pipfile.lock"]:
-                path = os.path.join(pipfileLocation, filename)
-                if os.path.isfile(path):
-                    shutil.copy(path, projectDir)
+        copy_pipfiles(pipfileLocation, projectDir)
 
         kw = dict(python=pythonPath, extra_index_url=[])
         pipenv_project = PipEnvProject()
