@@ -9,7 +9,17 @@ By convention, the "home" project defines a localhost instance and adds it to it
 """
 import os
 import os.path
-from typing import Any, Iterable, List, Optional, Tuple, Union, TYPE_CHECKING
+from typing import (
+    Any,
+    Iterable,
+    Dict,
+    List,
+    Optional,
+    OrderedDict,
+    Tuple,
+    Union,
+    TYPE_CHECKING,
+)
 
 from ansible.parsing.vault import VaultLib
 import six
@@ -17,11 +27,18 @@ from six import Iterator
 from unfurl.runtime import NodeInstance
 
 from .repo import GitRepo, Repo, split_git_url, RepoView, normalize_git_url_hard
-from .util import UnfurlError
+from .util import (
+    UnfurlError,
+    substitute_env,
+    wrap_sensitive_value,
+    save_to_tempfile,
+    is_sensitive,
+)
 from .merge import merge_dicts
 from .yamlloader import YamlConfig, make_vault_lib_ex, make_yaml
 from . import DefaultNames, get_home_config_path
-from six.moves.urllib.parse import urlparse
+from urllib.parse import urlparse, urlunsplit, urlsplit
+from ruamel.yaml.comments import CommentedMap
 from toscaparser.repositories import Repository
 
 if TYPE_CHECKING:
@@ -31,9 +48,9 @@ if TYPE_CHECKING:
 _basepath = os.path.abspath(os.path.dirname(__file__))
 
 
-import logging
+from .logs import getLogger
 
-logger = logging.getLogger("unfurl")
+logger = getLogger("unfurl")
 
 
 class Project:
@@ -42,11 +59,19 @@ class Project:
     one or more ensemble.yaml files which maybe optionally organized into one or more git repositories.
     """
 
-    def __init__(self, path: str, homeProject: Optional["Project"]=None):
+    def __init__(
+        self,
+        path: str,
+        homeProject: Optional["Project"] = None,
+        overrides: Optional[dict] = None,
+    ):
         assert isinstance(path, six.string_types), path
         self.projectRoot = os.path.abspath(os.path.dirname(path))
+        self.overrides = overrides or {}
         if os.path.exists(path):
-            self.localConfig = LocalConfig(path)
+            self.localConfig = LocalConfig(
+                path, yaml_include_hook=self.load_yaml_include
+            )
         else:
             self.localConfig = LocalConfig()
         self._set_repos()
@@ -61,7 +86,7 @@ class Project:
         )
         self.parentProject = parentProject
         if parentProject:
-            parentProject.localConfig.register_project(self)
+            parentProject.register_project(self)
         self._set_contexts()
         # depends on _set_contexts():
         self.project_repoview.yaml = make_yaml(self.make_vault_lib())  # type: ignore
@@ -92,7 +117,7 @@ class Project:
 
     def _set_repos(self) -> None:
         # abspath => RepoView:
-        self.workingDirs = Repo.find_git_working_dirs(self.projectRoot)
+        self.workingDirs = Repo.find_git_working_dirs(self.projectRoot, True)
         self._set_project_repoview()
 
         if self.project_repoview.repo:
@@ -128,7 +153,7 @@ class Project:
             return path
 
     @staticmethod
-    def find_path(testPath: str, stopPath: Optional[str]=None) -> Optional[str]:
+    def find_path(testPath: str, stopPath: Optional[str] = None) -> Optional[str]:
         """
         Walk parents looking for unfurl.yaml
         """
@@ -219,27 +244,50 @@ class Project:
             counter += 1
         return os.path.join(self.projectRoot, name)
 
-    def find_git_repo(self, repoURL: str, revision: Optional[str]=None) -> Optional[GitRepo]:
+    def _find_git_repo(
+        self, repoURL: str, revision: Optional[str] = None
+    ) -> Union[GitRepo, str]:
+        # if repo isn't found, return the repo url, possibly rewritten to include server credentials
         candidate = None
+        # repourl_parts = urlsplit(repoURL)
+        normalized = normalize_git_url_hard(repoURL)
         for dir, repository in self.workingDirs.items():
             repo = repository.repo
             if repoURL.startswith("git-local://"):
                 initialCommit = urlparse(repoURL).netloc.partition(":")[0]
                 match = initialCommit == repo.get_initial_revision()
             else:
-                match = normalize_git_url_hard(repoURL) == normalize_git_url_hard(repo.url)
+                match = normalized == normalize_git_url_hard(repo.url)
             if match:
                 if revision:
                     if repo.revision == repo.resolve_rev_spec(revision):
                         return repo
+                    # revisions don't match but we'll return it if we don't find a better candidate
                     candidate = repo
                 else:
                     return repo
-        return candidate
+            # elif not repourl_parts.username:  # if url doesn't have credentials already
+            #     candidate_parts = urlsplit(repo.url)
+            #     if (
+            #         candidate_parts.password
+            #         and candidate_parts.hostname == repourl_parts.hostname
+            #     ):
+            #         # rewrite repoUrl to add credentials
+            #         repoURL = urlunsplit(
+            #             repourl_parts._replace(netloc=candidate_parts.netloc)
+            #         )
+        return candidate or repoURL
 
-    def find_path_in_repos(self,
-        path: str,
-        importLoader: Optional[Any]=None
+    def find_git_repo(
+        self, repoURL: str, revision: Optional[str] = None
+    ) -> Optional[GitRepo]:
+        repo_or_url = self._find_git_repo(repoURL, revision)
+        if isinstance(repo_or_url, str):
+            return None
+        return repo_or_url
+
+    def find_path_in_repos(
+        self, path: str, importLoader: Optional[Any] = None
     ) -> Tuple[Optional[GitRepo], Optional[str], Optional[str], Optional[bool]]:
         """If the given path is part of the working directory of a git repository
         return that repository and a path relative to it"""
@@ -260,7 +308,7 @@ class Project:
             #         candidate = (repo, filePath, revision, bare)
         return candidate or None, None, None, None
 
-    def create_working_dir(self, gitUrl: str, ref: Optional[str]=None) -> GitRepo:
+    def create_working_dir(self, gitUrl: str, ref: Optional[str] = None) -> GitRepo:
         localRepoPath = self._create_path_for_git_repo(gitUrl)
         repo = Repo.create_working_dir(gitUrl, localRepoPath, ref)
         # add to workingDirs
@@ -272,30 +320,30 @@ class Project:
         return self.find_git_repo(split_git_url(repoUrl)[0])
 
     def find_or_clone(self, repo: GitRepo) -> GitRepo:
-        gitUrl = repo.url
-        existingRepo = self.find_git_repo(gitUrl)
-        if existingRepo:
-            return existingRepo
+        for dir, repository in self.workingDirs.items():
+            if repo.get_initial_revision() == repository.repo.get_initial_revision():
+                return repository.repo
 
-        # if not found:
+        gitUrl = repo.url
         localRepoPath = os.path.abspath(
             self._create_path_for_git_repo(repo.working_dir or gitUrl)
         )
+        logger.trace('Cloning "%s" to: %s', repo.working_dir or gitUrl, localRepoPath)
         newRepo = repo.clone(localRepoPath)
         # use gitUrl to preserve original origin
         self.workingDirs[localRepoPath] = RepoView(dict(name="", url=gitUrl), newRepo)
         return newRepo
 
-    def find_or_create_working_dir(self, repoURL: str, revision: Optional[str]=None) -> Optional[GitRepo]:
+    def find_or_create_working_dir(
+        self, repoURL: str, revision: Optional[str] = None
+    ) -> Optional[GitRepo]:
         repo = self.find_git_repo(repoURL, revision)
         if not repo:
             repo = self.create_working_dir(repoURL, revision)
         return repo
 
-    def get_manifest_path(self,
-        localEnv: "LocalEnv",
-        manifestPath: str,
-        can_be_empty: bool
+    def get_manifest_path(
+        self, localEnv: "LocalEnv", manifestPath: str, can_be_empty: bool
     ) -> Tuple[Optional["Project"], str, str]:
         if manifestPath:
             # at this point a named manifest
@@ -347,7 +395,9 @@ class Project:
             )
         self.contexts = contexts
 
-    def get_context(self, contextName: Optional[str], context: Optional[dict]=None) -> dict:
+    def get_context(
+        self, contextName: Optional[str], context: Optional[dict] = None
+    ) -> dict:
         # merge the named context with defaults and the given context
         localContext = self.contexts.get("defaults") or {}
         if contextName and contextName != "defaults" and self.contexts.get(contextName):
@@ -361,41 +411,59 @@ class Project:
             context or {}, localContext, replaceKeys=LocalConfig.replaceKeys
         )
 
-    def get_vault_password(self, contextName: Optional[str]=None, vaultId: str="default") -> Optional[str]:
-        secret = os.getenv(f"UNFURL_VAULT_{vaultId.upper()}_PASSWORD")
+    @staticmethod
+    def _check_vault_env_var(context, vaultId, default=None):
+        env_var = f"UNFURL_VAULT_{vaultId.upper()}_PASSWORD"
+        secret = os.getenv(env_var, default)
         if secret:
             return secret
+        return context.get("variables", {}).get(env_var, default)
+
+    def get_vault_password(
+        self, contextName: Optional[str] = None, vaultId: str = "default"
+    ) -> Optional[str]:
         context = self.get_context(contextName)  # type: ignore
-        secrets = context.get("secrets", {}).get(f"vault_secrets")
+        secret = self._check_vault_env_var(context, vaultId)
+        if secret:
+            return secret
+        secrets = context.get("secrets", {}).get("vault_secrets")
         if secrets:
             return secrets.get(vaultId)
         return None
 
-    def get_vault_passwords(self, contextName: Optional[str]=None) -> Iterable[Tuple[str, Union[str, bytes]]]:
+    def get_vault_passwords(
+        self, contextName: Optional[str] = None
+    ) -> Iterable[Tuple[str, Union[str, bytes]]]:
         # we want to support multiple vault ids
         context = self.get_context(contextName)  # type: ignore
-        secrets = context.get("secrets", {}).get(f"vault_secrets")
+        secrets = context.get("secrets", {}).get("vault_secrets")
         if not secrets:
             return
         for vaultId, password in secrets.items():
-            password = os.getenv(f"UNFURL_VAULT_{vaultId.upper()}_PASSWORD", password)
+            password = self._check_vault_env_var(context, vaultId, password)
             if password:
                 yield vaultId, password
 
-    def make_vault_lib(self, contextName: Optional[str]=None) -> Optional[VaultLib]:
+    def make_vault_lib(self, contextName: Optional[str] = None) -> Optional[VaultLib]:
         secrets = list(self.get_vault_passwords(contextName))
         if secrets:
             return make_vault_lib_ex(secrets)
         return None
 
-    def find_ensemble_by_path(self, path: str) -> Optional[dict]:
+    @staticmethod
+    def _find_ensemble_by_path(ensembles, projectRoot, path: str) -> Optional[dict]:
         path = os.path.abspath(path)
-        for tpl in self.localConfig.ensembles:
+        for tpl in ensembles:
             file = tpl.get("file")
-            if file and os.path.normpath(os.path.join(self.projectRoot, file)) == path:
+            if file and os.path.normpath(os.path.join(projectRoot, file)) == path:
                 return tpl
         return None
 
+    def find_ensemble_by_path(self, path: str) -> Optional[dict]:
+        return self._find_ensemble_by_path(
+            self.localConfig.ensembles, self.projectRoot, path
+        )
+
     def find_ensemble_by_name(self, name: str) -> Optional[dict]:
         for tpl in self.localConfig.ensembles:
             alias = tpl.get("alias")
@@ -436,7 +504,9 @@ class Project:
             % (default_project, context_name)
         )
 
-    def get_managed_project(self, location: dict, localEnv: "LocalEnv") -> Optional["Project"]:
+    def get_managed_project(
+        self, location: dict, localEnv: "LocalEnv"
+    ) -> Optional["Project"]:
         project_tpl = self.localConfig.projects[location["managedBy"]]
         path = self.localConfig.find_repository_path(project_tpl)
         externalProject = None
@@ -445,17 +515,20 @@ class Project:
             if externalProject:
                 externalProject._set_parent_project(self)
                 return externalProject
-        if not externalProject:
             localEnv.logger.warning(
                 'Could not find the project "%s" which manages "%s"',
                 location["managedBy"],
                 path,
             )
-            return self
-        return None
+        return self
 
     def register_ensemble(
-        self, manifestPath: str, *, project: "Project"=None, managedBy: Optional["Project"]=None, context: Optional[str]=None
+        self,
+        manifestPath: str,
+        *,
+        project: "Project" = None,
+        managedBy: Optional["Project"] = None,
+        context: Optional[str] = None,
     ) -> None:
         relPath = (project if project else self).get_relative_path(manifestPath)
         props = dict(file=relPath)
@@ -473,11 +546,75 @@ class Project:
             self.localConfig.ensembles.append(props)
         if managedBy or project:
             assert not (managedBy and project)
-            self.localConfig.register_project(managedBy or project, changed=True)
+            self.register_project(managedBy or project, changed=True)
         elif context:
             self.localConfig.config.config["ensembles"] = self.localConfig.ensembles
             self.localConfig.config.save()
 
+    def register_project(
+        self, project, for_context=None, changed=False, save_project=True
+    ):
+        self.localConfig.register_project(project, for_context, changed, save_project)
+        self.workingDirs[
+            project.project_repoview.working_dir
+        ] = project.project_repoview
+
+    def load_yaml_include(
+        self,
+        yamlConfig,
+        templatePath,
+        baseDir,
+        warnWhenNotFound=False,
+        expanded=None,
+        action=False,
+    ):
+        """
+        This is called while the YAML config is being loaded.
+        Returns (url or fullpath, parsed yaml)
+        """
+        if isinstance(templatePath, dict):
+            key = templatePath["file"]
+            merge = templatePath.get("merge")
+        else:
+            key = templatePath
+            merge = None
+
+        url_vars = os.environ.copy()
+        url_vars.update(self.overrides)
+        if action:
+            if isinstance(action, str):
+                return substitute_env(action, url_vars)
+            # check:
+            return True
+
+        if merge == "maplist" and "ENVIRONMENT" not in self.overrides:
+            # don't load remote includes for LocalEnv that don't specify an environment
+            # XXX (hackish way to avoid loads for transitory LocalEnv objects)
+            logger.trace("skipping retrieving url with ENVIRONMENT: %s", key)
+            return key, None
+
+        if not key:
+            return key, None
+        assert isinstance(key, str)
+        key = substitute_env(key, url_vars)
+        includekey, template = yamlConfig.load_yaml(key, baseDir, warnWhenNotFound)
+        if merge == "maplist" and template is not None:
+            environment = url_vars.get("ENVIRONMENT")
+            if not environment:
+                environment = expanded.get("default_environment")
+                ensembles = expanded.get("ensembles") or []
+                if "manifest_path" in url_vars:
+                    ensemble_tpl = self._find_ensemble_by_path(
+                        ensembles, self.projectRoot, url_vars["manifest_path"]
+                    )
+                else:
+                    ensemble_tpl = LocalConfig._get_default_manifest_tpl(ensembles)
+                if ensemble_tpl:
+                    environment = ensemble_tpl.get("environment", environment)
+            template = CommentedMap(_maplist(template, environment))
+            logger.debug("retrieved remote environment vars: %s", template)
+        return includekey, template
+
 
 class LocalConfig:
     """
@@ -494,12 +631,17 @@ class LocalConfig:
         "connections",
         "variables",
         "repositories",
+        "instances",
     ]
 
-    def __init__(self, path=None, validate=True):
+    def __init__(self, path=None, validate=True, yaml_include_hook=None):
         defaultConfig = {"apiVersion": "unfurl/v1alpha1", "kind": "Project"}
         self.config = YamlConfig(
-            defaultConfig, path, validate, os.path.join(_basepath, "unfurl-schema.json")
+            defaultConfig,
+            path,
+            validate,
+            os.path.join(_basepath, "unfurl-schema.json"),
+            yaml_include_hook,
         )
         self.ensembles = self.config.expanded.get("ensembles") or []
         self.projects = self.config.expanded.get("projects") or {}
@@ -511,22 +653,19 @@ class LocalConfig:
         """
         return os.path.join(self.config.get_base_dir(), path)
 
-    def get_default_manifest_tpl(self):
-        """
-        ensembles:
-          default:
-            file:
-            project:
-            context
-        """
-        if len(self.ensembles) == 1:
-            return self.ensembles[0]
+    @staticmethod
+    def _get_default_manifest_tpl(ensembles):
+        if len(ensembles) == 1:
+            return ensembles[0]
         else:
-            for tpl in self.ensembles:
+            for tpl in ensembles:
                 if tpl.get("default"):
                     return tpl
         return None
 
+    def get_default_manifest_tpl(self):
+        return self._get_default_manifest_tpl(self.ensembles)
+
     def create_local_instance(self, localName, attributes):
         # local or secret
         from .runtime import NodeInstance
@@ -583,7 +722,9 @@ class LocalConfig:
 
         return name
 
-    def register_project(self, project, for_context=None, changed=False, save_project=True):
+    def register_project(
+        self, project, for_context=None, changed=False, save_project=True
+    ):
         """
         Register an external project with current project.
         If the external project's repository is only local (without a remote origin repository) then save it in the local config if it exists.
@@ -640,7 +781,21 @@ class LocalConfig:
                 self.config.save_include(key)
             self.config.config["ensembles"] = self.ensembles
             self.config.save()
-        return name
+        return lock
+
+
+def _maplist(template, environment_scope=None):
+    for var in template:
+        scope = var.get("environment_scope", "*")
+        if scope != "*" and scope != environment_scope:
+            # match or if environment_scope is None skip any != *
+            continue
+        value = var["value"]
+        if var.get("variable_type") == "file":
+            value = save_to_tempfile(value, var["key"]).name
+        elif var.get("masked"):
+            value = wrap_sensitive_value(value)
+        yield var["key"], value
 
 
 class LocalEnv:
@@ -649,10 +804,12 @@ class LocalEnv:
     the local project it is part of and the home project.
     """
 
-    project = None
-    parent = None
+    project: Optional[Project] = None
+    parent: Optional["LocalEnv"] = None
 
-    def _find_external_project(self, manifestPath: str, project:  Project) -> Tuple[Optional[Project], Optional[str]]:
+    def _find_external_project(
+        self, manifestPath: str, project: Project
+    ) -> Tuple[Optional[Project], Optional[str]]:
         # We're pointing directly at a manifest path, check if it is might be part of an external project
         context_name = None
         location = project.find_ensemble_by_path(manifestPath)
@@ -677,6 +834,8 @@ class LocalEnv:
         if foundManifestPath:
             self.manifestPath = foundManifestPath
             if not project:
+                # set this because the yaml loader needs access to this while the project is being instantiated
+                self.overrides["manifest_path"] = foundManifestPath
                 project = self.find_project(os.path.dirname(foundManifestPath))
         elif project:
             # the manifestPath was pointing to a project, not a manifest
@@ -713,12 +872,12 @@ class LocalEnv:
 
     def __init__(
         self,
-        manifestPath: str=None,
-        homePath: Optional[str]=None,
-        parent: Optional["LocalEnv"]=None,
-        project: Optional[Project]=None,
-        can_be_empty: bool=False,
-        override_context: Optional[str]=None,
+        manifestPath: str = None,
+        homePath: Optional[str] = None,
+        parent: Optional["LocalEnv"] = None,
+        project: Optional[Project] = None,
+        can_be_empty: bool = False,
+        override_context: Optional[str] = None,
     ) -> None:
         """
         If manifestPath is None find the first unfurl.yaml or ensemble.yaml
@@ -733,6 +892,11 @@ class LocalEnv:
         logger = logging.getLogger("unfurl")
         self.logger = logger
         self.manifest_context_name = None
+        self.overrides: Dict[str, Any] = {}
+        if override_context is not None:
+            # aka the --use-environment option
+            # hackishly, "" is a valid option used by load_yaml_include
+            self.overrides["ENVIRONMENT"] = override_context
 
         if parent:
             self.parent = parent
@@ -768,10 +932,14 @@ class LocalEnv:
             or LocalConfig()
         )
         if override_context and override_context != "defaults":
-            assert override_context == self.manifest_context_name, self.manifest_context_name
+            assert (
+                override_context == self.manifest_context_name
+            ), self.manifest_context_name
             project = self.project or self.homeProject
             if not project or self.manifest_context_name not in project.contexts:
-                raise UnfurlError(f'No environment named "{self.manifest_context_name}" found.')
+                raise UnfurlError(
+                    f'No environment named "{self.manifest_context_name}" found.'
+                )
 
     def _get_home_project(self) -> Optional[Project]:
         homeProject = None
@@ -791,16 +959,33 @@ class LocalEnv:
                     self.logger.info('Using home project at: "%s"', self.homeConfigPath)
         return homeProject
 
-    def get_vault_password(self, vaultId: str="default") -> Optional[str]:
+    def get_vault_password(self, vaultId: str = "default") -> Optional[str]:
         # used by __main__.vaultclient
         project = self.project or self.homeProject
         if project:
             return project.get_vault_password(self.manifest_context_name, vaultId)
         return None
 
-    def get_manifest(self,
-        path: Optional[str]=None,
-        skip_validation: bool=False
+    def get_vault(self):
+        project = self.project or self.homeProject
+        if project:
+            vault = project.make_vault_lib(self.manifest_context_name)
+            if vault:
+                self.logger.info(
+                    "Vault password found, configuring vault ids: %s",
+                    [s[0] for s in vault.secrets],
+                )
+        else:
+            vault = None
+        if not vault:
+            msg = "No vault password found"
+            if self.manifest_context_name:
+                msg += f" for environment {self.manifest_context_name}"
+            self.logger.debug(msg)
+        return vault
+
+    def get_manifest(
+        self, path: Optional[str] = None, skip_validation: bool = False
     ) -> "YamlManifest":
         from .yamlmanifest import YamlManifest
 
@@ -813,21 +998,7 @@ class LocalEnv:
             manifest = self._manifests.get(self.manifestPath)
             if not manifest:
                 # should load vault ids from context
-                project = self.project or self.homeProject
-                if project:
-                    vault = project.make_vault_lib(self.manifest_context_name)
-                    if vault:
-                        self.logger.info(
-                            "Vault password found, configuring vault ids: %s",
-                            [s[0] for s in vault.secrets],
-                        )
-                else:
-                    vault = None
-                if not vault:
-                    msg = "No vault password found"
-                    if self.manifest_context_name:
-                        msg += f" for environment {self.manifest_context_name}"
-                    self.logger.debug(msg)
+                vault = self.get_vault()
                 manifest = YamlManifest(
                     localEnv=self, vault=vault, skip_validation=skip_validation
                 )
@@ -838,7 +1009,7 @@ class LocalEnv:
         path = Project.normalize_path(path)
         project = self._projects.get(path)
         if not project:
-            project = Project(path, homeProject)
+            project = Project(path, homeProject, self.overrides)
             self._projects[path] = project
         return project
 
@@ -849,9 +1020,8 @@ class LocalEnv:
         else:
             return None
 
-    def _get_external_localenv(self,
-        location: dict,
-        currentProject: Optional[Project]=None
+    def _get_external_localenv(
+        self, location: dict, currentProject: Optional[Project] = None
     ) -> Optional["LocalEnv"]:
         # currentProject because this method might be called before self.project was set
         assert "project" in location
@@ -887,7 +1057,9 @@ class LocalEnv:
             os.path.join(projectRoot, location.get("file") or ""), parent=self
         )
 
-    def _find_given_manifest_or_project(self, manifestPath: str) -> Tuple[Optional[str], Optional[Project]]:
+    def _find_given_manifest_or_project(
+        self, manifestPath: str
+    ) -> Tuple[Optional[str], Optional[Project]]:
         if not os.path.exists(manifestPath):
             return None, None
         if os.path.isdir(manifestPath):
@@ -934,7 +1106,9 @@ class LocalEnv:
         else:
             return repos
 
-    def _search_for_manifest_or_project(self, dir: str) -> Tuple[str, Optional[Project]]:
+    def _search_for_manifest_or_project(
+        self, dir: str
+    ) -> Tuple[str, Optional[Project]]:
         current = os.path.abspath(dir)
         while current and current != os.sep:
             test = os.path.join(current, DefaultNames.LocalConfig)
@@ -963,7 +1137,7 @@ class LocalEnv:
             return self.get_project(path, self.homeProject)
         return None
 
-    def get_context(self, context: Optional[dict]=None) -> dict:
+    def get_context(self, context: Optional[dict] = None) -> dict:
         """
         Return a new context that merges the given context with the local context.
         """
@@ -993,7 +1167,9 @@ class LocalEnv:
             spec,
         )
 
-    def find_git_repo(self, repoURL: str, revision: Optional[str]=None) -> Optional[GitRepo]:
+    def _find_git_repo(
+        self, repoURL: str, revision: Optional[str] = None
+    ) -> Union[GitRepo, str]:
         project = self.project or self.homeProject
         count = 0
         while project:
@@ -1002,54 +1178,75 @@ class LocalEnv:
                 project.projectRoot,
                 project.parentProject and project.parentProject.projectRoot,
             )
-            repo = project.find_git_repo(repoURL, revision)
-            if repo:
-                return repo
+            candidate = project._find_git_repo(repoURL, revision)
+            if isinstance(candidate, GitRepo):
+                return candidate
+            elif candidate != repoURL:
+                repoURL = candidate
             project = project.parentProject
-        return None
+        return repoURL
+
+    def find_git_repo(
+        self, repoURL: str, revision: Optional[str] = None
+    ) -> Optional[GitRepo]:
+        repo_or_url = self._find_git_repo(repoURL, revision)
+        if isinstance(repo_or_url, str):
+            return None
+        return repo_or_url
 
     def _create_working_dir(self, repoURL, revision, basepath):
         project = self.project or self.homeProject
         if not project:
+            logger.warning(
+                "Can not create clone repository, ensemble is not in an Unfurl project."
+            )
             return None
         while project:
             if basepath is None or project.is_path_in_project(basepath):
                 repo = project.create_working_dir(repoURL, revision)
                 break
             project = project.parentProject
-        else:
-            repo = (self.project or self.homeProject).create_working_dir(
+        else:  # no break
+            repo = (self.project or self.homeProject).create_working_dir(  # type: ignore
                 repoURL, revision
             )
         return repo
 
-    def find_or_create_working_dir(self,
+    def find_or_create_working_dir(
+        self,
         repoURL: str,
-        revision: Optional[str]=None,
-        basepath: Optional[str]=None
+        revision: Optional[str] = None,
+        basepath: Optional[str] = None,
     ) -> Tuple[Optional[GitRepo], Optional[str], Optional[bool]]:
-        repo = self.find_git_repo(repoURL, revision)
-        if repo:
-            logger.debug("Using existing repository at %s for %s", repo.working_dir, repoURL)
+        repo = self._find_git_repo(repoURL, revision)
+        if isinstance(repo, GitRepo):
+            logger.debug(
+                "Using existing repository at %s for %s", repo.working_dir, repoURL
+            )
             if not repo.is_dirty():
                 repo.pull(revision=revision)
+        else:
+            assert isinstance(
+                repo, str
+            ), repo  # it's the repoUrl (possibly rewritten) at this point
+            # git-local repos must already exist locally
+            if repo.startswith("git-local://"):
+                return None, None, None
 
-        # git-local repos must already exist
-        if not repo and not repoURL.startswith("git-local://"):
-            repo = self._create_working_dir(repoURL, revision, basepath)
-        if not repo:
-            return None, None, None
+            repo = self._create_working_dir(repo, revision, basepath)
+            if not repo:
+                return None, None, None
 
+        assert isinstance(repo, GitRepo)
         if revision:
-            # bare if HEAD isn't at the requested revision
-            bare = repo.revision != repo.resolve_rev_spec(revision)
-            return repo, repo.revision, bare
+            if repo.revision != repo.resolve_rev_spec(revision):
+                repo.checkout(revision)
+            return repo, repo.revision, False
         else:
             return repo, repo.revision, False
 
-    def find_path_in_repos(self,
-        path: str,
-        importLoader: Optional[Any]=None
+    def find_path_in_repos(
+        self, path: str, importLoader: Optional[Any] = None
     ) -> Tuple[Optional[GitRepo], Optional[str], Optional[str], Optional[bool]]:
         """If the given path is part of the working directory of a git repository
         return that repository and a path relative to it"""
@@ -1110,14 +1307,19 @@ class LocalEnv:
                 if os.path.isdir(homeAsdf):
                     asdfDataDir = homeAsdf
         if asdfDataDir:  # asdf is installed
+            os.environ["ASDF_DATA_DIR"] = asdfDataDir
             # current project has higher priority over home project
             project = self.project or self.homeProject
             count = 0
             while project:
-                paths += Project.get_asdf_paths(project.projectRoot, asdfDataDir, self.toolVersions)
+                paths += Project.get_asdf_paths(
+                    project.projectRoot, asdfDataDir, self.toolVersions
+                )
                 count += 1
                 assert count < 4
                 project = project.parentProject
             if os.getenv("HOME"):
-                paths += Project.get_asdf_paths(os.getenv("HOME"), asdfDataDir, self.toolVersions)
+                paths += Project.get_asdf_paths(
+                    os.getenv("HOME"), asdfDataDir, self.toolVersions
+                )
         return paths
