@@ -4,10 +4,18 @@ from collections.abc import Mapping
 import os.path
 import hashlib
 import json
+from typing import Tuple, Any
 from ruamel.yaml.comments import CommentedMap
 from .tosca import ToscaSpec, TOSCA_VERSION, ArtifactSpec
 
-from .support import ResourceChanges, AttributeManager, Status, Priority, NodeState
+from .support import (
+    ResourceChanges,
+    AttributeManager,
+    Status,
+    Priority,
+    NodeState,
+    Imports,
+)
 from .runtime import (
     OperationalInstance,
     NodeInstance,
@@ -31,6 +39,31 @@ logger = logging.getLogger("unfurl")
 _basepath = os.path.abspath(os.path.dirname(__file__))
 
 
+def relabel_dict(context, localEnv, key):
+    connections = context.get(key)
+    if not connections:
+        return {}
+    contexts = {}
+    if localEnv:
+        project = localEnv.project or localEnv.homeProject
+        if project:
+            contexts = project.contexts
+
+    # handle items like newname : oldname to rename merged connections
+    def follow_alias(v):
+        if isinstance(v, str):
+            env, sep, name = v.partition(":")
+            if sep:  # found a ":"
+                v = contexts[env][key][name]
+            else:  # look in current dict
+                v = connections[env]
+            return follow_alias(v)  # follow
+        else:
+            return v
+
+    return dict((n, follow_alias(v)) for n, v in connections.items())
+
+
 class Manifest(AttributeManager):
     """
     Base class for managing an ensemble.
@@ -50,6 +83,21 @@ class Manifest(AttributeManager):
         self.tosca = None
         self.specDigest = None
         self.repositories = {}
+        if self.localEnv:
+            # before we start parsing the manifest, add the repositories in the environment
+            self._add_repositories_from_environment()
+        self.imports = Imports()
+        self.imports.manifest = self
+        self._importedManifests = {}
+
+    def _add_repositories_from_environment(self):
+        assert self.localEnv
+        context = self.localEnv.get_context()
+        repositories = relabel_dict(context, self.localEnv, "repositories")
+        resolver = self.get_import_resolver(self)
+        for name, tpl in repositories.items():
+            toscaRepository = resolver.get_repository(name, tpl)
+            self.repositories[name] = RepoView(toscaRepository, None)
 
     def _set_spec(self, spec, more_spec=None, skip_validation=False):
         """
@@ -91,7 +139,9 @@ class Manifest(AttributeManager):
         if more_spec:
             # don't merge individual templates
             toscaDef = merge_dicts(
-                toscaDef, more_spec, replaceKeys=['node_templates', 'relationship_templates']
+                toscaDef,
+                more_spec,
+                replaceKeys=["node_templates", "relationship_templates"],
             )
         if not isinstance(toscaDef, CommentedMap):
             toscaDef = CommentedMap(toscaDef.items())
@@ -306,7 +356,6 @@ class Manifest(AttributeManager):
         return self.changeSets.get(jobId)
 
     def _create_entity_instance(self, ctor, name, status, parent):
-        operational = self.load_status(status)
         templateName = status.get("template", name)
 
         imported = None
@@ -315,19 +364,21 @@ class Manifest(AttributeManager):
             imported = self.imports.find_import(importName)
             if not imported:
                 raise UnfurlError(f"missing import {importName}")
-            if imported.shadow:
-                raise UnfurlError(f"already imported {importName}")
-            template = imported.template
-        else:
-            template = self.load_template(templateName)
 
+        if imported:
+            operational = self.load_status(dict(readyState=imported.local_status))
+        else:
+            operational = self.load_status(status)
+        template = self.load_template(templateName)
         if template is None:
             # not defined in the current model any more, try to retrieve the old version
             if operational.last_config_change:
                 changerecord = self._get_last_change(operational)
                 template = self.load_template(templateName, changerecord)
         if template is None:
-            raise UnfurlError(f"missing template definition for '{templateName}' while instantiating instance '{name}'")
+            raise UnfurlError(
+                f"missing template definition for '{templateName}' while instantiating instance '{name}'"
+            )
         # logger.debug("creating instance for template %s: %s", templateName, template)
 
         # omit keys that match <<REDACTED>> so can we use the computed property
@@ -339,8 +390,11 @@ class Manifest(AttributeManager):
         instance = ctor(name, attributes, parent, template, operational)
         if "created" in status:
             instance.created = status["created"]
+        if "protected" in status:
+            instance.protected = status["protected"]
         if imported:
-            self.imports.set_shadow(importName, instance)
+            instance.imported = importName
+            self.imports.set_shadow(importName, instance, imported)
         properties = status.get("properties")
         if isinstance(properties, dict):
             instance._properties = properties
@@ -393,7 +447,7 @@ class Manifest(AttributeManager):
         return repo, filePath, revision, bare
 
     def add_repository(self, repo, toscaRepository, file_name):
-        repository = self.repositories.get(toscaRepository.name)
+        repository: RepoView = self.repositories.get(toscaRepository.name)
         if repository:
             # already exist, make sure it's the same repo
             repo = repo or repository.repo
@@ -408,11 +462,14 @@ class Manifest(AttributeManager):
         )
         return repository
 
-    def update_repositories(self, config, inlineRepositories=None, resolver=None):
+    def _update_repositories(self, config, inlineRepositories=None, resolver=None):
+        # _update_repositories is called during parse time when including files
         if not resolver:
             resolver = self.get_import_resolver(self)
+        # we need to fetch this every call since the config might have changed:
         repositories = self._get_repositories(config)
         for name, tpl in repositories.items():
+            # only set if we haven't seen this repository before
             if name not in self.repositories:
                 toscaRepository = resolver.get_repository(name, tpl)
                 self.repositories[name] = RepoView(toscaRepository, None)
@@ -425,15 +482,18 @@ class Manifest(AttributeManager):
                     )
                     continue
                 if isinstance(repository, dict):
-                    repository = Repository(name, repository)
+                    repository = resolver.get_repository(name, repository)
                 self.add_repository(None, repository, "")
         return self._set_repositories()
 
     @staticmethod
     def _get_repositories(tpl):
-        return ((tpl.get("spec") or {}).get("service_template") or {}).get(
+        repositories = ((tpl.get("spec") or {}).get("service_template") or {}).get(
             "repositories"
         ) or {}
+        # these take priority:
+        repositories.update((tpl.get("environment") or {}).get("repositories") or {})
+        return repositories
 
     def _set_repositories(self):
         repositories = self.repositories
@@ -498,15 +558,17 @@ class Manifest(AttributeManager):
             repo = artifactTpl.get("repository")
             if isinstance(repo, dict):
                 # a full repository spec maybe part of the include
-                reponame = repo.setdefault("name", os.path.basename(path))
+                reponame = repo.get("name", os.path.basename(path))
                 # replace spec with just its name
                 artifactTpl["repository"] = reponame
-
-                inlineRepository = {reponame: repo}
+                assert repo.get("url"), f"bad inline repository definition: {repo}"
+                inlineRepository = {reponame: dict(url=repo['url'])}
         else:
             artifactTpl = dict(file=templatePath)
 
         if check:
+            if isinstance(check, str):
+                return check
             if not inlineRepository and "repository" in artifactTpl:
                 reponame = artifactTpl["repository"]
                 if reponame in ["spec", "self", "unfurl"]:  # builtin
@@ -516,11 +578,9 @@ class Manifest(AttributeManager):
             return True
 
         artifact = ArtifactSpec(artifactTpl, path=baseDir)
-
         tpl = CommentedMap()
-
-        resolver = self.get_import_resolver(warnWhenNotFound)
-        tpl["repositories"] = self.update_repositories(
+        resolver = self.get_import_resolver(warnWhenNotFound, config=expanded)
+        tpl["repositories"] = self._update_repositories(
             expanded or yamlConfig.config, inlineRepository, resolver
         )
         loader = toscaparser.imports.ImportsLoader(
@@ -529,10 +589,15 @@ class Manifest(AttributeManager):
             tpl=tpl,
             resolver=resolver,
         )
-        return loader._load_import_template(None, artifact.as_import_spec())
+        path, doc = loader._load_import_template(None, artifact.as_import_spec())
+        if doc is None:
+            logger.warning(
+                f"document include {templatePath} does not exist (base: {baseDir})"
+            )
+        return path, doc
 
-    def get_import_resolver(self, ignoreFileNotFound=False, expand=False):
-        return ImportResolver(self, ignoreFileNotFound, expand)
+    def get_import_resolver(self, ignoreFileNotFound=False, expand=False, config=None):
+        return ImportResolver(self, ignoreFileNotFound, expand, config)
 
 
 class SnapShotManifest(Manifest):
@@ -545,7 +610,7 @@ class SnapShotManifest(Manifest):
         self.manifest = YamlConfig(
             oldManifest, manifest.path, loadHook=self.load_yaml_include
         )
-        self.update_repositories(oldManifest)
+        self._update_repositories(oldManifest)
         expanded = self.manifest.expanded
         # just needs the spec, not root resource
         self._set_spec(expanded.get("spec", {}))
