@@ -17,7 +17,7 @@ def _mapCtor(self):
     return CommentedMap
 
 
-CommentedMap.mapCtor = property(_mapCtor)
+CommentedMap.mapCtor = property(_mapCtor)  # type: ignore
 
 # we can't subclass CommentedMap so we need to monkey patch
 __base__deepcopy__ = CommentedMap.__deepcopy__
@@ -31,7 +31,7 @@ def __deepcopy__(self, memo):
 
 
 if CommentedMap.__deepcopy__ is not __deepcopy__:
-    CommentedMap.__deepcopy__ = __deepcopy__
+    CommentedMap.__deepcopy__ = __deepcopy__  # type: ignore
 
 
 def copy(src):
@@ -66,7 +66,7 @@ def make_map_with_base(doc, baseDir):
 
 # XXX?? because json keys are strings allow number keys to merge with lists
 # other values besides delete not supported because current code can leave those keys in final result
-mergeStrategyKey = "+%"  # supported values: "whiteout", "nullout"
+mergeStrategyKey = "+%"  # supported values: "whiteout", "nullout", "merge", "error"
 
 # b is the merge patch, a is original dict
 def merge_dicts(
@@ -83,7 +83,8 @@ def merge_dicts(
 
     Similar to https://yaml.org/type/merge.html but does a recursive merge
     """
-    cls = getattr(b, "mapCtor", cls or b.__class__)
+    if cls is not dict:
+        cls = getattr(b, "mapCtor", cls or b.__class__)
     cp = cls()
     skip = []
     for key, val in a.items():
@@ -108,6 +109,7 @@ def merge_dicts(
                         cp[key] = merge_dicts(
                             bval,
                             val,
+                            cls,
                             defaultStrategy=childStrategy,
                             replaceKeys=replaceKeys,
                             listStrategy=listStrategy,
@@ -128,7 +130,30 @@ def merge_dicts(
             if isinstance(bval, MutableSequence) and listStrategy == "append_unique":
                 # XXX allow more strategies beyond append
                 #     if appendlists == 'all' or key in appendlists:
-                cp[key] = bval + [item for item in val if item not in bval]
+                bval = bval[:]
+                for i, item in enumerate(val):
+                    if (
+                        isinstance(item, Mapping)
+                        and item.get(mergeStrategyKey) == "merge"
+                    ):
+                        if i >= len(bval):
+                            bval.append(item)
+                        else:
+                            bitem = bval[i]
+                            if isinstance(bitem, Mapping):
+                                bval[i] = merge_dicts(
+                                    bitem,
+                                    item,
+                                    cls,
+                                    defaultStrategy=childStrategy,
+                                    replaceKeys=replaceKeys,
+                                    listStrategy=listStrategy,
+                                )
+                            else:
+                                bval[i] = item
+                    elif item not in bval:
+                        bval.append(item)
+                cp[key] = bval
                 continue
         #     elif mergelists == 'all' or key in mergelists:
         #       newlist = []
@@ -204,7 +229,7 @@ def get_template(doc, key, value, path, cls, includes=None):
     template = doc
     templatePath = None
     if key.include:
-        value, template, baseDir = doc.loadTemplate(value, key.maybe, doc)
+        value, template, baseDir = doc.loadTemplate(value, key.maybe, doc, path)
         if template is None:  # include wasn't not found and key.maybe
             return None
         cls = make_map_with_base(doc, baseDir)
@@ -292,6 +317,9 @@ def _find_template(doc, key, path, cls, fail):
 
 def has_template(doc, key, value, path, cls):
     if key.include:
+        if key.maybe:
+            # treat missing includes as null template instead of an error
+            return True
         loadTemplate = getattr(doc, "loadTemplate", None)
         if not loadTemplate:
             return False
@@ -432,6 +460,7 @@ def expand_doc(doc, current=None, cls=dict):
     if not isinstance(doc, Mapping) or not isinstance(current, Mapping):
         raise UnfurlError(f"top level element {doc} is not a dict")
     expanded = expand_dict(doc, (), includes, current, cls)
+    assert isinstance(expanded, Mapping), expanded
     if hasattr(doc, "_anchorCache"):
         expanded._anchorCache = doc._anchorCache
     last = 0
