@@ -1,6 +1,8 @@
 # Copyright (c) 2020 Adam Souzis
 # SPDX-License-Identifier: MIT
+from typing import Optional
 import six
+
 from .runtime import NodeInstance
 from .util import UnfurlError
 from .result import ChangeRecord
@@ -16,16 +18,26 @@ from .planrequests import (
     find_parent_resource,
     find_resources_from_template_name,
 )
-from .tosca import find_standard_interface
-
-import logging
-
-logger = logging.getLogger("unfurl")
+from .tosca import find_standard_interface, EntitySpec
+from .logs import getLogger
 
 
-def is_external_template_compatible(external, template):
+logger = getLogger("unfurl")
+
+
+def is_external_template_compatible(
+    import_name: str, external: EntitySpec, template: EntitySpec
+):
     # for now, require template names to match
-    if external.name == template.name:
+    imported = external.tpl.get("imported")
+    if imported:
+        i_name, sep, t_name = imported.partition(":")
+        if i_name != import_name:
+            return False
+    else:
+        t_name = template.name
+
+    if external.name == t_name:
         if not external.is_compatible_type(template.type):
             raise UnfurlError(
                 f'external template "{template.name}" not compatible with local template'
@@ -56,13 +68,14 @@ class Plan:
             discover=ReadOnlyPlan,
         ).get(workflow, WorkflowPlan)
 
-    interface = None
+    interface: Optional[str] = None
 
     def __init__(self, root, toscaSpec, jobOptions):
         self.jobOptions = jobOptions
         self.workflow = jobOptions.workflow
         self.root = root
         self.tosca = toscaSpec
+        self._checked_connection_task = False
         assert self.tosca
         if jobOptions.template:
             filterTemplate = self.tosca.get_template(jobOptions.template)
@@ -75,50 +88,58 @@ class Plan:
             self.filterTemplate = None
 
     def find_shadow_instance(self, template, match=is_external_template_compatible):
+        imported = template.tpl.get("imported")
+        if imported:
+            external = self.root.imports.find_import(imported)
+            if not external:
+                return None
+            if external.shadow and external.root is self.root:
+                # shadowed instance already created
+                return external
+            else:
+                import_name = imported.partition(":")[0]
+                return self.create_shadow_instance(external, import_name, template)
+
         searchAll = []
-        for name, value in self.root.imports.items():
-            external = value.resource
+        for name, record in self.root.imports.items():
+            external = record.external_instance
             # XXX if external is a Relationship and template isn't, get it's target template
             #  if no target, create with status == unknown
 
-            if match(external.template, template):
-                if external.shadow and external.root is self.root:
-                    # shadowed instance already created
-                    return external
+            if match(name, external.template, template):
+                if record.local_instance:
+                    return record.local_instance
                 else:
-                    return self.create_shadow_instance(external, name)
-            if value.spec.get("instance") == "*":
-                searchAll.append((name, value.resource))
+                    return self.create_shadow_instance(external, name, template)
+            if record.spec.get("instance") in ["root", "*"]:
+                # add root instance
+                searchAll.append((name, record.external_instance))
 
         # look in the topologies where were are importing everything
         for name, root in searchAll:
             for external in root.get_self_and_descendents():
-                if match(external.template, template):
-                    return self.create_shadow_instance(external, name)
+                if match(name, external.template, template):
+                    return self.create_shadow_instance(external, name, template)
 
         return None
 
-    def create_shadow_instance(self, external, importName):
-        if self.root.imports[importName].resource is external:
-            name = importName
-        else:
-            name = importName + ":" + external.name
-
-        if external.parent and external.parent.parent:
-            # assumes one-to-one correspondence instance and template
-            parent = self.find_shadow_instance(external.parent.template)
-            if not parent:  # parent wasn't in imports, add it now
-                parent = self.create_shadow_instance(external.parent, importName)
-        else:
-            parent = self.root
-
-        shadowInstance = external.__class__(
-            name, external.attributes, parent, external.template, external
+    def create_shadow_instance(self, external, import_name, template):
+        # create a local instance that "shadows" the external one we imported
+        name = import_name + ":" + external.name
+        instance_name = template.name
+        # XXX import the parent too by creating a template and setting its name to name?
+        parent = self.root
+        logger.debug(
+            "creating local instance %s for external instance %s",
+            instance_name or name,
+            external.name,
         )
-
-        shadowInstance.shadow = external
-        # Imports.__setitem__ will add or update:
-        self.root.imports[name] = shadowInstance
+        # attributes can be empty because AttributeManager delegates to the imported shadow instance
+        shadowInstance = external.__class__(
+            instance_name, {}, parent, template, external.status
+        )
+        shadowInstance.imported = name
+        self.root.imports.set_shadow(name, shadowInstance, external)
         return shadowInstance
 
     def find_resources_from_template(self, template):
@@ -126,9 +147,10 @@ class Plan:
             # XXX also match node_filter if present
             shadowInstance = self.find_shadow_instance(template)
             if shadowInstance:
+                logger.debug('found imported instance "%s"', shadowInstance.name)
                 yield shadowInstance
             else:
-                logger.info(
+                logger.warning(
                     "could not find external instance for template %s", template.name
                 )
             # XXX also yield newly created parents that needed to be checked?
@@ -224,14 +246,25 @@ class Plan:
             or self.jobOptions.force
             or (resource.status == Status.error and initialState)
         ):
-            yield from self._run_operation(
-                NodeState.creating, "Standard.create", resource, reason, inputs
+            req = create_task_request(
+                self.jobOptions,
+                "Standard.create",
+                resource,
+                reason,
+                inputs,
+                NodeState.creating,
             )
+            if req:
+                yield req
+            else:
+                # no create operation defined, run configure instead
+                initialState = True
 
         if (
             initialState
             or resource.state < NodeState.configured
             or (self.jobOptions.force and resource.state != NodeState.started)
+            or resource.status == Status.error
         ):
             yield from self._execute_default_configure(resource, reason, inputs)
 
@@ -260,7 +293,7 @@ class Plan:
             resource.state in [NodeState.starting, NodeState.started]
             or self.workflow == "stop"
         ):
-            nodeState = NodeState.stopping
+            nodeState: Optional[NodeState] = NodeState.stopping
             op = "Standard.stop"
 
             yield from self._run_operation(nodeState, op, resource, reason, inputs)
@@ -298,6 +331,8 @@ class Plan:
                 skip = "creation and deletion is managed by another instance"
             elif "protected" in resource.template.directives:
                 skip = 'instance with "protected" directive'
+            elif resource.protected:
+                skip = "protected instance"
             elif "virtual" in resource.template.directives:
                 skip = 'instance with "virtual" directive'
             elif resource.status in [Status.absent, Status.pending]:
@@ -323,10 +358,20 @@ class Plan:
             return self.execute_default_deploy(resource, reason, inputs)
         elif workflow == "undeploy" or workflow == "stop":
             return self.execute_default_undeploy(resource, reason, inputs)
-        elif workflow == "check" or workflow == "discover":
+        elif workflow == "check" or workflow == "discover" or workflow == "connect":
             return self.execute_default_install_op(workflow, resource, reason, inputs)
         return None
 
+    def _get_connection_task(self, taskRequest):
+        if not self._checked_connection_task:
+            self._checked_connection_task = True
+            if self.tosca.topology.primary_provider:
+                for rel in self.root.requirements:
+                    if rel.name == "primary_provider":
+                        yield from self.execute_default_install_op("check", rel)
+                        return
+                assert False, self.root.requirements
+
     def _generate_configurations(self, resource, reason, workflow=None):
         workflow = workflow or self.workflow
         # check if this workflow has been delegated to one explicitly declared
@@ -341,12 +386,19 @@ class Plan:
             group = TaskRequestGroup(resource, workflow)
         else:
             group = None
+        task_found = False
         for taskRequest in configGenerator:
             if taskRequest:
+                task_found = True
+                yield from self._get_connection_task(taskRequest)
                 if group and not isinstance(taskRequest, JobRequest):
                     group.children.append(taskRequest)
                 else:
                     yield taskRequest
+        if not task_found:
+            logger.verbose(
+                f'No operations for workflow "{workflow}" defined for instance "{resource.name}"'
+            )
         if group:
             yield group
 
@@ -467,7 +519,9 @@ class Plan:
         opts = self.jobOptions
         templates = self._get_templates()
 
-        logger.verbose("checking for tasks for templates %s", [t.name for t in templates])
+        logger.verbose(
+            "checking for tasks for templates %s", [t.name for t in templates]
+        )
         visited = set()
         for template in templates:
             found = False
@@ -581,17 +635,24 @@ class DeployPlan(Plan):
 
     def _generate_workflow_configurations(self, instance, oldTemplate):
         # if oldTemplate is not None this is an existing instance, so check if we should include
-        if oldTemplate:
+        if instance.shadow:
+            reason = "connect"
+        elif oldTemplate:
             reason = self.include_instance(oldTemplate, instance)
             if not reason:
                 logger.debug(
-                    "not including task for %s:%s", instance.name, oldTemplate.name
+                    "not including task for %s:%s (status: %s)",
+                    instance.name,
+                    oldTemplate.name,
+                    instance.status,
                 )
                 return
         else:  # this is newly created resource
             reason = Reason.add
 
-        if instance.status == Status.unknown or instance.shadow:
+        if instance.shadow:
+            installOp = "connect"
+        elif instance.status == Status.unknown:
             installOp = "check"
         elif "discover" in instance.template.directives and not instance.operational:
             installOp = "discover"
@@ -606,12 +667,18 @@ class DeployPlan(Plan):
 
         if reason == Reason.reconfigure:
             # XXX generate configurations: may need to stop, start, etc.
+            if "discover" in instance.template.directives:
+                op = "Install.discover"
+                startState = None
+            else:
+                op = "Standard.configure"
+                startState = NodeState.configuring
             req = create_task_request(
                 self.jobOptions,
-                "Standard.configure",
+                op,
                 instance,
                 reason,
-                startState=NodeState.configuring,
+                startState=startState,
             )
             if req:
                 yield req
