@@ -2,26 +2,30 @@
 # SPDX-License-Identifier: MIT
 import collections
 import re
+from typing import Iterator, List, Optional, Sequence, Tuple
 import six
 import shlex
 import sys
 import os
 import os.path
+
 from .util import (
     lookup_class,
     load_module,
     find_schema_errors,
     UnfurlError,
     UnfurlTaskError,
-    to_enum
+    to_enum,
 )
 from .result import serialize_value
-from .support import Defaults, NodeState
+from .support import Defaults, NodeState, Priority
+from .runtime import EntityInstance
 import logging
 
 logger = logging.getLogger("unfurl")
 
-# we want ConfigurationSpec to be standalone and easily serializable
+
+# we want ConfigurationSpec to be independent of our object model and easily serializable
 class ConfigurationSpec:
     @classmethod
     def getDefaults(cls):
@@ -42,6 +46,7 @@ class ConfigurationSpec:
             outputs=None,
             interface=None,
             entry_state=None,
+            base_dir=None,
         )
 
     def __init__(
@@ -64,6 +69,7 @@ class ConfigurationSpec:
         outputs=None,
         interface=None,
         entry_state=None,
+        base_dir=None,
     ):
         assert name and className, "missing required arguments"
         self.name = name
@@ -84,6 +90,7 @@ class ConfigurationSpec:
         self.dependencies = dependencies
         self.interface = interface
         self.entry_state = to_enum(NodeState, entry_state, NodeState.creating)
+        self.base_dir = base_dir
 
     def find_invalidate_inputs(self, inputs):
         if not self.inputSchema:
@@ -137,10 +144,11 @@ class ConfigurationSpec:
 
 class PlanRequest:
     error = None
-    future_dependencies = ()
+    future_dependencies: list = []
     task = None
+    render_errors = None
 
-    def __init__(self, target):
+    def __init__(self, target: "EntityInstance"):
         self.target = target
 
     @property
@@ -153,6 +161,52 @@ class PlanRequest:
     def get_operation_artifacts(self):
         return []
 
+    def include_in_plan(self):
+        if self.task and self.task.priority == Priority.critical:
+            return True  # XXX hackish, just used for primary_provider
+        if self.target.priority is not None:
+            return self.target.priority >= Priority.required
+        if self.target.created:
+            #  if already created then always include the resource
+            return True
+        return self.target.template.required
+
+    def has_unfulfilled_refs(self) -> bool:
+        for dep in self.get_unfulfilled_refs():
+            return True
+        return False
+
+    def get_unfulfilled_refs(self):
+        if not self.render_errors:
+            return
+        for error in self.render_errors:
+            dep = getattr(error, "dependency", None)
+            if dep:
+                yield dep
+
+    def update_unfulfilled_errors(self):
+        self.render_errors = list(self._get_unfulfilled_errors())
+        return self.has_unfulfilled_refs()
+
+    def _get_unfulfilled_errors(self):
+        if not self.render_errors:
+            return
+        for error in self.render_errors:
+            dep = getattr(error, "dependency", None)
+            if not dep or not dep.validate():
+                yield error
+
+    @property
+    def not_ready(self):
+        return self.future_dependencies or self.has_unfulfilled_refs()
+
+    @property
+    def name(self):
+        return type(self).__name__
+
+    def __str__(self) -> str:
+        return f'Planned task {self.name} for "{self.target.name}"'
+
 
 class TaskRequest(PlanRequest):
     """
@@ -315,6 +369,21 @@ class TaskRequestGroup(PlanRequest):
             artifacts.extend(req.get_operation_artifacts())
         return artifacts
 
+    def get_unfulfilled_refs(self):
+        for req in self.children:
+            yield from req.get_unfulfilled_refs()
+
+    def update_unfulfilled_errors(self):
+        for req in self.children:
+            req.update_unfulfilled_errors()
+        return self.has_unfulfilled_refs()
+
+    def __str__(self) -> str:
+        if self.children:
+            return str(self.children)
+        else:
+            return f'Planned task {self.workflow} for "{self.target.name}"'
+
     def __repr__(self):
         return f"TaskRequestGroup({self.target}:{self.workflow}:{self.children})"
 
@@ -385,7 +454,9 @@ def find_operation_host(target, operation_host):
     return target.root.find_instance_or_external(operation_host)
 
 
-def get_render_requests(requests):
+def get_render_requests(
+    requests: Sequence[PlanRequest],
+) -> Iterator[Tuple[Optional[TaskRequestGroup], TaskRequest]]:
     # returns requests that can be rendered grouped by its top-most task group
     for req in requests:
         if isinstance(req, TaskRequestGroup):
@@ -394,6 +465,7 @@ def get_render_requests(requests):
         elif isinstance(req, TaskRequest):
             yield None, req
         elif not isinstance(req, SetStateRequest):
+            # note: SetStateRequests don't render, only run
             assert not req, f"unexpected type of request: {req}"
 
 
@@ -419,14 +491,20 @@ def set_fulfilled(requests, completed):
     # as is future_dependencies
     ready, notReady = [], []
     for req in requests:
+        _not_ready = False
+        # always need to evaluate both:
         if req.update_future_dependencies(completed):
+            _not_ready = True
+        if req.update_unfulfilled_errors():
+            _not_ready = True
+        if _not_ready:
             notReady.append(req)
         else:  # list is now empty so request is ready
             ready.append(req)
     return ready, notReady
 
 
-def _prepare_request(job, req, errors):
+def _prepare_request(job, req: TaskRequest, errors):
     # req is a taskrequests, future_requests are (grouprequest, taskrequest) pairs
     if req.task:
         task = req.task
@@ -449,6 +527,8 @@ def _prepare_request(job, req, errors):
                 req.target.status,
                 msg,
             )
+            # we want to track and save skipped tasks
+            job.add_work(task)
     except Exception:
         proceed = False
         # note: failed rendering may be re-tried later if it has dependencies
@@ -474,14 +554,18 @@ def _render_request(job, parent, req, requests):
     try:
         task.logger.debug("rendering %s %s", task.target.name, task.name)
         task._rendering = True
-        if task._inputs:
-            # turn strictness off so we can detect dependencies
-            task._inputs.context.strict = False
+        task.inputs
+        assert not task._inputs.context.strict
         task.rendered = task.configurator.render(task)
     except Exception:
         # note: failed rendering may be re-tried later if it has dependencies
         error_info = sys.exc_info()
         error = UnfurlTaskError(task, "Configurator render failed", False)
+    if task._errors:
+        # we turned off strictness so templating errors got saved here instead
+        req.render_errors = task._errors
+        error = task._errors[0]
+        task._errors = []
     task._rendering = False
     task._attributeManager.mark_referenced_templates(task.target.template)
 
@@ -491,32 +575,36 @@ def _render_request(job, parent, req, requests):
     else:
         # key => (instance, list<attribute>)
         liveDependencies = task._attributeManager.find_live_dependencies()
-        logger.trace(f"liveDependencies for {task.target}: {liveDependencies}")
+        task.logger.trace(f"liveDependencies for {task.target}: {liveDependencies}")
         # a future request may change the value of these attributes
         deps = list(_get_deps(parent, req, liveDependencies, requests))
 
-    if deps:
+    dependent_refs = [dep.expr for dep in req.get_unfulfilled_refs()]
+    if deps or dependent_refs:
         req.future_dependencies = deps
         task.logger.debug(
             "%s:%s can not render yet, depends on %s",
             task.target.name,
             req.configSpec.operation,
-            str(deps), exc_info=error_info
+            deps or dependent_refs,
+            exc_info=error_info,
         )
         # rollback changes:
         task._errors = []
         task._inputs = None
         task._attributeManager.attributes = {}
         task.discard_work_folders()
-        return deps, None
+        return deps or dependent_refs, None
     elif error:
         task.fail_work_folders()
         task._inputs = None
-        task.logger.warning("Configurator render failed",  exc_info=error_info)
+        task.logger.warning("Configurator render failed", exc_info=error_info)
         task._attributeManager.attributes = {}  # rollback changes
+        return None, error
     else:
+        task.logger.trace(f"committing changes from rendering task {task.target}")
         task.commit_changes()
-    return deps, error
+        return None, None
 
 
 def _add_to_req_list(reqs, parent, request):
@@ -531,15 +619,20 @@ def _reevaluate_not_required(not_required, render_requests):
     # keep rendering if a not_required template was referenced and is now required
     new_not_required = []
     for (parent, request) in not_required:
-        if request.target.template.required:
-            render_requests.append( (parent, request) )
+        if request.include_in_plan():
+            request.target.validate()
+            render_requests.append((parent, request))
         else:
-            new_not_required.append( (parent, request) )
+            new_not_required.append((parent, request))
     return new_not_required
 
 
-def do_render_requests(job, requests):
-    ready, notReady, errors = [], [], []
+def do_render_requests(
+    job, requests: Sequence[PlanRequest]
+) -> Tuple[List[PlanRequest], List[PlanRequest], List[UnfurlError]]:
+    ready: List[PlanRequest] = []
+    notReady: List[PlanRequest] = []
+    errors: List[UnfurlError] = []
     flattened_requests = list(
         (p, r)
         for (p, r) in get_render_requests(requests)
@@ -551,14 +644,17 @@ def do_render_requests(job, requests):
         parent, request = render_requests.popleft()
         # we dont require default templates that aren't referenced
         # (but skip this check if the job already specified specific instances)
-        required = job.workflow != "deploy" or job.is_filtered() or request.target.template.required
+        required = (
+            job.workflow != "deploy" or job.is_filtered() or request.include_in_plan()
+        )
         if not required:
-            not_required.append( (parent, request) )
+            not_required.append((parent, request))
         else:
+            request.target.validate()
             deps, error = _render_request(job, parent, request, flattened_requests)
             if error:
                 errors.append(error)
-            if deps:
+            elif deps:
                 # remove parent from ready if added it there
                 if parent and ready and ready[-1] is parent:
                     ready.pop()
@@ -624,7 +720,7 @@ def find_parent_template(source):
         # special case "host" so it can be declared without full set of relationship / capability types
         if rel.is_derived_from("tosca.relationships.HostedOn") or "host" in req:
             return rel.target
-        return None
+    return None
 
 
 def find_parent_resource(root, source):
@@ -634,7 +730,9 @@ def find_parent_resource(root, source):
     for parent in find_resources_from_template_name(root, parentTemplate.name):
         # XXX need to evaluate matches
         return parent
-    raise UnfurlError(f"could not find instance of template: {parentTemplate.name}")
+    raise UnfurlError(
+        f"could not find parent instance {parentTemplate.name} for child {source.name}"
+    )
 
 
 def create_instance_from_spec(_manifest, target, rname, resourceSpec):
@@ -658,13 +756,19 @@ def create_instance_from_spec(_manifest, target, rname, resourceSpec):
         resourceSpec["created"] = target.key
 
     if "parent" not in resourceSpec and "template" in resourceSpec:
-        nodeSpec = _manifest.tosca.get_template(resourceSpec["template"])
+        tname = resourceSpec['template']
+        nodeSpec = _manifest.tosca.get_template(tname)
+        if not nodeSpec:
+            raise UnfurlError(
+                f'Can not find template "{tname}" when trying to create instance "{rname}".'
+            )
         parent = find_parent_resource(target.root, nodeSpec)
     else:
         parent = target.root
     # note: if resourceSpec[parent] is set it overrides the parent keyword
     return _manifest.create_node_instance(rname, resourceSpec, parent=parent)
 
+
 def _maybe_mock(iDef, template):
     if not os.getenv("UNFURL_MOCK_DEPLOY"):
         return iDef
@@ -675,9 +779,10 @@ def _maybe_mock(iDef, template):
     if not isinstance(iDef.implementation, dict):
         # it's a string naming an artifact
         iDef.implementation = dict(primary=iDef.implementation)
-    iDef.implementation['className'] = "unfurl.configurator.MockConfigurator"
+    iDef.implementation["className"] = "unfurl.configurator.MockConfigurator"
     return iDef
 
+
 def create_task_request(
     jobOptions,
     operation,
@@ -699,6 +804,12 @@ def create_task_request(
             inputs = dict(iDef.inputs, **inputs)
         else:
             inputs = iDef.inputs or {}
+        if iDef.invoke:
+            # get the implementation from the operation specified with the "invoke" key
+            iinterface, sep, iaction = iDef.invoke.rpartition(".")
+            iDef = _find_implementation(iinterface, iaction, resource.template)
+            inputs = dict(iDef.inputs, **inputs)
+
         kw = _get_config_spec_args_from_implementation(
             iDef, inputs, resource, operation_host
         )
@@ -782,7 +893,9 @@ def _set_config_spec_args(kw, target, base_dir):
     artifact = kw["primary"]
     className = kw.get("className")
     if not className and not artifact:  # malformed implementation
-        logger.warning("no artifact or className set on operation for %s: %s", target.name, kw)
+        logger.warning(
+            "no artifact or className set on operation for %s: %s", target.name, kw
+        )
         return None
     guessing = False
     if not className:
@@ -820,7 +933,12 @@ def _set_config_spec_args(kw, target, base_dir):
 
 def _get_config_spec_args_from_implementation(iDef, inputs, target, operation_host):
     implementation = iDef.implementation
-    kw = dict(inputs=inputs, outputs=iDef.outputs, operation_host=operation_host, entry_state=iDef.entry_state)
+    kw = dict(
+        inputs=inputs,
+        outputs=iDef.outputs,
+        operation_host=operation_host,
+        entry_state=iDef.entry_state,
+    )
     configSpecArgs = ConfigurationSpec.getDefaults()
     artifactTpl = None
     dependencies = None
@@ -847,6 +965,7 @@ def _get_config_spec_args_from_implementation(iDef, inputs, target, operation_ho
     # if not operation_instance:
     #     operation_instance = operation_instance or target.root
     base_dir = getattr(iDef.value, "base_dir", iDef._source)
+    kw["base_dir"] = base_dir
     if artifactTpl:
         artifact = target.template.find_or_create_artifact(
             artifactTpl, base_dir, predefined
