@@ -5,19 +5,34 @@ import os.path
 from pathlib import Path
 import sys
 import git
-import logging
-from six.moves.urllib.parse import urlparse
+from .logs import getLogger
+from urllib.parse import urlparse
 from .util import UnfurlError, save_to_file
 import toscaparser.repositories
 from ruamel.yaml.comments import CommentedMap
+import logging
+
+
+logger = getLogger("unfurl")
 
-logger = logging.getLogger("unfurl")
 
 def is_git_worktree(path, gitDir=".git"):
     # NB: if work tree is a submodule .git will be a file that looks like "gitdir: ./relative/path"
     return os.path.exists(os.path.join(path, gitDir))
 
-def normalize_git_url(url):
+
+def _add_user_to_url(url, username, password):
+    assert username
+    parts = urlparse(url)
+    user, sep, host = parts.netloc.rpartition("@")
+    if password:
+        netloc = f"{username}:{password}@{host}"
+    else:
+        netloc = f"{username}@{host}"
+    return parts._replace(netloc=netloc).geturl()
+
+
+def normalize_git_url(url, hard=0):
     if url.startswith("git-local://"):  # truncate url after commit digest
         return "git-local://" + urlparse(url).netloc.partition(":")[0]
 
@@ -27,12 +42,25 @@ def normalize_git_url(url):
         elif "@" in url:  # scp style used by git: user@server:project.git
             # convert to ssh://user@server/project.git
             return "ssh://" + url.replace(":", "/", 1)
+    if hard:
+        parts = urlparse(url)
+        # remove password and .git
+        user, sep, host = parts.netloc.rpartition("@")
+        if sep and hard == 1:
+            netloc = f"{user.partition(':')[0]}@{host}"
+        else:
+            netloc = host
+        if hard == 2:
+            path = parts.path.rstrip(".git")
+        else:
+            path = parts.path
+        return parts._replace(netloc=netloc, path=path).geturl()
     return url
 
+
 def normalize_git_url_hard(url):
-    parts = urlparse(normalize_git_url(url))
-    # remove scheme, user, and .git
-    return parts.netloc.rpartition("@")[2] + parts.path.rstrip('.git')
+    # remove scheme, .git and fragment
+    return normalize_git_url(url, hard=2).rpartition("://")[2].partition("#")[0]
 
 
 def is_url_or_git_path(url):
@@ -87,11 +115,12 @@ class Repo:
         return None
 
     @staticmethod
-    def find_git_working_dirs(rootDir, gitDir=".git"):
+    def find_git_working_dirs(rootDir, include_root, gitDir=".git"):
         working_dirs = {}
         for root, dirs, files in os.walk(rootDir):
             if Repo.update_git_working_dirs(working_dirs, root, dirs, gitDir):
-                del dirs[:]  # don't visit sub directories
+                if not include_root or rootDir != root:
+                    del dirs[:]  # don't visit sub directories
         return working_dirs
 
     @staticmethod
@@ -249,7 +278,16 @@ class RepoView:
 
     @property
     def url(self):
-        return self.repository.url if self.repository else self.repo.url
+        if self.repository:
+            url = self.repository.url
+            if self.repository.credential:
+                credential = self.repository.credential
+                return _add_user_to_url(url, credential["user"], credential["token"])
+            else:
+                return url
+        else:
+            assert self.repo
+            return self.repo.url
 
     def is_local_only(self):
         # if it doesn't have a repo then it most be local
@@ -364,7 +402,7 @@ class RepoView:
     def lock(self):
         record = CommentedMap(
             [
-                ("url", self.url),
+                ("url", normalize_git_url(self.url, 1)),
                 ("revision", self.get_current_revision()),
                 ("initial", self.get_initial_revision()),
             ]
@@ -408,12 +446,12 @@ class GitRepo(Repo):
         except:
             return None
 
-    def get_url_with_path(self, path):
+    def get_url_with_path(self, path, sanitize=False):
         if is_url_or_git_path(self.url):
             if os.path.isabs(path):
                 # get path relative to repository's root
                 path = os.path.relpath(path, self.working_dir)
-            return normalize_git_url(self.url) + "#:" + path
+            return normalize_git_url(self.url, sanitize) + "#:" + path
         else:
             return self.get_git_local_url(path)
 
@@ -522,9 +560,21 @@ class GitRepo(Repo):
                 cmd.append("--ff-only")
             code, out, err = self.run_cmd(cmd)
             if code:
-                logger.info("attempt to pull latest from %s into %s failed: %s %s", self.url, self.working_dir, out, err)
+                logger.info(
+                    "attempt to pull latest from %s into %s failed: %s %s",
+                    self.url,
+                    self.working_dir,
+                    out,
+                    err,
+                )
             else:
-                logger.verbose("pull latest from %s into %s: %s %s", self.url, self.working_dir, out, err)
+                logger.verbose(
+                    "pull latest from %s into %s: %s %s",
+                    self.url,
+                    self.working_dir,
+                    out,
+                    err,
+                )
             return not code
         else:
             return False
