@@ -3,21 +3,27 @@
 from collections.abc import Mapping, MutableSequence, MutableMapping
 from abc import ABC, abstractmethod
 from datetime import datetime, timedelta
-from typing import Match, Optional
+import logging
+from typing import TYPE_CHECKING, Any, Match, Optional
 import six
 import hashlib
 import re
+from toscaparser.common.exception import ValidationError
 
 from .merge import diff_dicts
 from .util import (
     UnfurlError,
+    UnfurlTaskError,
     is_sensitive,
     sensitive_dict,
     sensitive_list,
     dump,
+    wrap_sensitive_value,
 )
 from .logs import sensitive
 
+logger = logging.getLogger("unfurl")
+
 
 def _get_digest(value, kw):
     getter = getattr(value, "__digestable__", None)
@@ -75,7 +81,10 @@ def serialize_value(value, **kw):
 
 class ResourceRef(ABC):
 
-    parent = None  # parent must be defined by subclass
+    parent = None  # must be defined by subclass
+    # note: really an EntitySpec but would cause cicular imports
+    template: Optional["ResourceRef"] = None
+    base_dir = ""
 
     @abstractmethod
     def _resolve(self, key):
@@ -128,6 +137,10 @@ class ResourceRef(ABC):
     def templar(self):
         return self.root._templar
 
+    @property
+    def readonly(self) -> bool:
+        return False
+
 
 class ChangeRecord:
     """
@@ -145,12 +158,12 @@ class ChangeRecord:
     DateTimeFormat = "%Y-%m-%d-%H-%M-%S-%f"
 
     def __init__(
-        self, 
-        jobId: Optional[str]=None,
-        startTime: Optional[datetime]=None,
-        taskId: int=0,
-        previousId: Optional[str]=None, 
-        parse: Optional[str]=None
+        self,
+        jobId: Optional[str] = None,
+        startTime: Optional[datetime] = None,
+        taskId: int = 0,
+        previousId: Optional[str] = None,
+        parse: Optional[str] = None,
     ):
         if parse:
             self.parse(parse)
@@ -164,7 +177,7 @@ class ChangeRecord:
             else:
                 self.changeId = self.make_change_id(self.startTime, taskId, previousId)
 
-    def set_start_time(self, startTime: Optional[datetime]=None) -> None:
+    def set_start_time(self, startTime: Optional[datetime] = None) -> None:
         if not startTime:
             self.startTime = datetime.utcnow()
         elif isinstance(startTime, datetime):
@@ -214,7 +227,12 @@ class ChangeRecord:
         return re.match("^A[A-Za-z0-9]{11}$", test)
 
     @classmethod
-    def make_change_id(cls, timestamp: Optional[datetime]=None, taskid: int=0, previousId: Optional[str]=None) -> str:
+    def make_change_id(
+        cls,
+        timestamp: Optional[datetime] = None,
+        taskid: int = 0,
+        previousId: Optional[str] = None,
+    ) -> str:
         b62 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
         if not timestamp:
             timestamp = datetime.utcnow()
@@ -264,7 +282,7 @@ class ChangeRecord:
         terms = [changeId] + ["{}={}".format(k, v) for k, v in attributes.items()]
         return "\t".join(terms) + "\n"
 
-    def log(self, attributes: Optional[dict]=None) -> str:
+    def log(self, attributes: Optional[dict] = None) -> str:
         r"changeid\tkey=value\tkey=value"
         default = {
             k: getattr(self, k)
@@ -287,11 +305,11 @@ class ChangeAware:
 class ExternalValue(ChangeAware):
     __slots__ = ("type", "key")
 
-    def __init__(self, type, key):
+    def __init__(self, type: str, key: Any):
         self.type = type
         self.key = key
 
-    def get(self):
+    def get(self) -> Any:
         return self.key
 
     def __digestable__(self, options):
@@ -329,13 +347,13 @@ class Result(ChangeAware):
     # in __getitem__ and __setitem__
     __slots__ = ("original", "resolved", "external", "select")
 
-    def __init__(self, resolved):
+    def __init__(self, resolved: Any):
         self.select = ()
         self.original = _Deleted  # assume this is new to start
         if isinstance(resolved, ExternalValue):
             self.resolved = resolved.get()
             assert not isinstance(self.resolved, Result), self.resolved
-            self.external = resolved
+            self.external: Optional[ExternalValue] = resolved
         else:
             assert not isinstance(resolved, Result), resolved
             self.resolved = resolved
@@ -448,6 +466,23 @@ class Result(ChangeAware):
         return "Result(%r, %r, %r)" % (self.resolved, self.external, self.select)
 
 
+def is_sensitive_schema(defs, key):
+    defSchema = (key in defs and defs[key].schema) or {}
+    defMeta = defSchema.get("metadata", {})
+    return defMeta.get("sensitive")
+
+
+def _validation_error(src, context, prop_def, msg):
+    from .eval import Ref
+    from .configurator import Dependency
+
+    if Ref.is_ref(src):
+        dep = Dependency(src, target=context.currentResource, schema=prop_def)
+    else:
+        dep = None
+    UnfurlTaskError(context.task, msg, dependency=dep)
+
+
 class Results(ABC):
     """
     Evaluating expressions are not guaranteed to be idempotent (consider quoting)
@@ -469,7 +504,7 @@ class Results(ABC):
     def resolve_all(self):
         ...
 
-    def __init__(self, serializedOriginal, resourceOrCxt):
+    def __init__(self, serializedOriginal, resourceOrCxt, validate=False, defs=None):
         from .eval import RefContext
 
         assert not isinstance(serializedOriginal, Results), serializedOriginal
@@ -487,6 +522,12 @@ class Results(ABC):
             ctx.base_dir = newBaseDir
             ctx.trace("found baseDir", newBaseDir, "old", oldBaseDir)
         self.context = ctx
+        resource = ctx.currentResource
+        self.validate = validate
+        if defs is None:
+            self.defs = resource.template and resource.template.propertyDefs or {}
+        else:
+            self.defs = defs
 
     def get_copy(self, key, default=None):
         # return a copy of value or default if not found
@@ -498,7 +539,7 @@ class Results(ABC):
             return default
 
     @staticmethod
-    def _map_value(val, context, applyTemplates=True):
+    def _map_value(val, context, applyTemplates=True, defs=None):
         "Recursively and lazily resolves any references in a value"
         from .eval import map_value, Ref
 
@@ -509,9 +550,13 @@ class Results(ABC):
         elif isinstance(val, sensitive):
             return val
         elif isinstance(val, Mapping):
-            return ResultsMap(val, context)
+            # already validated
+            # always explicitly set defs
+            return ResultsMap(val, context, False, defs or {})
         elif isinstance(val, list):
-            return ResultsList(val, context)
+            # already validated
+            # always explicitly set defs
+            return ResultsList(val, context, False, defs or {})
         else:
             # at this point, just evaluates templates in strings or returns val
             return map_value(val, context.copy(wantList="result"), applyTemplates)
@@ -536,29 +581,86 @@ class Results(ABC):
         val = self._attributes[key]
         if isinstance(val, Result):
             # already resolved
+            self.context.trace("Results.get: already resolved", key, val)
             assert not isinstance(val.resolved, Result), val
             return val
         else:
             if self.doFullResolve:
+                self.context.trace("Results.doFullResolve", key, val)
                 if isinstance(val, Results):
                     resolved = val
                 else:  # evaluate records that aren't Results
                     resolved = map_value(val, self.context, self.applyTemplates)
             else:
                 # lazily evaluate lists and dicts
-                self.context.trace("Results._mapValue", val)
-                resolved = self._map_value(val, self.context, self.applyTemplates)
+                self.context.trace("Results._mapValue", key, val)
+                defs = self.get_datatype_defs(key)
+                resolved = self._map_value(val, self.context, self.applyTemplates, defs)
             # will return a Result if val was an expression that was evaluated
             if isinstance(resolved, Result):
                 result = resolved
+                result.resolved = self._transform(key, result.resolved)
                 resolved = result.resolved
             else:
+                resolved = self._transform(key, resolved)
                 result = Result(resolved)
             result.original = _Get
+            if isinstance(resolved, MutableSequence) and resolved:
+                assert not isinstance(resolved[0], Result), resolved[0]
+
+            if self.validate:
+                self._validate(key, resolved, val)
+            if self.defs and is_sensitive_schema(self.defs, key):
+                result.resolved = wrap_sensitive_value(resolved)
+
             self._attributes[key] = result
             assert not isinstance(resolved, Result), val
             return result
 
+    def get_datatype_defs(self, key):
+        property = self.defs.get(key)
+        if property:
+            return property.entity.properties
+        return None
+
+    def _transform(self, key, value):
+        from .eval import map_value
+
+        property = self.defs.get(key)
+        if property:
+            transform = property.schema.metadata.get("transform")
+            if transform:
+                logger.debug(
+                    "running transform on %s.%s", self.context.currentResource.name, key
+                )
+                return map_value(transform, self.context.copy(vars=dict(value=value)))
+        return value
+
+    def _validate(self, key, value, src=None, propDef=None):
+        propDef = propDef or self.defs.get(key)
+        if not propDef:
+            return True
+        resource = self.context.currentResource
+        try:
+            if value is None:
+                # required attributes might be null depending on the state of the resource
+                if propDef.required and key not in resource.template.attributeDefs:
+                    msg = f'Property "{key}" on "{resource.template.name}" cannot be null.'
+                    raise ValidationError(message=msg)
+            else:
+                propDef._validate(value)
+            self.context.trace(f'Validated "{key}" on "{resource.name}')
+        except Exception as err:
+            msg = f'Validation failure while evaluating "{key}" on "{resource.name}": {err}'
+            if self.context.task:
+                _validation_error(src, self.context, propDef, msg)
+            elif self.context.strict:
+                raise UnfurlError(msg, True)
+            else:
+                self.context.trace(msg)
+            return False
+        return True
+
     def _haskey(self, key):
         # can't use "in" operator for lists
         try:
@@ -569,22 +671,39 @@ class Results(ABC):
 
     def __setitem__(self, key, value):
         assert not isinstance(value, Result), (key, value)
+        if self.context.currentResource.readonly:
+            raise UnfurlError(
+                "Attempting to set {key} on a readonly instance {self.context.currentResource}"
+            )
         if self._haskey(key):
             resolved = self[key]
-            if resolved != value:
-                # exisiting value changed
+            if resolved != value:  # the existing value changed
+                if self.validate:
+                    self._validate(key, value)
+                if self.defs and is_sensitive_schema(self.defs, key):
+                    value = wrap_sensitive_value(value)
+
                 result = self._attributes[key]
                 if result.original is _Get:
                     # we haven't saved the original value yet
                     result.original = result.resolved
                 result.resolved = value
         else:
+            if self.validate:
+                self._validate(key, value)
+            if self.defs and is_sensitive_schema(self.defs, key):
+                value = wrap_sensitive_value(value)
             self._attributes[key] = Result(value)
 
         # remove from deleted if it's there
         self._deleted.pop(key, None)
 
     def __delitem__(self, index):
+        if self.context.currentResource.readonly:
+            raise UnfurlError(
+                "Attempting to delete item {item} on a readonly instance {self.context.currentResource}"
+            )
+        self.context.currentResource.name
         val = self._attributes[index]
         self._deleted[index] = val
         del self._attributes[index]
