@@ -8,7 +8,16 @@ Each instance have a status; attributes that describe its state; and a TOSCA tem
 which describes its capabilities, relationships and available interfaces for configuring and interacting with it.
 """
 from datetime import datetime
-from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, TYPE_CHECKING, cast
+from typing import (
+    Any,
+    Dict,
+    Iterable,
+    List,
+    Optional,
+    Union,
+    TYPE_CHECKING,
+    cast,
+)
 import six
 from collections.abc import Mapping
 
@@ -19,17 +28,19 @@ from .result import ResourceRef, ChangeAware
 
 from .support import AttributeManager, Defaults, Status, Priority, NodeState, Templar
 from .tosca import (
+    EntitySpec,
     CapabilitySpec,
     RelationshipSpec,
     NodeSpec,
     TopologySpec,
     ArtifactSpec,
 )
-from unfurl.logs import UnfurlLogger
+from .logs import UnfurlLogger
 import logging
 
 if TYPE_CHECKING:
     from unfurl.configurator import Dependency
+    import toscaparser.repositories
 
 # Tell mypy the logger is of type UnfurlLogger
 logger = cast(UnfurlLogger, logging.getLogger("unfurl"))
@@ -51,7 +62,7 @@ class Operational(ChangeAware):
         return Defaults.shouldRun
 
     @property
-    def local_status(self) -> Status:
+    def local_status(self) -> Optional[Status]:
         return Status.unknown
 
     @property
@@ -112,7 +123,9 @@ class Operational(ChangeAware):
             # return error, pending, or absent
             return status
 
-        dependentStatus = self.aggregate_status(self.get_operational_dependencies(), seen)
+        dependentStatus = self.aggregate_status(
+            self.get_operational_dependencies(), seen
+        )
         if self.is_computed():
             # if local_status is a no op (status purely determined by dependents) set to ok
             status = Status.ok
@@ -126,7 +139,7 @@ class Operational(ChangeAware):
 
     @property
     def status(self) -> Status:
-        return self._status( {id(self):self} )
+        return self._status({id(self): self})
 
     @property
     def required(self) -> bool:
@@ -166,7 +179,7 @@ class Operational(ChangeAware):
         return False
 
     @staticmethod
-    def aggregate_status(statuses: Union[Tuple["Operational"], List["Operational"]], seen) -> Optional[Status]:
+    def aggregate_status(statuses: Iterable["Operational"], seen) -> Optional[Status]:
         """
         Returns: ok, degraded, pending or None
 
@@ -192,7 +205,9 @@ class Operational(ChangeAware):
             if status.priority == Priority.ignore:
                 continue
             if id(status) in seen:
-                logger.verbose(f'Circular operational dependency when checking status {status} in {seen}')
+                logger.debug(
+                    f"Circular operational dependency when checking status {status} in {seen}"
+                )
                 continue
             seen[id(status)] = status
             _status = status._status(seen)
@@ -217,12 +232,12 @@ class OperationalInstance(Operational):
 
     def __init__(
         self,
-        status: Optional[Union["OperationalInstance", int, str]]=None,
-        priority: Optional[Union[int, str, Priority]]=None,
-        manualOveride: Optional[Union["OperationalInstance", int, str]]=None,
-        lastStateChange: Optional[datetime]=None,
-        lastConfigChange: Optional[datetime]=None,
-        state: NodeState=None,
+        status: Optional[Union["OperationalInstance", int, str]] = None,
+        priority: Optional[Union[int, str, Priority]] = None,
+        manualOveride: Optional[Union["OperationalInstance", int, str]] = None,
+        lastStateChange: Optional[datetime] = None,
+        lastConfigChange: Optional[datetime] = None,
+        state: NodeState = None,
     ) -> None:
         if isinstance(status, OperationalInstance):
             self._localStatus: Optional[Status] = status._localStatus
@@ -246,7 +261,7 @@ class OperationalInstance(Operational):
     def get_operational_dependencies(self) -> Iterable["Operational"]:
         return self.dependencies
 
-    def local_status() -> Dict[str, Any]:  # type: ignore
+    def __local_status():  # type: ignore
         doc = "The local_status property."
 
         def fget(self: "OperationalInstance") -> Optional[Status]:
@@ -260,9 +275,9 @@ class OperationalInstance(Operational):
 
         return locals()
 
-    local_status = property(**local_status())  # type: ignore
+    local_status: Optional[Status] = property(**__local_status())  # type: ignore
 
-    def manual_overide_status() -> Dict[str, Any]:  # type: ignore
+    def __manual_overide_status() -> Dict[str, Any]:  # type: ignore
         doc = "The manualOverideStatus property."
 
         def fget(self: "OperationalInstance") -> Optional[Status]:
@@ -271,28 +286,22 @@ class OperationalInstance(Operational):
         def fset(self: "OperationalInstance", value: Status) -> None:
             self._manualOverideStatus = value
 
-        def fdel(self: "OperationalInstance") -> None:
-            del self._manualOverideStatus
-
         return locals()
 
-    manual_overide_status = property(**manual_overide_status())  # type: ignore
+    manual_overide_status: Optional[Status] = property(**__manual_overide_status())  # type: ignore
 
-    def priority():  # type: ignore
+    def __priority():  # type: ignore
         doc = "The priority property."
 
         def fget(self: "OperationalInstance") -> Optional[Priority]:
-            return Defaults.shouldRun if self._priority is None else self._priority
+            return self._priority
 
-        def fset(self: "OperationalInstance", value: Priority) -> None:
-            self._priority = value
-
-        def fdel(self: "OperationalInstance") -> None:
-            del self._priority
+        def fset(self: "OperationalInstance", value):  # type: ignore
+            self._priority = to_enum(Priority, value)
 
         return locals()
 
-    priority = property(**priority())  # type: ignore
+    priority: Optional[Priority] = property(**__priority())  # type: ignore
 
     @property
     def last_state_change(self) -> Optional[datetime]:
@@ -302,7 +311,7 @@ class OperationalInstance(Operational):
     def last_config_change(self) -> Optional[datetime]:
         return self._lastConfigChange
 
-    def state():  # type: ignore
+    def __state():  # type: ignore
         doc = "The state property."
 
         def fget(self: "OperationalInstance") -> Optional[NodeState]:
@@ -313,7 +322,7 @@ class OperationalInstance(Operational):
 
         return locals()
 
-    state = property(**state())  # type: ignore
+    state: Optional[NodeState] = property(**__state())  # type: ignore
 
 
 class _ChildResources(Mapping):
@@ -332,11 +341,13 @@ class _ChildResources(Mapping):
 
 class EntityInstance(OperationalInstance, ResourceRef):
     attributeManager = None
-    created = None
-    shadow = None
+    created: Optional[Union[bool, str]] = None
+    protected: Optional[bool] = None
     imports = None
+    imported = None
     envRules = None
     _baseDir = ""
+    templateType = EntitySpec
 
     def __init__(
         self, name="", attributes=None, parent=None, template=None, status=Status.ok
@@ -351,20 +362,36 @@ class EntityInstance(OperationalInstance, ResourceRef):
             p = getattr(parent, self.parentRelation)
             p.append(self)
 
-        self.template = template or self.templateType()
+        self.template: EntitySpec = template or self.templateType()
+        assert isinstance(self.template, self.templateType)
         self._properties = {}
 
+
+    @property
+    def repository(self) -> Optional["toscaparser.repositories.Repository"]:
+        return None
+
+    def _get_prop(self, name):
+        # hack because `.repository` isn't a ResourceRef and so can't be evaluated in expressions
+        if name == ".repository":
+            if self.repository:
+                return self.repository.tpl
+            else:
+                return None
+        else:
+            return super()._get_prop(name)
+
     def _resolve(self, key):
         # might return a Result
         self.attributes[key]  # force resolve
         return self.attributes._attributes[key]
 
-    def query(self, expr, vars=None, wantList=False):
+    def query(self, expr, vars=None, wantList=False, trace=None):
         from .eval import Ref, RefContext
 
-        return Ref(expr).resolve(RefContext(self, vars=vars), wantList)
+        return Ref(expr).resolve(RefContext(self, vars=vars, trace=trace), wantList)
 
-    def local_status():
+    def __local_status():  # type: ignore
         doc = "The working_dir property."
 
         def fget(self):
@@ -380,7 +407,7 @@ class EntityInstance(OperationalInstance, ResourceRef):
 
         return locals()
 
-    local_status = property(**local_status())
+    local_status: Optional[Status] = property(**__local_status())  # type: ignore
 
     def get_operational_dependencies(self):
         if self.parent and self.parent is not self.root:
@@ -390,7 +417,7 @@ class EntityInstance(OperationalInstance, ResourceRef):
             yield d
 
     def is_computed(self):
-      return self.template.aggregate_only()
+        return self.template.aggregate_only()
 
     @property
     def key(self):
@@ -443,6 +470,31 @@ class EntityInstance(OperationalInstance, ResourceRef):
     def get_default_relationships(self, relation=None):
         return self.root.get_default_relationships(relation)
 
+    @property
+    def shadow(self):
+        if self.imported:
+            imports = self.root.imports
+            if imports and self.imported in imports:
+                return imports[self.imported].external_instance
+            else:
+                raise UnfurlError(
+                    f'could not find imported instance "{self.imported}" for local instance "{self.name}"'
+                )
+        return None
+
+    @property
+    def readonly(self) -> bool:
+        return bool(self.imported)  # imported instances are readonly
+
+    def validate(self) -> None:
+        """
+        Raises UnfurlValidationError on failure.
+        """
+        if self.shadow:
+            self.shadow.validate()
+        elif self.template:
+            self.template.validate()
+
     def __eq__(self, other):
         if self is other:
             return True
@@ -475,6 +527,83 @@ class EntityInstance(OperationalInstance, ResourceRef):
         return f"{self.__class__}('{self.name}')"
 
 
+class HasInstancesInstance(EntityInstance):
+    parentRelation = "instances"
+
+    def __init__(
+        self, name="", attributes=None, parent=None, template=None, status=None
+    ):
+        if parent:
+            # only node instances have unique names
+            if parent.root.find_resource(name):
+                raise UnfurlError(
+                    f'can not create node instance "{name}", its name is already in use'
+                )
+
+        self.instances = []
+        EntityInstance.__init__(self, name, attributes, parent, template, status)
+        if self.root is self:
+            self._all = _ChildResources(self)
+            self._templar = Templar(DataLoader())
+
+    @property
+    def key(self):
+        return f"::{self.name}"
+
+    def get_self_and_descendents(self):
+        "Recursive descendent including self"
+        yield self
+        for r in self.instances:
+            for descendent in r.get_self_and_descendents():
+                yield descendent
+
+    @property
+    def descendents(self):
+        return list(self.get_self_and_descendents())
+
+    # XXX use find_instance instead and remove find_resource
+    def find_resource(self, resourceid):
+        if self.name == resourceid:
+            return self
+        for r in self.instances:
+            child = r.find_resource(resourceid)
+            if child:
+                return child
+        return None
+
+    find_instance = find_resource
+
+    def find_instance_or_external(self, resourceid):
+        instance = self.find_instance(resourceid)
+        if instance:
+            return instance
+        if self.imports:
+            return self.imports.find_import(resourceid)
+        return None
+
+    def get_requirements(self, match):
+        if match is None:
+            return self.requirements
+        if isinstance(match, six.string_types):
+            return [r for r in self.requirements if r.template.name == match]
+        elif isinstance(match, NodeInstance):
+            return [r for r in self.requirements if r.target == match]
+        elif isinstance(match, CapabilityInstance):
+            return [r for r in self.requirements if r.parent == match]
+        elif isinstance(match, ArtifactInstance):
+            return []
+        else:
+            raise UnfurlError(f'invalid match for get_requirements: "{match}"')
+
+    def get_operational_dependents(self, seen=None):
+        if seen is None:
+            seen = set()
+        for instance in self.instances:
+            if id(instance) not in seen:
+                seen.add(id(instance))
+                yield instance
+
+
 # both have occurrences
 # only need to configure capabilities as required by a relationship
 class CapabilityInstance(EntityInstance):
@@ -482,12 +611,13 @@ class CapabilityInstance(EntityInstance):
     # 3.8.1 Capability assignment p. 114
     parentRelation = "_capabilities"
     templateType = CapabilitySpec
-    _relationships = None
+    _relationships: Optional[List["RelationshipInstance"]] = None
 
     @property
-    def relationships(self):
+    def relationships(self) -> List["RelationshipInstance"]:
         if self._relationships is None:
             self._relationships = []
+        assert isinstance(self.template, CapabilitySpec)
         if len(self._relationships) != len(self.template.relationships):
             # instantiate missing relationships (they are only added, never deleted)
             instantiated = {id(c.template) for c in self._relationships}
@@ -516,33 +646,45 @@ class RelationshipInstance(EntityInstance):
     # 3.8.2 Requirements assignment p. 115
     parentRelation = "_relationships"
     templateType = RelationshipSpec
+    if TYPE_CHECKING:
+        template = templateType()
     source = None
 
     @property
-    def target(self):
+    def target(self) -> Optional["NodeInstance"]:
         # parent is a capability, return it's parent (a Node)
-        return self.parent.parent if self.parent else None
+        if self.parent:
+            return self.parent.parent
+        else:
+            return None
 
     @property
-    def key(self):
+    def capability(self) -> Optional[CapabilityInstance]:
+        return self.parent
+
+    @property
+    def key(self) -> str:
         # XXX implement something like _ChildResources to enable ::name instead of [.name=name]
         if self.source:
             return f"{self.source.key}::.requirements::[.name={self.name}]"
-        elif self.parent is self.root:  # it's a default relationship
+        elif not self.parent or self.parent is self.root:  # it's a default relationship
             return f"::.requirements::[.name={self.name}]"
-        else:  # capability is parent
+        elif self.parent:  # capability is parent
             return f"{self.parent.key}::.relationships::[.name={self.name}]"
 
-    def merge_props(self, matchfn, delete_if_none=False):
-        env = {}
+    def merge_props(self, matchfn, delete_if_none=False) -> Dict[str, str]:
+        env: Dict[str, str] = {}
         capability = self.parent
-        for name, val in capability.template.find_props(capability.attributes, matchfn):
-            if val is None:
-                if delete_if_none:
-                    env.pop(name, None)
-            else:
-                env[name] = val
-        for name, val in self.template.find_props(self.attributes, matchfn):
+        if capability:
+            for name, val in matchfn(
+                capability.template.find_props(capability.attributes)
+            ):
+                if val is None:
+                    if delete_if_none:
+                        env.pop(name, None)
+                else:
+                    env[name] = val
+        for name, val in matchfn(self.template.find_props(self.attributes)):
             if val is None:
                 if delete_if_none:
                     env.pop(name, None)
@@ -553,7 +695,9 @@ class RelationshipInstance(EntityInstance):
 
 class ArtifactInstance(EntityInstance):
     parentRelation = "_artifacts"
-    templateType = ArtifactSpec
+    templateType = ArtifactSpec  # type: ignore # XXX type error doesn't make sense
+    if TYPE_CHECKING:
+        template = ArtifactSpec({})
 
     def __init__(
         self, name="", attributes=None, parent=None, template=None, status=None
@@ -569,9 +713,9 @@ class ArtifactInstance(EntityInstance):
         return self.template.file
 
     @property
-    def repository(self):
+    def repository(self) -> Optional["toscaparser.repositories.Repository"]:
         return self.template.repository
-
+  
     def get_path(self, resolver=None):
         return self.template.get_path_and_fragment(resolver)
 
@@ -587,9 +731,10 @@ class ArtifactInstance(EntityInstance):
             yield d
 
 
-class NodeInstance(EntityInstance):
+class NodeInstance(HasInstancesInstance):
     templateType = NodeSpec
-    parentRelation = "instances"
+    if TYPE_CHECKING:
+        template = templateType()
 
     def __init__(
         self, name="", attributes=None, parent=None, template=None, status=None
@@ -606,19 +751,14 @@ class NodeInstance(EntityInstance):
         self._requirements = []
         self._artifacts = []
         self._named_artifacts = None
-        self.instances = []
-        EntityInstance.__init__(self, name, attributes, parent, template, status)
-
-        if self.root is self:
-            self._all = _ChildResources(self)
-            self._templar = Templar(DataLoader())
+        HasInstancesInstance.__init__(self, name, attributes, parent, template, status)
 
         self._interfaces = {}
         # preload
         self.get_interface("inherit")
         self.get_interface("default")
 
-    def _find_relationship(self, relationship):
+    def _find_relationship(self, relationship) -> Optional[RelationshipInstance]:
         """
         Find RelationshipInstance that has the give relationship template
         """
@@ -636,8 +776,8 @@ class NodeInstance(EntityInstance):
         return None
 
     @property
-    def requirements(self):
-        # if self._requirements is None:
+    def requirements(self) -> List[RelationshipInstance]:
+        assert isinstance(self.template, NodeSpec)
         if len(self._requirements) != len(self.template.requirements):
             # instantiate missing relationships (they are only added, never deleted)
             instantiated = {id(r.template) for r in self._requirements}
@@ -657,22 +797,8 @@ class NodeInstance(EntityInstance):
 
         return self._requirements
 
-    def get_requirements(self, match):
-        if match is None:
-            return self.requirements
-        if isinstance(match, six.string_types):
-            return [r for r in self.requirements if r.template.name == match]
-        elif isinstance(match, NodeInstance):
-            return [r for r in self.requirements if r.target == match]
-        elif isinstance(match, CapabilityInstance):
-            return [r for r in self.requirements if r.parent == match]
-        elif isinstance(match, ArtifactInstance):
-            return []
-        else:
-            raise UnfurlError(f'invalid match for get_requirements: "{match}"')
-
     @property
-    def capabilities(self):
+    def capabilities(self) -> List[CapabilityInstance]:
         if len(self._capabilities) != len(self.template.capabilities):
             # instantiate missing capabilities (they are only added, never deleted)
             instantiated = {id(c.template) for c in self._capabilities}
@@ -689,7 +815,7 @@ class NodeInstance(EntityInstance):
 
         return self._capabilities
 
-    def get_capabilities(self, name):
+    def get_capabilities(self, name) -> List[CapabilityInstance]:
         return [
             capability
             for capability in self.capabilities
@@ -701,18 +827,26 @@ class NodeInstance(EntityInstance):
         # only include named artifacts
         if self._named_artifacts is None:
             self._named_artifacts = {}
-            instantiated = {a.name: a for a in self._artifacts}
-            for name, template in self.template.artifacts.items():
-                artifact = instantiated.get(name)
-                if not artifact:
-                    artifact = ArtifactInstance(
-                        template.name, parent=self, template=template
-                    )
-                    assert artifact in self._artifacts
+            if self.template:
+                instantiated = {a.name: a for a in self._artifacts}
+                for name, template in self.template.artifacts.items():
+                    artifact = instantiated.get(name)
+                    if not artifact:
+                        artifact = ArtifactInstance(
+                            template.name, parent=self, template=template
+                        )
+                        assert artifact in self._artifacts
 
-                self._named_artifacts[template.name] = artifact
+                    self._named_artifacts[template.name] = artifact
         return self._named_artifacts
 
+    @property
+    def repository(self) -> Optional["toscaparser.repositories.Repository"]:
+        if self.template.is_compatible_type("unfurl.nodes.Repository"):
+            # this is a reified repository -- return the Repository with the same name
+            return self.template.spec.template.repositories.get(self.template.name)
+        return None
+
     def _get_default_relationships(self, relation=None):
         if self.root is self:
             return
@@ -743,24 +877,27 @@ class NodeInstance(EntityInstance):
         for cap in self.capabilities:
             for rel in cap.relationships:
                 if rel.source:
-                      if rel.template.is_compatible_type("unfurl.relationships.Configures"):
-                          yield rel.source
-                      yield from rel.source._configured_by()
+                    if rel.template.is_compatible_type(
+                        "unfurl.relationships.Configures"
+                    ):
+                        yield rel.source
+                    yield from rel.source._configured_by()
 
     @property
     def configured_by(self):
-      return list(self._configured_by())
+        return list(self._configured_by())
 
     @property
-    def targets(self):
-        dep = {}
+    def targets(self) -> Dict[str, Union["NodeInstance", List["NodeInstance"]]]:
+        dep: Dict[str, Union[NodeInstance, List[NodeInstance]]] = {}
         for rel in self.requirements:
             if rel.target:
                 if rel.name in dep:
-                    if isinstance(dep[rel.name], list):
-                        dep[rel.name].append(rel.target)
+                    target = dep[rel.name]
+                    if isinstance(target, list):
+                        target.append(rel.target)
                     else:
-                        dep[rel.name] = [dep[rel.name], rel.target]
+                        dep[rel.name] = [target, rel.target]
                 else:
                     dep[rel.name] = rel.target
         return dep
@@ -793,20 +930,16 @@ class NodeInstance(EntityInstance):
                 else:
                     raise
 
-    @property
-    def key(self):
-        return f"::{self.name}"
-
     def get_operational_dependencies(self):
-        for dep in super().get_operational_dependencies():
-            yield dep
+        yield from super().get_operational_dependencies()
 
         for instance in self.requirements:
-            if instance is not self.parent:
+            if instance.target is not self.parent:
                 yield instance
 
-    def get_operational_dependents(self):
-        seen = set()
+    def get_operational_dependents(self, seen=None):
+        if seen is None:
+            seen = set()
         for cap in self.capabilities:
             for rel in cap.relationships:
                 dep = rel.source
@@ -814,41 +947,7 @@ class NodeInstance(EntityInstance):
                     seen.add(id(dep))
                     yield dep
 
-        for instance in self.instances:
-            if id(instance) not in seen:
-                seen.add(id(instance))
-                yield instance
-
-    def get_self_and_descendents(self):
-        "Recursive descendent including self"
-        yield self
-        for r in self.instances:
-            for descendent in r.get_self_and_descendents():
-                yield descendent
-
-    @property
-    def descendents(self):
-        return list(self.get_self_and_descendents())
-
-    # XXX use find_instance instead and remove find_resource
-    def find_resource(self, resourceid):
-        if self.name == resourceid:
-            return self
-        for r in self.instances:
-            child = r.find_resource(resourceid)
-            if child:
-                return child
-        return None
-
-    find_instance = find_resource
-
-    def find_instance_or_external(self, resourceid):
-        instance = self.find_resource(resourceid)
-        if instance:
-            return instance
-        if self.imports:
-            return self.imports.find_import(resourceid)
-        return None
+        yield from super().get_operational_dependents(seen)
 
     def add_interface(self, klass, name=None):
         if not isinstance(klass, six.string_types):
@@ -879,14 +978,12 @@ class NodeInstance(EntityInstance):
         return f"NodeInstance('{self.name}')"
 
 
-class TopologyInstance(NodeInstance):
+class TopologyInstance(HasInstancesInstance):
     templateType = TopologySpec
 
     def __init__(self, template, status=None):
         attributes = dict(inputs=template.inputs, outputs=template.outputs)
-        NodeInstance.__init__(
-            self, "root", attributes, template=template, status=status
-        )
+        HasInstancesInstance.__init__(self, "root", attributes, None, template, status)
 
         self._relationships = None
         self._tmpDir = None
@@ -899,7 +996,7 @@ class TopologyInstance(NodeInstance):
             self._templar = Templar(loader)
 
     @property
-    def requirements(self):
+    def requirements(self) -> List[RelationshipInstance]:
         """
         The root node returns RelationshipInstances representing default relationship templates
         """
