@@ -1,4 +1,3 @@
-
 # Copyright (c) 2020 Adam Souzis
 # SPDX-License-Identifier: MIT
 """
@@ -13,27 +12,37 @@ import os.path
 import six
 import re
 import ast
-from enum import IntEnum
+import time
+from typing import TYPE_CHECKING, cast, Dict, Optional, Any
+from enum import Enum
+from urllib.parse import urlsplit
+
+
 from .eval import RefContext, set_eval_func, Ref, map_value
 from .result import (
     Results,
+    ResultsList,
     ResultsMap,
     Result,
     ExternalValue,
     serialize_value,
+    is_sensitive_schema,
 )
 from .util import (
     ChainMap,
     find_schema_errors,
     UnfurlError,
     UnfurlValidationError,
+    UnfurlTaskError,
     assert_form,
     wrap_sensitive_value,
     is_sensitive,
     load_module,
     load_class,
     sensitive,
-    filter_env
+    filter_env,
+    env_var_value,
+    get_random_password,
 )
 from .merge import intersect_dict, merge_dicts
 from unfurl.projectpaths import get_path
@@ -41,7 +50,7 @@ import ansible.template
 from ansible.parsing.dataloader import DataLoader
 from ansible.utils import unsafe_proxy
 from ansible.utils.unsafe_proxy import wrap_var, AnsibleUnsafeText, AnsibleUnsafeBytes
-from jinja2.runtime import ChainableUndefined
+from jinja2.runtime import DebugUndefined, make_logging_undefined
 from toscaparser.elements.portspectype import PortSpec
 
 import logging
@@ -49,7 +58,7 @@ import logging
 logger = logging.getLogger("unfurl")
 
 
-class Status(IntEnum):
+class Status(int, Enum):
     unknown = 0
     ok = 1
     degraded = 2
@@ -57,28 +66,52 @@ class Status(IntEnum):
     pending = 4
     absent = 5
 
+    @property
+    def color(self):
+        return {
+            Status.unknown: "white",
+            Status.ok: "green",
+            Status.degraded: "yellow",
+            Status.error: "red",
+            Status.pending: "white",
+            Status.absent: "yellow",
+        }[self]
+ 
 
 # see "3.4.1 Node States" p74
-NodeState = IntEnum(
-    "NodeState",
-    "initial creating created configuring configured starting started stopping stopped deleting deleted error",
-    module=__name__,
-)
+class NodeState(int, Enum):
+    initial = 1
+    creating = 2
+    created = 3
+    configuring = 4
+    configured = 5
+    starting = 6
+    started = 7
+    stopping = 8
+    stopped = 9
+    deleting = 10
+    deleted = 11
+    error = 12
 
-# ignore may must
-Priority = IntEnum(
-    "Priority", "ignore optional required critical", start=0, module=__name__
-)
+
+class Priority(int, Enum):
+    ignore = 0
+    optional = 1
+    required = 2
+    critical = 3
 
 
 class Reason:
-    pass
-
-
-for (
-    r
-) in "add reconfigure force upgrade update missing error degraded prune run".split():
-    setattr(Reason, r, r)
+    add = "add"
+    reconfigure = "reconfigure"
+    force = "force"
+    upgrade = "upgrade"
+    update = "update"
+    missing = "missing"
+    error = "error"
+    degraded = "degraded"
+    prune = "prune"
+    run = "run"
 
 
 class Defaults:
@@ -258,15 +291,6 @@ def _wrap_sequence(v):
 unsafe_proxy._wrap_sequence = _wrap_sequence
 
 
-class UnfurlUndefined(ChainableUndefined):
-     def __getattr__(self, name):
-          if name == '__UNSAFE__':
-              # AnsibleUndefined should never be assumed to be unsafe
-              # This prevents ``hasattr(val, '__UNSAFE__')`` from evaluating to ``True``
-              raise AttributeError(name)
-          # Return original Undefined object to preserve the first failure context
-          return self
-
 def apply_template(value, ctx, overrides=None):
     if not isinstance(value, six.string_types):
         msg = f"Error rendering template: source must be a string, not {type(value)}"
@@ -275,6 +299,50 @@ def apply_template(value, ctx, overrides=None):
         else:
             return f"<<{msg}>>"
     value = value.strip()
+    if ctx.task:
+        logger = ctx.task.logger
+    else:
+        logger = logging.getLogger("unfurl")
+
+    # local class to bind with logger and ctx
+    class _UnfurlUndefined(DebugUndefined):
+        __slots__ = ()
+
+        def __getattr__(self, name):
+            if name == "__UNSAFE__":
+                # see AnsibleUndefined
+                # self should never be assumed to be unsafe
+                # This prevents ``hasattr(val, '__UNSAFE__')`` from evaluating to ``True``
+                raise AttributeError(name)
+            # Return original Undefined object to preserve the first failure context
+            return self  # see ChainableUndefined
+
+        __getitem__ = __getattr__  # type: ignore
+
+        def _log_message(self) -> None:
+            msg = "Template: %s", self._undefined_message
+            # XXX override _undefined_message: if self._undefined_obj is a Results then add its ctx._lastResource to the msg
+            #     see https://github.com/pallets/jinja/blob/7d72eb7fefb7dce065193967f31f805180508448/src/jinja2/runtime.py#L824
+            logger.warning(msg)
+            if ctx.task:  # already logged, so don't log
+                ctx.task._errors.append(UnfurlTaskError(ctx.task, msg, False))
+
+        # see LoggingUndefined:
+        def __str__(self) -> str:
+            self._log_message()
+            return super().__str__()  # type: ignore
+
+        def __iter__(self):
+            self._log_message()
+            return super().__iter__()  # type: ignore
+
+        def __bool__(self) -> bool:
+            self._log_message()
+            return super().__bool__()  # type: ignore
+
+        # see ChainableUndefined
+        def __html__(self) -> str:
+            return str(self)
 
     # implementation notes:
     #   see https://github.com/ansible/ansible/test/units/template/test_templar.py
@@ -300,9 +368,13 @@ def apply_template(value, ctx, overrides=None):
     # templar.environment.lstrip_blocks = False
     fail_on_undefined = ctx.strict
     if not fail_on_undefined:
-        templar.environment.undefined = UnfurlUndefined
+        templar.environment.undefined = _UnfurlUndefined
 
-    vars = _VarTrackerDict(__unfurl=ctx, __python_executable=sys.executable)
+    vars = _VarTrackerDict(
+        __unfurl=ctx, __python_executable=sys.executable, __now=time.time()
+    )
+    if hasattr(ctx.currentResource, "attributes"):
+        vars["SELF"] = ctx.currentResource.attributes
     vars.update(ctx.vars)
     vars.ctx = ctx
 
@@ -322,9 +394,14 @@ def apply_template(value, ctx, overrides=None):
             value = templar.template(value, fail_on_undefined=fail_on_undefined)
         except Exception as e:
             msg = str(e)
+            # XXX have _UnfurlUndefined throw an exception with the missing obj and key
             match = re.search(r"has no attribute '(\w+)'", msg)
             if match:
                 msg = f'missing attribute or key: "{match.group(1)}"'
+            else:
+                match = re.search(r"'(\w+)' is undefined", msg)
+                if match:
+                    msg = f'missing variable: "{match.group(1)}"'
             value = f"<<Error rendering template: {msg}>>"
             if ctx.strict:
                 logger.debug(value, exc_info=True)
@@ -332,6 +409,8 @@ def apply_template(value, ctx, overrides=None):
             else:
                 logger.warning(value[2:100] + "... see debug log for full report")
                 logger.debug(value, exc_info=True)
+                if ctx.task:
+                    ctx.task._errors.append(UnfurlTaskError(ctx.task, msg))
         else:
             if value != oldvalue:
                 ctx.trace("successfully processed template:", value)
@@ -433,9 +512,19 @@ set_eval_func("lookup", lookup_func)
 
 
 def get_input(arg, ctx):
+    if isinstance(arg, list):
+        name = arg[0]
+        default = arg[1]
+        has_default = True
+    else:
+        name = arg
+        has_default = False
+
     try:
-        return ctx.currentResource.root.attributes["inputs"][arg]
+        return ctx.currentResource.root.attributes["inputs"][name]
     except KeyError:
+        if has_default:
+            return map_value(default, ctx)
         raise UnfurlError(f"undefined input '{arg}'")
 
 
@@ -482,7 +571,7 @@ def has_env(arg, ctx):
 set_eval_func("has_env", has_env, True)
 
 
-def get_env(args, ctx):
+def get_env(args, ctx: RefContext):
     """
     Return the value of the given environment variable name.
     If NAME is not present in the environment, return the given default value if supplied or return None.
@@ -507,7 +596,133 @@ def get_env(args, ctx):
 
 set_eval_func("get_env", get_env, True)
 
-set_eval_func("to_env", lambda args, ctx: filter_env(map_value(args, ctx), addOnly=True))
+
+def set_context_vars(vars, resource):
+    root = resource.root
+    ROOT = {}
+    vars.update(dict(NODES=TopologyMap(root), ROOT=ROOT, TOPOLOGY=ROOT))
+    if "inputs" in root._attributes:
+        ROOT.update(
+            dict(
+                inputs=root._attributes["inputs"],
+                outputs=root._attributes["outputs"],
+            )
+        )
+    app_template = root.template.spec.substitution_template
+    if app_template:
+        app = root.find_instance(app_template.name)
+        if app:
+            ROOT["app"] = app.attributes
+        for name, req in app_template.requirements.items():
+            if req.relationship and req.relationship.target:
+                target = root.find_instance(req.relationship.target.name)
+                if target:
+                    ROOT[name] = target.attributes
+    return vars
+
+
+class _EnvMapper(dict):
+    """Resolve environment variable name to instance properties via the root template's requirements.
+    Pattern should match _generate_env_names in to_json.py and set_context_vars above.
+    """
+
+    def __missing__(self, key):
+        objname, sep, prop = key.partition("_")
+        root = self.ctx.currentResource.root
+        app = root.template.spec.substitution_template
+        if app and objname and prop:
+            obj = None
+            if objname == "APP":
+                obj = app
+            else:
+                for name, req in app.requirements.items():
+                    if name.upper() == objname:
+                        if req.relationship and req.relationship.target:
+                            obj = req.relationship.target
+            if obj:
+                instance = root.find_instance(obj.name)
+                if instance:
+                    for key in instance.attributes:
+                        if key.upper() == prop:
+                            return env_var_value(instance.attributes[key], self)
+        raise KeyError(key)
+
+
+def to_env(args, ctx: RefContext):
+    env = None
+    if ctx.task:
+        env = ctx.task.get_environment(False)
+    sub = _EnvMapper(env or {})
+    sub.ctx = ctx  # type: ignore
+
+    rules = map_value(args or {}, ctx)
+    assert isinstance(rules, Mapping)
+    result = filter_env(rules, env, True, sub)
+    if ctx.kw.get("update_os_environ"):
+        os.environ.update(result)
+        for key, value in rules.items():
+            if value is None and key in os.environ:
+                del os.environ[key]
+    return result
+
+
+set_eval_func("to_env", to_env)
+
+
+def to_label(arg, extra="", max=63, lower=False, replace=""):
+    if isinstance(arg, Mapping):
+        return {
+            to_label(n, extra, max, lower, replace): to_label(
+                v, extra, max, lower, replace
+            )
+            for n, v in arg.items()
+            if v is not None
+        }
+    elif isinstance(arg, str):
+        val = re.sub(f"[^\\w{extra}]", replace, arg)
+        if lower:
+            val = val.lower()
+        return val[:max]
+
+
+def to_dns_label(arg):
+    """
+    The maximum length of each label is 63 characters, and a full domain name can have a maximum of 253 characters.
+    Alphanumeric characters and hyphens can be used in labels, but a domain name must not commence or end with a hyphen.
+    """
+    return to_label(arg, extra="-", replace="--")
+
+
+set_eval_func(
+    "to_dns_label",
+    lambda arg, ctx: to_dns_label(map_value(arg, ctx)),
+)
+
+
+def to_kubernetes_label(arg):
+    """
+    See https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#syntax-and-character-set
+    """
+    return to_label(arg, extra="_.-", replace="__")
+
+
+set_eval_func(
+    "to_kubernetes_label",
+    lambda arg, ctx: to_kubernetes_label(map_value(arg, ctx)),
+)
+
+
+def to_googlecloud_label(arg):
+    """
+    See https://cloud.google.com/resource-manager/docs/creating-managing-labels#requirements
+    """
+    return to_label(arg, extra="_-", lower=True, replace="__")
+
+
+set_eval_func(
+    "to_googlecloud_label",
+    lambda arg, ctx: to_googlecloud_label(map_value(arg, ctx)),
+)
 
 _toscaKeywordsToExpr = {
     "SELF": ".",
@@ -519,7 +734,7 @@ _toscaKeywordsToExpr = {
 }
 
 
-def get_attribute(args, ctx):
+def get_attribute(args, ctx: RefContext):
     args = map_value(args, ctx)
     entity_name = args.pop(0)
     candidate_name = args.pop(0)
@@ -543,7 +758,7 @@ def get_attribute(args, ctx):
 set_eval_func("get_attribute", get_attribute, True)
 
 
-def get_nodes_of_type(type_name, ctx):
+def get_nodes_of_type(type_name, ctx: RefContext):
     return [
         r
         for r in ctx.currentResource.root.get_self_and_descendents()
@@ -555,53 +770,215 @@ def get_nodes_of_type(type_name, ctx):
 set_eval_func("get_nodes_of_type", get_nodes_of_type, True)
 
 
-def get_artifact(ctx, entity_name, artifact_name, location=None, remove=None):
+set_eval_func("_generate", lambda arg, ctx: get_random_password(10, ""), True)
+
+
+class ContainerImage(ExternalValue):
+    """
+    Represents a container image.
+    get() returns name of the image, which may be qualified
+    with the registry url, repository name, tag, or digest
+
+    All of the following are valid:
+
+    "name"
+    "repository/name"
+    "name:tag"
+    "repository/name:tag"
+    "registry.docker.io/repository/name
+    "registry.example.com:8080/repository/name:tag"
+    "registry-1.docker.io/repository/name@sha256:digest"
+    """
+
+    # https://docs.docker.com/engine/reference/commandline/tag/
+    # An image name is made up of slash-separated name components,
+    # optionally prefixed by a registry hostname. The hostname must comply with standard DNS rules,
+    # but may not contain underscores. If a hostname is present, it may optionally be followed by
+    # a port number in the format :8080. If not present, the command uses Docker’s public registry
+    # located at registry-1.docker.io by default. Name components may contain lowercase letters, digits
+    # and separators. A separator is defined as a period, one or two underscores, or one or more dashes.
+    # A name component may not start or end with a separator.
+    # A tag name must be valid ASCII and may contain lowercase and uppercase letters, digits, underscores, periods and dashes.
+    # # A tag name may not start with a period or a dash and may contain a maximum of 128 characters.
+
+    def __init__(
+        self,
+        name: str,
+        tag=None,
+        digest=None,
+        registry_host=None,
+        username=None,
+        password=None,
+        source_digest=None,
+    ):
+        self.name = name.lstrip("/").lower()
+        self.tag = tag
+        self.digest = digest
+        self.registry_host = registry_host
+        self.username = username
+        self.password = password
+        self.source_digest = source_digest
+        super().__init__("container_image", self.get())
+
+    # XXX
+    # def resolve_key(self, name=None, currentResource=None):
+    #     # hostname: registry-1.docker.io, name, tag, digest
+
+    def get(self) -> str:
+        if self.registry_host:
+            name = os.path.join(self.registry_host, self.name)
+        else:
+            name = self.name
+
+        if self.tag:
+            return f"{name}:{self.tag}"
+        if self.digest:
+            if "@" == self.digest[0]:
+                return name + self.digest
+            else:
+                return f"{name}@sha256:{self.digest}"
+        return name
+
+    @staticmethod
+    def split(artifact_name):
+        if not artifact_name:
+            return None, None, None, None
+        hostname = None
+        namespace, sep, name = artifact_name.partition("/")
+        if sep and (":" in namespace or artifact_name.count("/") > 1):
+            # heuristic because name can look like a hostname
+            hostname = namespace
+        else:
+            name = artifact_name
+
+        tag = None
+        name, sep, digest = name.partition("@")
+        if not sep:
+            digest = None
+            name, sep, qualifier = artifact_name.partition(":")
+            if sep:
+                tag = qualifier
+        return name.lower(), tag, digest, hostname
+
+    @staticmethod
+    def make(artifact_name):
+        return ContainerImage(*ContainerImage.split(artifact_name))
+
+    @staticmethod
+    def resolve_name(base_name: str, artifact_name: str):
+        if not base_name:
+            return artifact_name
+        # support more qualified name such as image name or tag
+        name, sep, qualifier = artifact_name.partition("@")
+        if not sep:
+            name, sep, qualifier = artifact_name.partition(":")
+
+        # if beginning of name overlaps with end of self.name discard it
+        segs, new_segs = base_name.split("/"), name.split("/")
+        while segs[-len(new_segs) :] == new_segs:
+            new_segs.pop(0)
+        if new_segs:
+            name = base_name + "/" + "/".join(new_segs)
+        else:
+            name = base_name
+        return name.lower() + sep + qualifier
+
+    def __digestable__(self, options):
+        if self.digest:
+            return self.digest
+        else:
+            return (self.source_digest, self.get())
+
+
+set_eval_func(
+    "container_image", lambda arg, ctx: ContainerImage.make(map_value(arg, ctx))
+)
+
+
+def _get_instances_from_keyname(ctx, entity_name):
+    ctx = ctx.copy(ctx._lastResource)
+    query = _toscaKeywordsToExpr.get(entity_name, "::" + entity_name)
+    instances = cast(ResultsList, ctx.query(query, wantList=True))
+    if instances:
+        return instances
+    else:
+        ctx.trace("entity_name not found", entity_name)
+        return None
+
+
+if TYPE_CHECKING:
+    from .tosca import ArtifactSpec
+
+
+def _find_artifact(instances, artifact_name) -> "Optional[ArtifactSpec]":
+    for instance in instances:
+        # XXX implement instance.artifacts
+        artifact = instance.template.artifacts.get(artifact_name)
+        if artifact:
+            return artifact
+    return None
+
+
+def _get_container_image_from_repository(entity, artifact_name):
+    # aka get_artifact_as_value
+    name, tag, digest, hostname = ContainerImage.split(artifact_name)
+    attr = entity.attributes
+
+    repository_id = attr.get("repository_id")
+    if repository_id and name:
+        name = ContainerImage.resolve_name(repository_id, name)
+    else:
+        name = repository_id or name
+
+    tag = tag or attr.get("repository_tag")
+
+    if attr.get("registry_url"):
+        hostname = attr["registry_url"]
+        if "//" in hostname:
+            hostname = urlsplit(attr["registry_url"]).netloc
+    username = attr.get("username")
+    password = attr.get("password")
+    source_digest = attr.get("revision")
+    return ContainerImage(
+        name, tag, digest, hostname, username, password, source_digest
+    )
+
+
+def get_artifact(ctx: RefContext, entity, artifact_name, location=None, remove=None):
     """
     Returns either an URL or local path to the artifact
     See section "4.8.1 get_artifact" in TOSCA 1.3 (p. 189)
 
-    If the artifact is a Docker image, return the image name in the form of
+    If the artifact is a container image, return the image name in the form of
     "registry/repository/name:tag" or "registry/repository/name@sha256:digest"
 
     If entity_name or artifact_name is not found return None.
     """
-    ctx = ctx.copy(ctx._lastResource)
-    query = _toscaKeywordsToExpr.get(entity_name, "::" + entity_name)
-    instances = ctx.query(query, wantList=True)
+    from .runtime import NodeInstance, ArtifactInstance
 
-    if not instances:
-        ctx.trace("entity_name not found", entity_name)
-        return None
+    if not entity:
+        return ContainerImage.make(artifact_name)  # XXX assume its a container image
+    if isinstance(entity, ArtifactInstance):
+        return entity.template.as_value()
+    if isinstance(entity, str):
+        instances = _get_instances_from_keyname(ctx, entity)
+    elif isinstance(entity, NodeInstance):
+        if entity.template.is_compatible_type("unfurl.nodes.Repository"):
+            # XXX retrieve method from template definition
+            return _get_container_image_from_repository(entity, artifact_name)
+        else:
+            instances = [entity]
     else:
-        for instance in instances:
-            # XXX implement instance.artifacts
-            artifact = instance.template.artifacts.get(artifact_name)
-            if artifact:
-                break
+        return None
+
+    if instances:
+        artifact = _find_artifact(instances, artifact_name)
+        if artifact:
+            return artifact.as_value()
         else:
             ctx.trace("artifact not found", artifact_name)
-            return None
-
-    artifactDef = artifact.toscaEntityTemplate
-    if artifactDef.is_derived_from("tosca.artifacts.Deployment.Image.Container.Docker"):
-        # if artifact is an image in a registry return image path and location isn't specified
-        if artifactDef.checksum:
-            name = (
-                artifactDef.file
-                + "@"
-                + artifactDef.checksum_algorithm
-                + ":"
-                + artifactDef.checksum
-            )
-        elif "tag" in artifact.properties:
-            name = artifactDef.file + ":" + artifact.properties["tag"]
-        else:
-            name = artifactDef.file
-        if artifact.repository:
-            return artifact.repository.hostname + "/" + name
-        return name
-    raise UnfurlError("get_artifact not implemented")
-    # XXX
+    return None
+    # XXX TOSCA 1.3 stuff:
     # if not location:
     #     #if  deploy_path # update artifact instance
     #     return artifact.getUrl()
@@ -610,10 +987,10 @@ def get_artifact(ctx, entity_name, artifact_name, location=None, remove=None):
     #     # if location == 'LOCAL_FILE':
 
 
-set_eval_func("get_artifact", lambda args, ctx: get_artifact(ctx, *args), True)
+set_eval_func("get_artifact", lambda args, ctx: get_artifact(ctx, *(map_value(args, ctx))), True)  # type: ignore
 
 
-def get_import(arg, ctx):
+def get_import(arg: RefContext, ctx):
     """
     Returns the external resource associated with the named import
     """
@@ -630,35 +1007,65 @@ def get_import(arg, ctx):
 set_eval_func("external", get_import)
 
 
-_Import = collections.namedtuple("_Import", ["resource", "spec"])
+class _Import:
+    if TYPE_CHECKING:
+        from .runtime import EntityInstance
+
+    def __init__(
+        self,
+        external_instance: "EntityInstance",
+        spec: dict,
+        local_instance: Optional["EntityInstance"] = None,
+    ):
+        self.external_instance = external_instance
+        self.spec = spec
+        self.local_instance = local_instance
 
 
 class Imports(collections.OrderedDict):
-    def find_import(self, name):
+    def find_import(self, qualified_name):
+        # return a local shadow of the imported instance
+        # or the imported instance itself if no local shadow exist (yet).
+        imported = self._find_import(qualified_name)
+        if imported:
+            return imported
+        iName, sep, rName = qualified_name.partition(":")
+        localEnv = self.manifest.localEnv
+        if iName not in self and localEnv:
+            project = localEnv.project or localEnv.homeProject
+            tpl = project and project.find_ensemble_by_name(iName)
+            if tpl:
+                self.manifest.load_external_ensemble(iName, dict(manifest=tpl))
+                return self._find_import(qualified_name)
+        return None
+
+    def _find_import(self, name):
         if name in self:
-            return self[name].resource
+            # fully qualified name already added
+            return self[name].local_instance or self[name].external_instance
         iName, sep, rName = name.partition(":")
         if iName not in self:
             return None
-        imported = self[iName].resource.find_resource(rName or "root")
+        # do a unqualified look up to find the declared import
+        imported = self[iName].external_instance.root.find_resource(rName or "root")
         if imported:
-            # add for future reference
-            self[name] = imported  # see __setitem__
+            self.add_import(iName, imported)
         return imported
 
-    def set_shadow(self, key, instance):
-        instance.shadow = self[key].resource
-        self[key] = instance
-
-    def __setitem__(self, key, value):
-        if isinstance(value, tuple):
-            value = _Import(*value)
+    def set_shadow(self, key, local_instance, external_instance):
+        if key not in self:
+            record = self.add_import(key, external_instance)
         else:
-            if key in self:
-                value = self[key]._replace(resource=value)
-            else:
-                value = _Import(value, {})
-        return super().__setitem__(key, value)
+            record = self[key]
+
+        if record.local_instance:
+            raise UnfurlError(f"already imported {key}")
+        record.local_instance = local_instance
+        return record
+
+    def add_import(self, key, external_instance, spec=None):
+        self[key] = _Import(external_instance, spec or {})
+        return self[key]
 
 
 class ExternalResource(ExternalValue):
@@ -668,7 +1075,7 @@ class ExternalResource(ExternalValue):
 
     def __init__(self, name, importSpec):
         super().__init__("external", name)
-        self.resource = importSpec.resource
+        self.resource = importSpec.external_instance
         self.schema = importSpec.spec.get("schema")
 
     def _validate(self, obj, schema, name):
@@ -722,7 +1129,7 @@ class SecretResource(ExternalResource):
 
 # shortcuts for local and secret
 def shortcut(arg, ctx):
-    return Ref(dict(ref=dict(external=ctx.currentFunc), foreach=arg)).resolve(
+    return Ref(dict(ref=dict(external=ctx.currentFunc), select=arg)).resolve(
         ctx, wantList="result"
     )
 
@@ -851,14 +1258,6 @@ class TopologyMap(dict):
         return len(tuple(self.resource.get_self_and_descendents()))
 
 
-def _is_sensitive_schema(defs, key, value):
-    defSchema = (key in defs and defs[key].schema) or {}
-    defMeta = defSchema.get("metadata", {})
-    # attribute marked as sensitive and value isn't a secret so mark value as sensitive
-    # but externalvalues are ok since they don't reveal much
-    return defMeta.get("sensitive") and not value.external
-
-
 class AttributeManager:
     """
     Tracks changes made to Resources
@@ -874,10 +1273,12 @@ class AttributeManager:
 
     # what about an attribute that is added to the spec that already exists in status?
     # XXX2 tests for the above behavior
-    def __init__(self, yaml=None):
+    def __init__(self, yaml=None, task=None):
         self.attributes = {}
         self.statuses = {}
         self._yaml = yaml  # hack to safely expose the yaml context
+        self.task = task
+        self._context_vars = None
 
     @property
     def yaml(self):
@@ -902,11 +1303,23 @@ class AttributeManager:
 
     def mark_referenced_templates(self, template):
         for (resource, attr) in self.attributes.values():
-            if resource.template is not template and template not in resource.template._isReferencedBy:
+            if (
+                resource.template is not template
+                and template not in resource.template._isReferencedBy
+            ):
                 resource.template._isReferencedBy.append(template)
 
+    def _get_context(self, resource):
+        if self._context_vars is None:
+            self._context_vars = {}
+            set_context_vars(self._context_vars, resource)
+        return RefContext(resource, self._context_vars, task=self.task)
+
     def get_attributes(self, resource):
         if resource.key not in self.attributes:
+            if resource.shadow:
+                return resource.shadow.attributes
+
             # deepcopy() because lazily created ResultMaps and ResultLists will mutate
             # the underlying nested structures when resolving values
             if resource.template:
@@ -918,19 +1331,17 @@ class AttributeManager:
                 )
             else:
                 _attributes = ChainMap(copy.deepcopy(resource._attributes))
-
-            vars = dict(NODES=TopologyMap(resource.root))
-            if "inputs" in resource.root._attributes:
-                vars["TOPOLOGY"] = dict(
-                    inputs=resource.root._attributes["inputs"],
-                    outputs=resource.root._attributes["outputs"],
-                )
-            ctx = RefContext(resource, vars)
-            attributes = ResultsMap(_attributes, ctx)
+            ctx = self._get_context(resource)
+            validate = True
+            mode = os.getenv("UNFURL_VALIDATION_MODE")
+            if mode is not None and "nopropcheck" in mode:
+                validate = False
+            attributes = ResultsMap(_attributes, ctx, validate=validate)
             self.attributes[resource.key] = (resource, attributes)
             return attributes
         else:
-            return self.attributes[resource.key][1]
+            attributes = self.attributes[resource.key][1]
+            return attributes
 
     # def revertChanges(self):
     #   self.attributes = {}
@@ -938,8 +1349,10 @@ class AttributeManager:
     #   #   resource._localStatus = old
 
     @staticmethod
-    def _save_sensitive(defs, key, value, instance):
-        sensitive = _is_sensitive_schema(defs, key, value)
+    def _save_sensitive(defs, key, value):
+        # attribute marked as sensitive and value isn't a secret so mark value as sensitive
+        # but externalvalues are ok since they don't reveal much
+        sensitive = is_sensitive_schema(defs, key) and not value.external
         if sensitive:
             savedValue = wrap_sensitive_value(value.resolved)
         else:
@@ -994,7 +1407,7 @@ class AttributeManager:
                     _attributes[key] = value
                 else:
                     changed, isLive = self._check_attribute(specd, key, value, resource)
-                    savedValue = self._save_sensitive(defs, key, value, resource)
+                    savedValue = self._save_sensitive(defs, key, value)
                     is_sensitive = isinstance(savedValue, sensitive)
                     # save the Result not savedValue because we need the ExternalValue
                     live[key] = (isLive, savedValue if is_sensitive else value)
@@ -1020,7 +1433,7 @@ class AttributeManager:
                 continue
             for key in foundSensitive:
                 if key in diff:
-                    diff[key] = resource._attributes[key]
+                    diff[key] = _attributes[key]
             changes[resource.key] = diff
 
         self.attributes = {}
