@@ -6,26 +6,37 @@ TOSCA implementation
 import functools
 import copy
 from .tosca_plugins import TOSCA_VERSION
-from .util import UnfurlError, UnfurlValidationError, get_base_dir, check_class_registry
+from .util import (
+    UnfurlError,
+    UnfurlValidationError,
+    get_base_dir,
+    check_class_registry,
+    env_var_value,
+)
 from .eval import Ref, RefContext, map_value
-from .result import ResourceRef, ResultsList
+from .result import ExternalValue, ResourceRef, ResultsList
 from .merge import patch_dict, merge_dicts
 from .logs import get_console_log_level
+from .support import is_template, ContainerImage
 from toscaparser.tosca_template import ToscaTemplate
+from toscaparser.entity_template import EntityTemplate
 from toscaparser.properties import Property
 from toscaparser.elements.entity_type import EntityType
 from toscaparser.elements.statefulentitytype import StatefulEntityType
 import toscaparser.workflow
 import toscaparser.imports
 import toscaparser.artifacts
+import toscaparser.repositories
 from toscaparser.common.exception import ExceptionCollector
-import six
+import os
+from .logs import getLogger
 import logging
 import re
+from typing import Dict, Optional
 
 from ruamel.yaml.comments import CommentedMap
 
-logger = logging.getLogger("unfurl")
+logger = getLogger("unfurl")
 
 from toscaparser import functions
 
@@ -45,7 +56,7 @@ toscaIsFunction = functions.is_function
 
 
 def is_function(function):
-    return toscaIsFunction(function) or Ref.is_ref(function)
+    return toscaIsFunction(function) or Ref.is_ref(function) or is_template(function)
 
 
 functions.is_function = is_function
@@ -92,6 +103,7 @@ def create_default_topology():
     )
     return ToscaTemplate(yaml_dict_tpl=tpl)
 
+
 def _patch(node, patchsrc, quote=False, tpl=None):
     if tpl is None:
         tpl = node.toscaEntityTemplate.entity_tpl
@@ -106,9 +118,11 @@ def _patch(node, patchsrc, quote=False, tpl=None):
     logger.trace("patched node %s: now %s", node.name, patched)
     return patched
 
+
 class ToscaSpec:
     InstallerType = "unfurl.nodes.Installer"
     topology = None
+    substitution_template = None
 
     def evaluate_imports(self, toscaDef):
         if not toscaDef.get("imports"):
@@ -148,8 +162,10 @@ class ToscaSpec:
                 for prop, value in req.get_nodefilter_properties():
                     # annotate the target's properties
                     target = req.relationship and req.relationship.target
-                    if target and isinstance(value, dict) and 'eval' in value:
-                        value.setdefault('vars', {})['SOURCE'] = dict(eval="::"+nodespec.name)
+                    if target and isinstance(value, dict) and "eval" in value:
+                        value.setdefault("vars", {})["SOURCE"] = dict(
+                            eval="::" + nodespec.name
+                        )
                         patch = dict(properties={prop: value})
                         _patch(target, patch, quote=True)
                         patched = True
@@ -158,7 +174,9 @@ class ToscaSpec:
                     target = req.relationship and req.relationship.target
                     if target:
                         matching_target_req = target.requirements.get(name)
-                        _patch(nodespec, value, tpl=matching_target_req.entity_tpl[name])
+                        _patch(
+                            nodespec, value, tpl=matching_target_req.entity_tpl[name]
+                        )
                         patched = True
 
         return patched
@@ -191,6 +209,12 @@ class ToscaSpec:
 
     def _parse_template(self, path, inputs, toscaDef, resolver):
         # need to set a path for the import loader
+        mode = os.getenv("UNFURL_VALIDATION_MODE")
+        additionalProperties = False
+        if mode is not None:
+            additionalProperties = "additionalProperties" in mode
+            ToscaTemplate.strict = "reqcheck" in mode
+        EntityTemplate.additionalProperties = additionalProperties
         self.template = ToscaTemplate(
             path=path,
             parsed_params=inputs,
@@ -198,14 +222,14 @@ class ToscaSpec:
             import_resolver=resolver,
             verify=False,  # we display the error messages ourselves so we don't need to verify here
         )
-        ExceptionCollector.collecting = True # don't stop collecting validation errors
-        ExceptionCollector.near = ' while instantiating the spec'
+        ExceptionCollector.collecting = True  # don't stop collecting validation errors
+        ExceptionCollector.near = " while instantiating the spec"
         self.nodeTemplates = {}
         self.installers = {}
         self.relationshipTemplates = {}
         for template in self.template.nodetemplates:
             if not template.type_definition:
-                continue # invalidate template
+                continue  # invalidate template
             nodeTemplate = NodeSpec(template, self)
             if template.is_derived_from(self.InstallerType):
                 self.installers[template.name] = nodeTemplate
@@ -219,9 +243,9 @@ class ToscaSpec:
         self.topology = TopologySpec(self, inputs)
         substitution_mappings = self.template.topology_template.substitution_mappings
         if substitution_mappings and substitution_mappings.node:
-            self.substitution_template = self.nodeTemplates.get(substitution_mappings.node)
-        else:
-            self.substitution_template = None
+            self.substitution_template = self.nodeTemplates.get(
+                substitution_mappings.node
+            )
         self.load_workflows()
         self.groups = {
             g.name: GroupSpec(g, self) for g in self.template.topology_template.groups
@@ -232,15 +256,13 @@ class ToscaSpec:
         }
         ExceptionCollector.collecting = False
 
-    def _patch(self, toscaDef, path):
+    def _patch(self, toscaDef, path, errorsSoFar):
         matches = None
         decorators = self.load_decorators()
         if decorators:
             logger.debug("applying decorators %s", decorators)
-            # copy errors before we clear them in _overlay
-            errorsSoFar = ExceptionCollector.exceptions[:]
-            # overlay uses ExceptionCollector
             matches = self._overlay(decorators)
+            # overlay uses ExceptionCollector
             if ExceptionCollector.exceptionsCaught():
                 # abort if overlay caused errors
                 # report previously collected errors too
@@ -258,7 +280,6 @@ class ToscaSpec:
         annotated = self.enforce_filters()
         return matches or modified_imports or annotated
 
-
     def __init__(
         self, toscaDef, spec=None, path=None, resolver=None, skip_validation=False
     ):
@@ -285,13 +306,21 @@ class ToscaSpec:
             try:
                 self._parse_template(path, inputs, toscaDef, resolver)
             except:
-                if not ExceptionCollector.exceptionsCaught() or not self.template or not self.topology:
+                if (
+                    not ExceptionCollector.exceptionsCaught()
+                    or not self.template
+                    or not self.topology
+                ):
                     raise  # unexpected error
 
-            patched = self._patch(toscaDef, path)
+            # copy errors because self._patch() might clear them
+            errorsSoFar = ExceptionCollector.exceptions[:]
+            patched = self._patch(toscaDef, path, errorsSoFar)
             if patched:
                 # overlay and evaluate_imports modifies tosaDef in-place, try reparsing it
                 self._parse_template(path, inputs, toscaDef, resolver)
+            else:  # restore previously errors
+                ExceptionCollector.exceptions[:0] = errorsSoFar
 
             if ExceptionCollector.exceptionsCaught():
                 message = "\n".join(
@@ -492,12 +521,16 @@ class ToscaSpec:
                 )
 
         for name, impl in tpl.get("instances", {}).items():
-            if name not in node_templates and isinstance(impl, dict):
-                # add this as a template
-                if "template" not in impl:
-                    node_templates[name] = self.instance_to_template(impl.copy())
-                elif isinstance(impl["template"], dict):
-                    node_templates[name] = impl["template"]
+            if not isinstance(impl, dict):
+                continue
+            if name in node_templates:
+                if "default" not in node_templates[name].get("directives", []):
+                    continue  # allow default templates to be overridden
+            # add this as a template
+            if "template" not in impl:
+                node_templates[name] = self.instance_to_template(impl.copy())
+            elif isinstance(impl["template"], dict):
+                node_templates[name] = impl["template"]
 
         if "discovered" in tpl:
             # node templates added dynamically by configurators
@@ -531,58 +564,74 @@ class ToscaSpec:
                 self.relationshipTemplates[template.name] = relTemplate
 
 
-def find_props(attributes, propertyDefs, matchfn):
+def find_env_vars(props_iter):
+    for propdef, value in props_iter:
+        datatype = propdef.entity
+        if datatype.type == "unfurl.datatypes.EnvVar":
+            yield propdef.name, env_var_value(value)
+        elif (
+            datatype.type == "map"
+            and propdef.entry_schema_entity
+            and propdef.entry_schema_entity.type == "unfurl.datatypes.EnvVar"
+        ):
+            if value:
+                for key, item in value.items():
+                    yield key, env_var_value(item)
+        else:
+            metadata = propdef.schema.get("metadata", {})
+            if metadata.get("env_vars"):
+                for name in metadata["env_vars"]:
+                    yield name, env_var_value(value)
+
+
+def find_props(attributes, propertyDefs, flatten=False):
     if not attributes:
         return
     for propdef in propertyDefs.values():
         if propdef.name not in attributes:
             continue
-        match = matchfn(propdef.entry_schema_entity or propdef.entity)
-        if not propdef.entry_schema and not propdef.entity.properties:
-            # it's a simple value type
-            if match:
-                yield propdef.name, attributes[propdef.name]
-            continue
-
-        if not propdef.entry_schema:
+        if propdef.entity.properties:
             # it's complex datatype
             value = attributes[propdef.name]
-            if match:
-                yield propdef.name, value
-            elif value:
+            if value:
                 # descend into its properties
-                for name, v in find_props(value, propdef.entity.properties, matchfn):
-                    yield name, v
-            continue
-
-        properties = propdef.entry_schema_entity.properties
-        if not match and not properties:
-            # entries are simple value types and didn't match
-            continue
-
-        value = attributes[propdef.name]
-        if not value:
-            continue
-        if propdef.type == "map":
-            for key, val in value.items():
-                if match:
-                    yield key, val
-                elif properties:
-                    for name, v in find_props(val, properties, matchfn):
-                        yield name, v
-        elif propdef.type == "list":
-            for val in value:
-                if match:
-                    yield None, val
-                elif properties:
-                    for name, v in find_props(val, properties, matchfn):
-                        yield name, v
+                yield from find_props(value, propdef.entity.properties, flatten)
+            else:
+                yield propdef, value
+        elif not flatten or not propdef.entry_schema:
+            yield propdef, attributes[propdef.name]
+        else:
+            # its a list or map
+            assert propdef.entry_schema
+            properties = propdef.entry_schema_entity.properties
+            value = attributes[propdef.name]
+            if not value:
+                yield propdef, value
+                continue
+            if propdef.type == "map":
+                for key, val in value.items():
+                    if properties:
+                        # descend into its properties
+                        yield from find_props(val, properties, flatten)
+                    else:
+                        yield propdef, (key, val)
+            elif propdef.type == "list":
+                for val in value:
+                    if properties:
+                        # descend into its properties
+                        yield from find_props(val, properties, flatten)
+                    else:
+                        yield propdef, val
 
 
 # represents a node, capability or relationship
 class EntitySpec(ResourceRef):
     # XXX need to define __eq__ for spec changes
-    def __init__(self, toscaNodeTemplate, spec=None):
+    def __init__(self, toscaNodeTemplate=None, spec=None):
+        if not toscaNodeTemplate:
+            toscaNodeTemplate = next(
+                iter(create_default_topology().topology_template.nodetemplates)
+            )
         self.toscaEntityTemplate = toscaNodeTemplate
         self.spec = spec
         self.name = toscaNodeTemplate.name
@@ -595,7 +644,9 @@ class EntitySpec(ResourceRef):
             )
 
         self.type = toscaNodeTemplate.type
-        self._isReferencedBy = [] # this is referenced by another template or via property traversal
+        self._isReferencedBy = (
+            []
+        )  # this is referenced by another template or via property traversal
         # nodes have both properties and attributes
         # as do capability properties and relationships
         # but only property values are declared
@@ -646,11 +697,6 @@ class EntitySpec(ResourceRef):
     def get_interfaces(self):
         return self.toscaEntityTemplate.interfaces
 
-    def get_interface_requirements(self):
-        return self.toscaEntityTemplate.type_definition.get_interface_requirements(
-            self.toscaEntityTemplate.entity_tpl
-        )
-
     @property
     def groups(self):
         if not self.spec:
@@ -699,7 +745,7 @@ class EntitySpec(ResourceRef):
     def find_or_create_artifact(self, nameOrTpl, path=None, predefined=False):
         if not nameOrTpl:
             return None
-        if isinstance(nameOrTpl, six.string_types):
+        if isinstance(nameOrTpl, str):
             name = nameOrTpl
             artifact = self.artifacts.get(nameOrTpl)
             if artifact:
@@ -751,9 +797,12 @@ class EntitySpec(ResourceRef):
     def directives(self):
         return []
 
-    def find_props(self, attributes, matchfn):
-        for name, val in find_props(attributes, self.propertyDefs, matchfn):
-            yield name, val
+    @property
+    def tpl(self):
+        return self.toscaEntityTemplate.entity_tpl
+
+    def find_props(self, attributes):
+        yield from find_props(attributes, self.propertyDefs)
 
     @property
     def base_dir(self):
@@ -774,19 +823,26 @@ class EntitySpec(ResourceRef):
         # no implementations found
         return True
 
+    def validate(self) -> None:
+        """
+        Raises UnfurlValidationError on failure.
+        """
+        pass
+
     @property
     def required(self):
-        # if this template is required by another template
+        # check if this template is required by another template
         for root in _get_roots(self):
             if self.spec.substitution_template:
                 if self.spec.substitution_template is root:
-                    # if don't require if a root is the substitution_mappings
+                    # require if a root of this template is the substitution_template
                     return True
-            elif 'default' not in root.directives:
-                # if don't require if this only has defaults templates as a root
+            elif "default" not in root.directives:
+                # otherwise require when there is a root that isn't a defaults template
                 return True
         return False
 
+
 def _get_roots(node, seen=None):
     # node can reference each other's properties, so we need to handle circular references
     if seen is None:
@@ -794,9 +850,10 @@ def _get_roots(node, seen=None):
     yield node
     for parent in node._isReferencedBy:
         if parent.name not in seen:
-            seen.add( node.name )
+            seen.add(node.name)
             yield from _get_roots(parent, seen)
 
+
 class NodeSpec(EntitySpec):
     # has attributes: tosca_id, tosca_name, state, (3.4.1 Node States p.61)
     def __init__(self, template=None, spec=None):
@@ -827,6 +884,11 @@ class NodeSpec(EntitySpec):
             ]
             return relationship
 
+    def get_interface_requirements(self):
+        return self.toscaEntityTemplate.type_definition.get_interface_requirements(
+            self.toscaEntityTemplate.entity_tpl
+        )
+
     @property
     def artifacts(self):
         if self._artifacts is None:
@@ -850,7 +912,7 @@ class NodeSpec(EntitySpec):
                     yield p
 
     @property
-    def requirements(self):
+    def requirements(self) -> Dict[str, "RequirementSpec"]:
         if self._requirements is None:
             self._requirements = {}
             nodeTemplate = self.toscaEntityTemplate
@@ -928,9 +990,20 @@ class NodeSpec(EntitySpec):
                 relSpec.toscaEntityTemplate.source.name
                 == reqSpec.parentNode.toscaEntityTemplate.name
             ):
-                assert not reqSpec.relationship or reqSpec.relationship is relSpec, (reqSpec.relationship, relSpec)
+                assert (
+                    not reqSpec.relationship
+                    or reqSpec.relationship.name == relSpec.name
+                ), (
+                    reqSpec.relationship,
+                    relSpec,
+                )
                 reqSpec.relationship = relSpec
-                assert not relSpec.requirement or relSpec.requirement is reqSpec, (relSpec.requirement, reqSpec)
+                assert (
+                    not relSpec.requirement or relSpec.requirement.name == reqSpec.name
+                ), (
+                    relSpec.requirement,
+                    reqSpec,
+                )
                 if not relSpec.requirement:
                     relSpec.requirement = reqSpec
                 break
@@ -940,6 +1013,8 @@ class NodeSpec(EntitySpec):
 
     @property
     def abstract(self):
+        if self.tpl.get("imported"):
+            return "select"
         for name in ("select", "substitute"):
             if name in self.toscaEntityTemplate.directives:
                 return name
@@ -949,6 +1024,14 @@ class NodeSpec(EntitySpec):
     def directives(self):
         return self.toscaEntityTemplate.directives
 
+    def validate(self):
+        super().validate()
+        missing = self.toscaEntityTemplate.missing_requirements
+        if missing:
+            raise UnfurlValidationError(
+                f"Node template {self.name} is missing requirements: {','.join(missing)}"
+            )
+
 
 class RelationshipSpec(EntitySpec):
     """
@@ -1014,7 +1097,9 @@ class RelationshipSpec(EntitySpec):
         if not defaultFor:
             return False
         nodeTemplate = capability.parentNode.toscaEntityTemplate
-
+        if defaultFor == self.toscaEntityTemplate.ANY and capability.name == "feature":
+            # XXX get_matching_capabilities() buggy in this case
+            return True  # optimization
         if (
             defaultFor == self.toscaEntityTemplate.ANY
             or defaultFor == nodeTemplate.name
@@ -1048,7 +1133,7 @@ class RequirementSpec:
         # occurrences
 
     def __repr__(self):
-        return f"{self.__class__.__name__}('{self.name}')"
+        return f"{self.__class__.__name__}('{self.name}'):{self.entity_tpl}"
 
     @property
     def artifacts(self):
@@ -1062,21 +1147,23 @@ class RequirementSpec:
 
     def get_nodefilter_properties(self):
         # XXX should merge type_tpl with entity_tpl
-        return get_nodefilters(self.type_tpl, 'properties')
+        return get_nodefilters(self.type_tpl, "properties")
 
     def get_nodefilter_requirements(self):
         # XXX should merge type_tpl with entity_tpl
         return get_nodefilters(self.type_tpl, "requirements")
 
+
 def get_nodefilters(entity_tpl, key):
     if not isinstance(entity_tpl, dict):
         return
-    nodefilter = entity_tpl.get('node_filter')
+    nodefilter = entity_tpl.get("node_filter")
     if nodefilter and key in nodefilter:
         for filter in nodefilter[key]:
             name, value = next(iter(filter.items()))
             yield name, value
 
+
 class CapabilitySpec(EntitySpec):
     def __init__(self, parent=None, capability=None):
         if not parent:
@@ -1202,6 +1289,10 @@ class TopologySpec(EntitySpec):
                 raise KeyError(key)
             return matches
 
+    @property
+    def tpl(self):
+        return self.toscaEntityTemplate.tpl
+
 
 class Workflow:
     def __init__(self, workflow):
@@ -1268,7 +1359,7 @@ class ArtifactSpec(EntitySpec):
                 name, artifact_tpl, custom_defs, path
             )
         EntitySpec.__init__(self, artifact, spec)
-        self.repository = (
+        self.repository: Optional[toscaparser.repositories.Repository] = (
             spec
             and artifact.repository
             and spec.template.repositories.get(artifact.repository)
@@ -1317,6 +1408,20 @@ class ArtifactSpec(EntitySpec):
     def as_import_spec(self):
         return dict(file=self.file, repository=self.toscaEntityTemplate.repository)
 
+    def as_value(self) -> Optional[ExternalValue]:
+        if self.is_compatible_type("tosca.artifacts.Deployment.Image.Container.Docker"):
+            artifactDef = self.toscaEntityTemplate
+            assert not artifactDef.checksum or artifactDef.checksum_algorithm == 256
+            kw = dict(tag=self.properties.get("tag"), digest=artifactDef.checksum)
+            if self.repository:
+                kw["registry_host"] = self.repository.hostname
+                if self.repository.credential:
+                    kw["username"] = self.repository.credential.get("user")
+                    kw["password"] = self.repository.credential.get("token")
+            return ContainerImage(artifactDef.file, **kw)
+        # XXX return File or FilePath
+        return None
+
 
 class GroupSpec(EntitySpec):
     def __init__(self, template, spec):
