@@ -2,10 +2,24 @@
 # SPDX-License-Identifier: MIT
 import sys
 from types import ModuleType
-from typing import IO, Dict, Generator, Iterable, List, Optional, Set, Tuple, Union, Iterator, TYPE_CHECKING, cast
+from typing import (
+    IO,
+    Any,
+    Dict,
+    Generator,
+    Iterable,
+    List,
+    MutableMapping,
+    Optional,
+    Set,
+    Tuple,
+    Union,
+    Iterator,
+    TYPE_CHECKING,
+    cast,
+)
 from typing_extensions import SupportsIndex
 import git
-import six
 import traceback
 import itertools
 import tempfile
@@ -20,7 +34,6 @@ import os.path
 from jsonschema import Draft7Validator, validators, RefResolver
 import jsonschema.exceptions
 from ruamel.yaml.scalarstring import ScalarString, FoldedScalarString
-from ansible.parsing.yaml.objects import AnsibleVaultEncryptedUnicode
 from ansible.parsing.vault import VaultEditor
 from ansible.module_utils._text import to_text, to_bytes, to_native  # BSD licensed
 from ansible.utils.unsafe_proxy import AnsibleUnsafeText, AnsibleUnsafeBytes, wrap_var
@@ -28,7 +41,6 @@ import warnings
 import codecs
 import io
 from contextlib import contextmanager
-from logging import Logger
 import string
 import random
 
@@ -51,7 +63,7 @@ try:
     imp = None
 except ImportError:
     import imp  # type: ignore
-from .logs import UnfurlLogger, sensitive
+from .logs import LogExtraLevels, sensitive, is_sensitive
 import logging
 
 # Used for python typing, prevents circular imports
@@ -71,6 +83,7 @@ except ImportError:
 
 _basepath = os.path.abspath(os.path.dirname(__file__))
 
+
 def get_package_digest() -> Union[git.Repo, str, object]:
     from git import Repo
 
@@ -86,7 +99,9 @@ def get_package_digest() -> Union[git.Repo, str, object]:
 
 
 class UnfurlError(Exception):
-    def __init__(self, message: object, saveStack: bool=False, log: bool=False) -> None:
+    def __init__(
+        self, message: object, saveStack: bool = False, log: bool = False
+    ) -> None:
         if saveStack:
             (etype, value, traceback) = sys.exc_info()
             if value:
@@ -103,42 +118,57 @@ class UnfurlError(Exception):
 
 
 class UnfurlValidationError(UnfurlError):
-    def __init__(self, message: object, errors: Optional[List[Exception]]=None, log: bool=False) -> None:
+    def __init__(
+        self,
+        message: object,
+        errors: Optional[List[Exception]] = None,
+        log: bool = False,
+    ) -> None:
         super().__init__(message, log=log)
         self.errors = errors or []
 
 
 class UnfurlTaskError(UnfurlError):
-    def __init__(self, task: "TaskView", message: object, log: int=logging.ERROR):
+    def __init__(
+        self,
+        task: "TaskView",
+        message: object,
+        log: int = logging.ERROR,
+        dependency=None,
+    ):
         task = cast("ConfigTask", task)
         message = f"{task.changeId} on {task.target.name} {task.name}: {message}"
         super().__init__(message, True, False)
         self.task = task
         task._errors.append(self)
         self.severity = log
+        self.dependency = dependency  # what caused the error
         if log:
             task.logger.log(log, message, exc_info=True)
 
 
 class UnfurlAddingResourceError(UnfurlTaskError):
-    def __init__(self, task: "TaskView", resourceSpec: Mapping, name: str, log: int=logging.DEBUG) -> None:
+    def __init__(
+        self,
+        task: "TaskView",
+        resourceSpec: Mapping,
+        name: str,
+        log: int = logging.DEBUG,
+    ) -> None:
         message = f"error updating resource {name}"
         super().__init__(task, message, log)
         self.resourceSpec = resourceSpec
 
 
-def wrap_sensitive_value(obj: object) -> object:
-    # we don't remember the vault and vault id associated with this value
+def wrap_sensitive_value(obj: object):
+    # convert to sensitive obj if possible or return the original object
+    # Note: we don't remember the vault and vault id associated with this value
     # so the value will be rekeyed with whichever vault is associated with the serializing yaml
-    if isinstance(obj, bytes):
-        if six.PY3:
-            return sensitive_bytes(obj)
-        else:
-            try:
-                return sensitive_str(unicode(obj, "utf-8"))
-            except UnicodeDecodeError:
-                return sensitive_bytes(obj)
-    elif isinstance(obj, six.string_types):
+    if isinstance(obj, sensitive):
+        return obj
+    elif isinstance(obj, bytes):
+        return sensitive_bytes(obj)
+    elif isinstance(obj, str):
         return sensitive_str(obj)
     elif isinstance(obj, Mapping):
         return sensitive_dict(obj)
@@ -148,19 +178,6 @@ def wrap_sensitive_value(obj: object) -> object:
         return obj
 
 
-def is_sensitive(obj: object) -> bool:
-    test = getattr(obj, "__sensitive__", None)
-    if test:
-        return test()
-    if isinstance(obj, AnsibleVaultEncryptedUnicode):
-        return True
-    elif isinstance(obj, Mapping):
-        return any(is_sensitive(i) for i in obj.values())
-    elif isinstance(obj, MutableSequence):
-        return any(is_sensitive(i) for i in obj)
-    return False
-
-
 # also mark AnsibleUnsafe so wrap_var doesn't strip out sensitive status
 class sensitive_bytes(AnsibleUnsafeBytes, sensitive):
     """Transparent wrapper class to mark bytes as sensitive"""
@@ -187,19 +204,21 @@ class sensitive_list(list, sensitive):
     """Transparent wrapper class to mark a list as sensitive"""
 
 
-def to_yaml_text(val: object) -> Union[sensitive, ScalarString, FoldedScalarString, str]:
+def to_yaml_text(
+    val: object,
+) -> Union[sensitive, ScalarString, FoldedScalarString, str]:
     if isinstance(val, (ScalarString, sensitive)):
         return val
     # convert or copy string (copy to deal with things like AnsibleUnsafeText)
     val = str(val)
     if "\n" in val:
-        if six.PY2 and isinstance(val, str):
-            val = val.decode("utf-8")
         return FoldedScalarString(val)
     return val
 
 
-def assert_form(src: object, types: Union[type, Tuple[type, ...]]=Mapping, test: bool=True) -> object:
+def assert_form(
+    src: object, types: Union[type, Tuple[type, ...]] = Mapping, test: bool = True
+) -> object:
     if not isinstance(src, types) or not test:
         raise UnfurlError(f"Malformed definition: {src}")
     return src
@@ -213,7 +232,9 @@ def register_short_names(shortNames: Union[Mapping, Iterable]) -> None:
     _shortNameRegistry.update(shortNames)
 
 
-def register_class(className: str, factory: object, short_name: str=None, replace: bool=True) -> None:
+def register_class(
+    className: str, factory: object, short_name: str = None, replace: bool = True
+) -> None:
     if short_name:
         _shortNameRegistry[short_name] = className
     if not replace and className in _ClassRegistry:
@@ -222,7 +243,7 @@ def register_class(className: str, factory: object, short_name: str=None, replac
     _ClassRegistry[className] = factory
 
 
-def load_module(path: str, full_name: str=None) -> ModuleType:
+def load_module(path: str, full_name: str = None) -> ModuleType:
     if full_name is None:
         full_name = re.sub(r"\W", "_", path)  # generate a name from the path
     if full_name in sys.modules:
@@ -247,7 +268,7 @@ def load_module(path: str, full_name: str=None) -> ModuleType:
     return module
 
 
-def load_class(klass: str, defaultModule: str="__main__") -> object:
+def load_class(klass: str, defaultModule: str = "__main__") -> object:
     import importlib
 
     prefix, sep, suffix = klass.rpartition(".")
@@ -261,8 +282,11 @@ _shortNameRegistry = {}
 def check_class_registry(kind: str) -> bool:
     return kind in _ClassRegistry or kind in _shortNameRegistry
 
+
 # XXX: can't infer arg "default"'s type for now
-def lookup_class(kind: str, apiVersion: Optional[str]=None, default: Optional[str]=None) -> object:
+def lookup_class(
+    kind: str, apiVersion: Optional[str] = None, default: Optional[str] = None
+) -> object:
     if kind in _ClassRegistry:
         return _ClassRegistry[kind]
     elif kind in _shortNameRegistry:
@@ -281,7 +305,7 @@ def lookup_class(kind: str, apiVersion: Optional[str]=None, default: Optional[st
 
 def to_enum(enum, value, default=None):  # type: ignore
     # from string: Status[name]; to string: status.name
-    if isinstance(value, six.string_types):
+    if isinstance(value, str):
         return enum[value]
     elif default is not None and not value:
         return default
@@ -305,40 +329,37 @@ def to_dotenv(env: dict) -> str:
     return "\n".join(keys)
 
 
-def dump(obj: object, tp: IO[bytes], suffix: str="", yaml: Optional[ModuleType]=None, encoding: Optional[str]=None) -> None:
+def dump(
+    obj: object,
+    tp: IO[bytes],
+    suffix: str = "",
+    yaml: Optional[ModuleType] = None,
+    encoding: Optional[str] = None,
+) -> None:
     from .yamlloader import yaml as _yaml
 
     try:
-        if six.PY3:
-            textEncoding = (
-                "utf-8"
-                if encoding in [None, "yaml", "vault", "json", "env"]
-                else encoding
-            )
-            f = io.TextIOWrapper(tp, textEncoding)
-        else:
-            f = tp
-        if suffix.endswith(".yml") or suffix.endswith(".yaml") or encoding == "yaml":
+        textEncoding = (
+            "utf-8" if encoding in [None, "yaml", "vault", "json", "env"] else encoding
+        )
+        f = io.TextIOWrapper(tp, textEncoding)
+        explicit = encoding is not None
+        if encoding == "yaml" or (
+            not explicit and (suffix.endswith(".yml") or suffix.endswith(".yaml"))
+        ):
             (yaml or _yaml).dump(obj, f)
             return
-        elif suffix.endswith(".json") or encoding == "json":
+        elif encoding == "json" or (not explicit and suffix.endswith(".json")):
             json.dump(obj, f, indent=2)  # str to bytes
             return
 
-        if six.PY3:
-            if isinstance(obj, str):
-                obj = codecs.encode(obj, encoding or "utf-8")
-            if isinstance(obj, bytes):
-                tp.write(obj)
-                return
-        else:
-            if isinstance(obj, unicode):
-                obj = codecs.encode(obj, encoding or "utf-8")
-            if isinstance(obj, str):
-                tp.write(obj)
-                return
+        if isinstance(obj, str):
+            obj = codecs.encode(obj, encoding or "utf-8")
+        if isinstance(obj, bytes):
+            tp.write(obj)
+            return
 
-        if suffix.endswith(".env") or encoding == "env" and isinstance(obj, dict):
+        if (suffix.endswith(".env") or encoding == "env") and isinstance(obj, dict):
             tp.write(codecs.encode(to_dotenv(obj), "utf-8"))  # type: ignore
             return
 
@@ -346,11 +367,16 @@ def dump(obj: object, tp: IO[bytes], suffix: str="", yaml: Optional[ModuleType]=
         if obj is not None:  # treat None as 0 byte file
             json.dump(obj, f, indent=2, sort_keys=True)
     finally:
-        if six.PY3:
-            f.detach()
+        f.detach()
 
 
-def _save_to_vault(path: str, obj: object, yaml: Optional[ModuleType]=None, encoding: Optional[str]=None, fd: Union[str, int]=None) -> bool:
+def _save_to_vault(
+    path: str,
+    obj: object,
+    yaml: Optional[ModuleType] = None,
+    encoding: Optional[str] = None,
+    fd: Union[str, int] = None,
+) -> bool:
     vaultExt = path.endswith(".vault")
     if vaultExt or encoding == "vault":
         assert yaml and yaml.representer.vault and yaml.representer.vault.secrets
@@ -358,7 +384,7 @@ def _save_to_vault(path: str, obj: object, yaml: Optional[ModuleType]=None, enco
         vault = VaultEditor(vaultlib)
         f = io.BytesIO()
         vpath = path[: -len(".vault")] if vaultExt else path
-        dump(obj, f, vpath, yaml, encoding)
+        dump(obj, f, vpath, yaml, None if encoding == "vault" else encoding)
         # the first vaultid is the most specific to the current project so encrypt with that one
         vault_id, secret = vaultlib.secrets[0]
         b_vaulttext = vaultlib.encrypt(f.getvalue(), secret=secret, vault_id=vault_id)
@@ -367,7 +393,12 @@ def _save_to_vault(path: str, obj: object, yaml: Optional[ModuleType]=None, enco
     return False
 
 
-def save_to_file(path: str, obj: object, yaml: Optional[ModuleType]=None, encoding: Optional[str]=None) -> None:
+def save_to_file(
+    path: str,
+    obj: object,
+    yaml: Optional[ModuleType] = None,
+    encoding: Optional[str] = None,
+) -> None:
     dir = os.path.dirname(path)
     if dir and not os.path.isdir(dir):
         os.makedirs(dir)
@@ -376,13 +407,14 @@ def save_to_file(path: str, obj: object, yaml: Optional[ModuleType]=None, encodi
             dump(obj, f, path, yaml, encoding)
 
 
-def save_to_tempfile(obj: object,
-        suffix: str="",
-        delete: bool=True,
-        dir: Optional[str]=None,
-        yaml: Optional[ModuleType]=None,
-        encoding: Optional[str]=None
-    ) -> tempfile._TemporaryFileWrapper:
+def save_to_tempfile(
+    obj: object,
+    suffix: str = "",
+    delete: bool = True,
+    dir: Optional[str] = None,
+    yaml: Optional[ModuleType] = None,
+    encoding: Optional[str] = None,
+) -> tempfile._TemporaryFileWrapper:
     tp = tempfile.NamedTemporaryFile(
         "w+b",
         suffix=suffix or "",
@@ -400,7 +432,7 @@ def save_to_tempfile(obj: object,
     return tp
 
 
-def make_temp_dir(delete: bool=True, prefix: str="unfurl") -> str:
+def make_temp_dir(delete: bool = True, prefix: str = "unfurl") -> str:
     tempDir = tempfile.mkdtemp(prefix, dir=os.environ.get("UNFURL_TMPDIR"))
     if delete:
         atexit.register(lambda: os.path.isdir(tempDir) and shutil.rmtree(tempDir))  # type: ignore
@@ -417,24 +449,27 @@ def get_base_dir(path: str) -> str:
     else:
         return os.path.normpath(os.path.dirname(path))
 
+
 def truncate_str(v: str) -> str:
     s = str(v)
     if len(s) > 1000:
         return f"{s[:494]} [{len(s) - 1000} omitted...]  {s[-494:]}"
     return v
 
+
 def get_random_password(count=12, prefix="uv", extra=None):
     srandom = random.SystemRandom()
-    start = string.ascii_letters + string.digits
+    start = string.ascii_letters
     if extra is None:
-        extra = "%&()*+,-./:<>?=@^_`~"
+        extra = "%()*+,-./<>?=@^_~"
     source = string.ascii_letters + string.digits + extra
     return prefix + "".join(
         srandom.choice(source if i else start) for i in range(count)
     )
 
+
 @contextmanager
-def change_cwd(new_path: str, log: UnfurlLogger = None) -> Iterator:
+def change_cwd(new_path: str, log: LogExtraLevels = None) -> Iterator:
     """Temporarily change current working directory"""
     old_path = os.getcwd()
     if new_path:
@@ -463,7 +498,12 @@ def extend_with_default(validator_class: Draft7Validator) -> None:
     """
     validate_properties = validator_class.VALIDATORS["properties"]
 
-    def set_defaults(validator: Draft7Validator, properties: Mapping, instance: Draft7Validator, schema: Mapping) -> Union[Generator, Draft7Validator]:
+    def set_defaults(
+        validator: Draft7Validator,
+        properties: Mapping,
+        instance: Draft7Validator,
+        schema: Mapping,
+    ) -> Union[Generator, Draft7Validator]:
         if not validator.is_type(instance, "object"):
             return
 
@@ -483,11 +523,13 @@ DefaultValidatingLatestDraftValidator = (
 )
 
 
-def validate_schema(obj: Mapping, schema: Mapping, baseUri: Optional[str]=None) -> bool:
+def validate_schema(obj: Any, schema: Mapping, baseUri: Optional[str] = None) -> bool:
     return not find_schema_errors(obj, schema)
 
 
-def find_schema_errors(obj: Mapping, schema: Mapping, baseUri: Optional[str]=None) -> Optional[Tuple[str, List[object]]]:
+def find_schema_errors(
+    obj: Any, schema: Mapping, baseUri: Optional[str] = None
+) -> Optional[Tuple[str, List[object]]]:
     # XXX2 have option that includes definitions from manifest's schema
     if baseUri is not None:
         resolver = RefResolver(base_uri=baseUri, referrer=schema)
@@ -506,17 +548,15 @@ def find_schema_errors(obj: Mapping, schema: Mapping, baseUri: Optional[str]=Non
     return message, errors
 
 
-class ChainMap(Mapping):
+class ChainMap(MutableMapping):
     """
     Combine multiple mappings for sequential lookup.
     """
 
-    def __init__(self, *maps: List[object], **kw: Mapping) -> None:
+    def __init__(self, *maps: MutableMapping) -> None:
         self._maps = maps
-        self._track = kw.get("track")
-        self.accessed: Set[object] = set()
 
-    def split(self) -> Tuple[object, "ChainMap"]:
+    def split(self) -> Tuple[MutableMapping, "ChainMap"]:
         return self._maps[0], ChainMap(*self._maps[1:])
 
     def __getitem__(self, key: Union[SupportsIndex, slice]) -> object:
@@ -525,14 +565,14 @@ class ChainMap(Mapping):
                 return mapping[key]
             except KeyError:
                 pass
-            else:
-                if self._track is mapping:
-                    self.accessed.add(key)
         raise KeyError(key)
 
-    def __setitem__(self, key: SupportsIndex, value: object) -> None:
+    def __setitem__(self, key, value: object) -> None:
         self._maps[0][key] = value
 
+    def __delitem__(self, key) -> None:
+        del self._maps[0][key]
+
     def __iter__(self) -> Iterator:
         return iter(frozenset(itertools.chain(*self._maps)))
 
@@ -570,16 +610,50 @@ class Generate:
             return False
 
 
-def _env_var_value(val):
+def substitute_env(contents, env=None):
+    """
+    Replace ${NAME} or ${NAME:default value} with the value of the environment variable $NAME
+    Use \${NAME} to ignore
+    """
+    if env is None:
+        env = os.environ
+
+    def replace(m):
+        if m.group(1):  # \ found
+            return m.group(0)[len(m.group(1)) - 1 or 1 :]
+        default_value = m.group(3)[1:] if m.group(3) else ""
+        for name in m.group(2).split("|"):
+            if name in env:
+                value = env[name]
+                if callable(value):
+                    return value()
+                return value
+        return default_value
+
+    return re.sub(r"(\\+)?\$\{([\w|]+)(\:.+?)?\}", replace, contents)
+
+
+def env_var_value(val, sub=None):
+    if val is None:
+        return None
     if isinstance(val, bool):
         return val and "true" or ""
-    elif is_sensitive(val):
-        return wrap_sensitive_value(str(val))
+    sensitive = is_sensitive(val)
+    val = str(val)
+    if sub is not None:
+        val = substitute_env(val, sub)
+    if sensitive:
+        return wrap_sensitive_value(val)
     else:
-        return str(val)
+        return val
 
 
-def filter_env(rules: Mapping, env: Optional[Union[Dict, "os._Environ[str]"]]=None, addOnly: Optional[bool]=False) -> Dict[str, str]:
+def filter_env(
+    rules: Mapping,
+    env: Optional[Union[Dict, "os._Environ[str]"]] = None,
+    addOnly: bool = False,
+    sub: Optional[MutableMapping] = None,
+) -> Dict[str, str]:
     """
     Applies the given list of rules to a dictionary of environment variables and returns a new dictionary.
 
@@ -606,10 +680,15 @@ def filter_env(rules: Mapping, env: Optional[Union[Dict, "os._Environ[str]"]]=No
         start: Dict[str, str] = {}
     else:
         start = env.copy()
+    if sub is not None:
+        # include start so substitutions can refer to the new entries being defined
+        sub = ChainMap(start, sub)
+
     for name, val in rules.items():
         if name[:1] in "+-^":
             # add or remove from env
-            remove = name[1] == "-"
+            remove = name[0] == "-"
+            prepend = name[0] == "^"
             name = name[1:]
             neg = name[:1] == "!"
             if neg:
@@ -618,9 +697,7 @@ def filter_env(rules: Mapping, env: Optional[Union[Dict, "os._Environ[str]"]]=No
                 source = start  # if remove, look in the current set, not original
             else:
                 source = env  # type: ignore
-            match = {
-                k: v for k, v in source.items() if fnmatch.fnmatchcase(k, name) ^ neg
-            }
+            match = {k for k in source if fnmatch.fnmatchcase(k, name) ^ neg}
             if remove:
                 [
                     start.pop(k, None)
@@ -630,14 +707,16 @@ def filter_env(rules: Mapping, env: Optional[Union[Dict, "os._Environ[str]"]]=No
                 ]
             else:
                 if match:
-                    if name[1] == "^" and val:
+                    if prepend and val:
                         # treat as PATH-like and prepend val
-                        match = {k: (val + os.pathsep + v) for k, v in match.items()}
-                    start.update(match)
-                elif val is not None:  # name isn't in existing, use default is set
-                    start[name] = _env_var_value(val)
+                        update = {k: (val + os.pathsep + source[k]) for k in match}
+                    else:
+                        update = {k: source[k] for k in match}
+                    start.update(update)
+                elif val is not None:  # name isn't in existing, use default if set
+                    start[name] = env_var_value(val, sub)
         elif val is not None:  # don't set if val is None
-            start[name] = _env_var_value(val)
+            start[name] = env_var_value(val, sub)
     return start
 
 
@@ -664,6 +743,8 @@ required_envvars = [
     "UNFURL_APPROVE",
     "UNFURL_MOCK_DEPLOY",
     "UNFURL_LOGFILE",
+    "UNFURL_VAULT_SKIP_DECRYPT",
+    "UNFURL_VALIDATION_MODE",
 ]
 # hack for sphinx ext documentedlist
 _sphinx_envvars = [(i,) for i in required_envvars]
