@@ -7,12 +7,12 @@ import json
 import os
 from typing import Union, Tuple, List
 import six
-from six.moves import urllib
+import urllib
+from urllib.parse import urljoin, urlsplit
+import ssl
+import certifi
+
 
-try:
-    from urlparse import urljoin  # Python2
-except ImportError:
-    from urllib.parse import urljoin  # Python3
 pathname2url = urllib.request.pathname2url
 from jsonschema import RefResolver
 
@@ -22,6 +22,7 @@ from ruamel.yaml.representer import RepresenterError, SafeRepresenter
 from ruamel.yaml.constructor import ConstructorError
 
 from .util import (
+    filter_env,
     to_bytes,
     to_text,
     sensitive_str,
@@ -42,6 +43,7 @@ from .merge import (
     restore_includes,
 )
 from .repo import is_url_or_git_path
+from .logs import getLogger
 from toscaparser.common.exception import URLException, ExceptionCollector
 from toscaparser.utils.gettextutils import _
 import toscaparser.imports
@@ -50,9 +52,8 @@ from toscaparser.repositories import Repository
 from ansible.parsing.vault import VaultLib, VaultSecret
 from ansible.utils.unsafe_proxy import AnsibleUnsafeText, AnsibleUnsafeBytes
 
-import logging
+logger = getLogger("unfurl")
 
-logger = logging.getLogger("unfurl")
 
 CLEARTEXT_VAULT = VaultLib(secrets=[("cleartext", b"")])
 
@@ -102,9 +103,9 @@ def represent_sensitive_bytes(dumper, data):
 
 def _construct_vault(constructor, node, tag):
     value = constructor.construct_scalar(node)
+    if os.getenv("UNFURL_VAULT_SKIP_DECRYPT"):
+        return sensitive_str.redacted_str
     if not constructor.vault.secrets:
-        if os.getenv("UNFURL_VAULT_SKIP_DECRYPT"):
-            return sensitive_str.redacted_str
         raise ConstructorError(
             context=None,
             context_mark=None,
@@ -189,15 +190,22 @@ def make_vault_lib_ex(secrets: List[Tuple[str, Union[str, bytes]]]):
     return VaultLib(secrets=vault_secrets)
 
 
+def urlopen(url):
+    return urllib.request.urlopen(
+        url, context=ssl.create_default_context(cafile=certifi.where())
+    )
+
+
 _refResolver = RefResolver("", None)
 
 
 class ImportResolver(toscaparser.imports.ImportResolver):
-    def __init__(self, manifest, ignoreFileNotFound=False, expand=False):
+    def __init__(self, manifest, ignoreFileNotFound=False, expand=False, config=None):
         self.manifest = manifest
         self.ignoreFileNotFound = ignoreFileNotFound
         self.loader = manifest.loader
         self.expand = expand
+        self.config = config or {}
 
     def __getstate__(self):
         state = self.__dict__.copy()
@@ -205,6 +213,12 @@ class ImportResolver(toscaparser.imports.ImportResolver):
         return state
 
     def get_repository(self, name, tpl):
+        if self.manifest:
+            # don't create another Repository instance
+            repo_view = self.manifest.repositories.get(name)
+            if repo_view:
+                return repo_view.repository
+
         if isinstance(tpl, dict) and "url" in tpl:
             url = tpl["url"]
             if (
@@ -213,6 +227,21 @@ class ImportResolver(toscaparser.imports.ImportResolver):
                 # convert to ssh://user@server/project.git
                 url = "ssh://" + url.replace(":", "/", 1)
             tpl["url"] = url
+
+        if self.manifest and tpl.get('credential'):
+            credential = tpl['credential']
+            # support expressions to resolve credential secrets
+            if self.manifest.rootResource:
+                from .eval import map_value
+                tpl['credential'] = map_value(credential, self.manifest.rootResource)
+            elif self.manifest.localEnv:
+                # we're including or importing before we finished initializing
+                # update os.environ so get_env works
+                context = self.manifest.localEnv.get_context(self.config.get("environment"))
+                if context.get("variables"):
+                    os.environ.update(filter_env(context["variables"]))
+                tpl['credential'] = self.manifest.localEnv.map_value(credential)
+
         return Repository(name, tpl)
 
     def get_url(self, importLoader, repository_name, file_name, isFile=None):
@@ -249,12 +278,13 @@ class ImportResolver(toscaparser.imports.ImportResolver):
             isFile = True
             if bare:
                 # XXX support empty filePath or when filePath is a directory -- need to checkout the tree
-                if (
-                    not filePath
-                    or repo.repo.rev_parse(revision + ":" + filePath).type != "blob"
-                ):
+                if not filePath:
                     raise UnfurlError(
-                        f"can't retrieve local repository for '{path}' with revision '{revision}'"
+                        f"local repository for '{path}' can not checkout revision '{revision}'"
+                    )
+                elif repo.repo.rev_parse(revision + ":" + filePath).type != "blob":
+                    raise UnfurlError(
+                        f"can't retrieve '{filePath}' with revision '{revision}' from local repository for '{path}'"
                     )
                 importLoader.stream = six.StringIO(repo.show(filePath, revision))
                 path = os.path.join(filePath, file_name or "")
@@ -262,11 +292,12 @@ class ImportResolver(toscaparser.imports.ImportResolver):
                 path = os.path.join(repo.working_dir, filePath, file_name or "").rstrip(
                     "/"
                 )
-                repository = None
-                if repository_name:
-                    repository = Repository(
-                        repository_name, importLoader.repositories[repository_name]
-                    )
+                if repo_view:
+                    assert not repo_view.repo
+                    repo_view.repo = repo
+                elif repository_name:
+                    # first time seeing this repository
+                    repository = self.get_repository(repository_name, importLoader.repositories[repository_name])
                     self.manifest.add_repository(repo, repository, filePath)
         elif file_name:
             path = os.path.join(path, file_name)
@@ -294,7 +325,7 @@ class ImportResolver(toscaparser.imports.ImportResolver):
                         else:
                             f = codecs.open(path, encoding="utf-8", errors="strict")
                     else:
-                        f = urllib.request.urlopen(path)
+                        f = urlopen(path)
                 except urllib.error.URLError as e:
                     if hasattr(e, "reason"):
                         msg = _(
@@ -311,7 +342,10 @@ class ImportResolver(toscaparser.imports.ImportResolver):
                                 'The server "%(path)s" couldn\'t fulfill the request. '
                                 + 'Error code: "%(code)s".'
                             )
-                        ) % {"path": path, "code": e.code}
+                        ) % {
+                            "path": path,
+                            "code": e.code,
+                        }  # type: ignore
                         ExceptionCollector.appendException(URLException(what=msg))
                         return
                     else:
@@ -408,12 +442,29 @@ class YamlConfig:
             raise UnfurlError(msg, True)
 
     def load_yaml(self, path, baseDir=None, warnWhenNotFound=False):
-        path, sep, fragment = path.partition("#")
-        path = os.path.abspath(os.path.join(baseDir or self.get_base_dir(), path))
-        if warnWhenNotFound and not os.path.isfile(path):
-            return path, None
-        logger.trace("attempting to load YAML file: %s", path)
-        with open(path, "r") as f:
+        url = urlsplit(path)
+        if url.scheme.startswith("http") and url.netloc:  # looks like an absolute url
+            fragment = url.fragment
+            logger.trace("attempting to load YAML url: %s", path)
+            try:
+                f = urlopen(path)
+            except urllib.error.URLError:
+                if warnWhenNotFound:
+                    logger.warning(f"document include {path} could not be retreived")
+                    return path, None
+                raise
+        else:
+            path, sep, fragment = path.partition("#")
+            path = os.path.abspath(os.path.join(baseDir or self.get_base_dir(), path))
+            if warnWhenNotFound and not os.path.isfile(path):
+                logger.warning(
+                    f"document include {path} does not exist (base: {baseDir})"
+                )
+                return path, None
+            logger.trace("attempting to load YAML file: %s", path)
+            f = open(path, "r")
+
+        with f:
             config = self.yaml.load(f)
         if fragment and config:
             return path, _refResolver.resolve_fragment(config, fragment)
@@ -431,7 +482,7 @@ class YamlConfig:
     def dump(self, out=sys.stdout):
         try:
             self.yaml.dump(self.config, out)
-        except:
+        except Exception:
             raise UnfurlError(f"Error saving {self.path}", True)
 
     def save(self):
@@ -499,7 +550,7 @@ class YamlConfig:
         output = six.StringIO()
         try:
             self.yaml.dump(template, output)
-        except:
+        except Exception:
             raise UnfurlError(f"Error saving include {path}", True)
         with open(path, "w") as f:
             f.write(output.getvalue())
@@ -507,7 +558,8 @@ class YamlConfig:
     def load_include(
         self, templatePath, warnWhenNotFound=False, expanded=None, check=False
     ):
-        if check:
+        if check and isinstance(check, bool):
+            # called by merge.has_template
             if self.loadHook:
                 return self.loadHook(
                     self,
@@ -533,11 +585,24 @@ class YamlConfig:
             key = templatePath["file"]
             when = templatePath.get("when")
             if when and not os.getenv(when):
-                return value, None, ''
+                return value, None, ""
         else:
             value = None
             key = templatePath
 
+        if self.loadHook:
+            # give loadHook change to transform key
+            key = self.loadHook(
+                self,
+                templatePath,
+                self.baseDirs[-1],
+                warnWhenNotFound,
+                expanded,
+                key,
+            )
+            if not key:
+                return value, None, ""
+
         if key in self._cachedDocIncludes:
             path, template = self._cachedDocIncludes[key]
             baseDir = os.path.dirname(path)
@@ -552,17 +617,18 @@ class YamlConfig:
                 )
             else:
                 path, template = self.load_yaml(key, baseDir, warnWhenNotFound)
-            newBaseDir = os.path.dirname(path)
+
             if template is None:
-                if warnWhenNotFound:
-                    logger.warning(
-                        f"document include {templatePath} does not exist (base: {baseDir})"
-                    )
-                return value, template, newBaseDir
-            elif isinstance(template, CommentedMap):
+                return value, template, baseDir
+
+            if path.startswith("http"):  # its an url, don't change baseDir
+                newBaseDir = baseDir
+            else:
+                newBaseDir = os.path.dirname(path)
+            if isinstance(template, CommentedMap):
                 template.base_dir = newBaseDir
             _cache_anchors(self.config._anchorCache, template)
-        except:
+        except Exception:
             raise UnfurlError(
                 f"unable to load document include: {templatePath} (base: {baseDir})",
                 True,
