@@ -16,11 +16,11 @@ from .util import UnfurlError, to_yaml_text, filter_env
 from .merge import patch_dict, intersect_dict
 from .yamlloader import YamlConfig, make_yaml
 from .result import serialize_value, ChangeRecord
-from .support import ResourceChanges, Defaults, Imports, Status
+from .support import ResourceChanges, Defaults, Status
 from .localenv import LocalEnv
 from .lock import Lock
-from .manifest import Manifest
-from .tosca import ArtifactSpec
+from .manifest import Manifest, relabel_dict
+from .tosca import ArtifactSpec, find_env_vars
 from .runtime import TopologyInstance
 from .eval import map_value
 from .planrequests import create_instance_from_spec
@@ -102,7 +102,7 @@ def save_status(operational, status=None):
         readyState["effective"] = operational.status.name
     if operational.state is not None:
         readyState["state"] = operational.state.name
-    if operational.priority:  # and operational.priority != Defaults.shouldRun:
+    if operational.priority is not None:
         status["priority"] = operational.priority.name
     status["readyState"] = readyState
 
@@ -127,7 +127,7 @@ def save_result(value):
         return value
 
 
-def save_task(task):
+def save_task(task, skip_result=False):
     """
     Convert dictionary suitable for serializing as yaml
       or creating a Changeset.
@@ -165,7 +165,7 @@ def save_task(task):
     if task.result:
         if task.result.outputs:
             output["outputs"] = save_result(task.result.outputs)
-        if task.result.result:
+        if task.result.result and not skip_result:
             output["result"] = save_result(task.result.result)
     else:
         output["result"] = "skipped"
@@ -174,31 +174,6 @@ def save_task(task):
     return output
 
 
-def relabel_dict(context, localEnv, key):
-    connections = context.get(key)
-    if not connections:
-        return {}
-    contexts = {}
-    if localEnv:
-        project = localEnv.project or localEnv.homeProject
-        if project:
-            contexts = project.contexts
-
-    # handle items like newname : oldname to rename merged connections
-    def follow_alias(v):
-        if isinstance(v, str):
-            env, sep, name = v.partition(":")
-            if sep:  # found a ":"
-                v = contexts[env][key][name]
-            else:  # look in current dict
-                v = connections[env]
-            return follow_alias(v)  # follow
-        else:
-            return v
-
-    return dict((n, follow_alias(v)) for n, v in connections.items())
-
-
 class ReadOnlyManifest(Manifest):
     """Loads an ensemble from a manifest but doesn't instantiate the instance model."""
 
@@ -229,9 +204,9 @@ class ReadOnlyManifest(Manifest):
         if context_inputs:
             inputs.update(context_inputs)
         spec["inputs"] = inputs
-        self.update_repositories(
-            manifest, relabel_dict(self.context, localEnv, "repositories")
-        )
+        # _update_repositories might not have been called while parsing
+        # call it now to make sure we set up the built-in repositories
+        self._update_repositories(manifest)
 
     @property
     def uris(self):
@@ -280,7 +255,7 @@ class ReadOnlyManifest(Manifest):
 def clone(localEnv, destPath):
     clone = ReadOnlyManifest(localEnv=localEnv)
     config = clone.manifest.config
-    for key in ["status", "changes", "lastJob"]:
+    for key in ["status", "changes", "lastJob", "lock"]:
         config.pop(key, None)
     if "metadata" in config:
         config["metadata"].pop("uri", None)
@@ -319,27 +294,28 @@ class YamlManifest(ReadOnlyManifest):
             manifest.get("spec", {}).get("deployment_blueprints") or {}
         )
         if deployment_blueprint:
-            if deployment_blueprint not in deployment_blueprints:
-                raise UnfurlError(
-                    f"Can not find requested deployment blueprint: '{deployment_blueprint}' is missing from the ensemble."
-                )
+            self._add_deployment_blueprint_template(
+                deployment_blueprints, deployment_blueprint, more_spec
+            )
             logger.info('Using deployment blueprint "%s"', deployment_blueprint)
-            resource_templates = deployment_blueprints[deployment_blueprint].get("resource_templates")
-            if resource_templates:
-                node_templates = more_spec["topology_template"]["node_templates"]
-                self._load_resource_templates(resource_templates, node_templates, False)
         elif deployment_blueprints:
-            logger.warning("This ensemble contains deployment blueprints but none were specified for use.")
+            logger.warning(
+                "This ensemble contains deployment blueprints but none were specified for use."
+            )
         if self.context.get("instances"):
-            # add context instances to spec instances
+            # add context instances to spec instances but skip ones that are just in there because they were shared
+            env_instances = {
+                k: v
+                for k, v in self.context["instances"].items()
+                if "imported" not in v
+            }
             self._load_resource_templates(
-                self.context["instances"], spec.setdefault("instances", {}), True
+                env_instances, spec.setdefault("instances", {}), True
             )
-
         self._set_spec(spec, more_spec, skip_validation)
         assert self.tosca
         if self.localEnv:
-            msg = f'Loading ensemble in environment "{self.localEnv.manifest_context_name}"'
+            msg = f'Loading ensemble "{self.path}" in environment "{self.localEnv.manifest_context_name}"'
             if self.tosca.topology and self.tosca.topology.primary_provider:
                 msg += f' with a primary_provider of type "{self.tosca.topology.primary_provider.type}"'
             logger.info(msg)
@@ -354,14 +330,10 @@ class YamlManifest(ReadOnlyManifest):
 
         self.lastJob = manifest.get("lastJob")
 
-        self.imports = Imports()
-        self._importedManifests = {}
-
         if localEnv:
             for name in ["locals", "secrets"]:
-                self.imports[name.rstrip("s")] = localEnv.get_local_instance(
-                    name, self.context
-                )
+                instance, local_spec = localEnv.get_local_instance(name, self.context)
+                self.imports.add_import(name.rstrip("s"), instance, local_spec)
 
         rootResource = self.create_topology_instance(status)
 
@@ -375,6 +347,32 @@ class YamlManifest(ReadOnlyManifest):
         self._configure_root(rootResource)
         self._ready(rootResource)
 
+    def _add_deployment_blueprint_template(
+        self, deployment_blueprints, deployment_blueprint, more_spec
+    ):
+        if deployment_blueprint not in deployment_blueprints:
+            raise UnfurlError(
+                f"Can not find requested deployment blueprint: '{deployment_blueprint}' is missing from the ensemble."
+            )
+        deployment_blueprint_tpl = deployment_blueprints[deployment_blueprint]
+        resource_templates = deployment_blueprint_tpl.get("resource_templates")
+        resourceTemplates = deployment_blueprint_tpl.get("resourceTemplates")
+        if resourceTemplates is not None:
+            # resourceTemplates and ResourceTemplate keys exist when imported from json
+            resource_templates = {}
+            local_resource_templates = (
+                deployment_blueprint_tpl.get("ResourceTemplate") or {}
+            )
+            for template_name in resourceTemplates:
+                if template_name in local_resource_templates:
+                    resource_templates[template_name] = local_resource_templates[
+                        template_name
+                    ]
+
+        if resource_templates:
+            node_templates = more_spec["topology_template"]["node_templates"]
+            self._load_resource_templates(resource_templates, node_templates, False)
+
     def _configure_root(self, rootResource):
         rootResource.imports = self.imports
         if (
@@ -434,7 +432,7 @@ class YamlManifest(ReadOnlyManifest):
                 env = os.environ.copy()
             for rel in root.requirements:
                 t = lambda datatype: datatype.type == "unfurl.datatypes.EnvVar"
-                env.update(rel.merge_props(t, True))
+                env.update(rel.merge_props(find_env_vars, True))
             intersect_dict(os.environ, env)  # remove keys not in env
             os.environ.update(env)
             paths = self.localEnv and self.localEnv.get_paths()
@@ -444,7 +442,7 @@ class YamlManifest(ReadOnlyManifest):
                 )
                 logger.debug("PATH set to %s", os.environ["PATH"])
 
-        # self.load_external_ensemble("localhost", tpl)
+        # self.load_external_ensemble("localhost", tpl) # declared in templates/home/unfurl.yaml.j2
         importsSpec = self.context.get("external", {})
         # note: external "localhost" is defined in UNFURL_HOME's context by convention
         for name, value in importsSpec.items():
@@ -486,7 +484,8 @@ class YamlManifest(ReadOnlyManifest):
         """
         :manifest: artifact template (file and optional repository name)
         :instance: "*" or name # default is root
-        :schema: # expected schema for attributes
+        :schema: expected schema for attributes
+        :url: uri of manifest
         """
         # load the manifest for the imported resource
         location = value.get("manifest")
@@ -504,9 +503,16 @@ class YamlManifest(ReadOnlyManifest):
         else:
             # ensemble is in the same project
             baseDir = getattr(location, "base_dir", self.get_base_dir())
-            artifact = ArtifactSpec(location, path=baseDir, spec=self.tosca)
+            artifact_tpl = dict(file=location["file"])
+            if "repository" in location:
+                artifact_tpl = location["repository"]
+            artifact = ArtifactSpec(artifact_tpl, path=baseDir, spec=self.tosca)
             path = artifact.get_path()
-            localEnv = LocalEnv(path, parent=self.localEnv)
+            localEnv = LocalEnv(
+                path,
+                parent=self.localEnv,
+                override_context=location.get("environment", ""),
+            )
             if self.path and os.path.abspath(self.path) == os.path.abspath(
                 localEnv.manifestPath
             ):
@@ -530,7 +536,7 @@ class YamlManifest(ReadOnlyManifest):
             raise UnfurlError(
                 f"Can not import external ensemble '{name}': instance '{rname}' not found"
             )
-        self.imports[name] = (resource, value)
+        self.imports.add_import(name, resource, value)
         self._importedManifests[id(root)] = importedManifest
 
     def load_changes(self, changes, changeLogPath):
@@ -617,13 +623,13 @@ class YamlManifest(ReadOnlyManifest):
         # save computed values for properties as they were observed
         if resource._properties:
             status["properties"] = resource._properties
-        if resource.shadow:
-            # name will be the same as the import name
-            status["imported"] = resource.name
+        if resource.imported:
+            status["imported"] = resource.imported
         save_status(resource, status)
         if resource.created is not None:
             status["created"] = resource.created
-
+        if resource.protected is not None:
+            status["protected"] = resource.protected
         return (resource.name, status)
 
     def save_requirement(self, resource):
@@ -652,8 +658,10 @@ class YamlManifest(ReadOnlyManifest):
 
     def save_resource(self, resource, discovered):
         # XXX checkstatus break unit tests so skip mostly
-        checkstatus = (resource.template.type == "unfurl.nodes.LocalRepository"
-                        or "default" in resource.template.directives)
+        checkstatus = (
+            resource.template.type == "unfurl.nodes.LocalRepository"
+            or "default" in resource.template.directives
+        )
         ret = self._save_entity_if_instantiated(resource, checkstatus)
         if not ret:
             return ret
@@ -791,7 +799,11 @@ class YamlManifest(ReadOnlyManifest):
         jobRecord = self.save_job_record(job)
         if job.workDone:
             self.manifest.config["lastJob"] = jobRecord
-            changes = list(map(save_task, job.workDone.values()))
+            exclude_result = not self.changeLogPath
+            # don't save result.results into this yaml, it might contain sensitive data
+            changes = list(
+                map(lambda t: save_task(t, exclude_result), job.workDone.values())
+            )
             if self.changeLogPath:
                 self.manifest.config["jobsLog"] = self.changeLogPath
 
@@ -913,10 +925,14 @@ class YamlManifest(ReadOnlyManifest):
             f.write(job.log(attrs))
 
             for change in changes:
-                status = change["readyState"]
+                if "readyState" not in change:
+                    continue  # never ran (skipped)
+                status = change["readyState"].get("effective") or change[
+                    "readyState"
+                ].get("local")
                 attrs = dict(
                     previousId=change.get("previousId", ""),
-                    status=status.get("effective") or status.get("local"),
+                    status=status,
                     target=change["target"],
                     operation=change["implementation"]["operation"],
                 )
